#!/bin/bash
exec guile -e main -s "$0" "$@"
!#

(use-modules (openssl))

(load "lib.scm")

(define die (lambda (message)
  (display message (current-error-port)) (newline (current-error-port))
  (exit 1)
))

(define main (lambda (args)
  (define privatekey-loaded #f)
  (define privatekey #f)
  (define ca-loaded #f)
  (define cert-loaded #f)
  (define port-string #f)
  (define rest-args '())

  (define master-context (or (SSL_CTX_new (SSLv23_method)) (die "unable to allocate SSL context")))

  ; >128 bits, require authentication, specifically include RC4-SHA, sort by strength
  (SSL_CTX_set_cipher_list master-context "HIGH:!ADH:RC4-SHA:@STRENGTH")

  (let helper ((cursor (cdr args)))
    (if (not (null? cursor))
      (cond
        ((string=? "ca" (car cursor))
          (SSL_CTX_load_verify_locations master-context (cadr cursor) (caddr cursor))
          (set! ca-loaded #t)
          (helper (cdddr cursor))
        )
        ((string=? "pkey" (car cursor))
          (SSL_CTX_use_PrivateKey_file master-context (cadr cursor) SSL_FILETYPE_PEM) 
          (set! privatekey-loaded #t)
          (helper (cddr cursor))
        )
        ((string=? "cert" (car cursor))
          (SSL_CTX_use_certificate_file master-context (cadr cursor) SSL_FILETYPE_PEM)
          (set! cert-loaded #t)
          (helper (cddr cursor))
        )
        ((string=? "port" (car cursor))
          (set! port-string (cadr cursor))
          (helper (cddr cursor))
        )
        (else
          (display "UNKNOWN ARGUMENT\n")
          (set! rest-args cursor) 
        )
      )
    )
  )
  
  (if (not privatekey-loaded)
    (let ()
      (define pkey (EVP_PKEY_new))
      (define rsakey (RSA_generate_key 1024 #x10001))
      (EVP_PKEY_set1_RSA pkey rsakey)
      (SSL_CTX_use_PrivateKey master-context pkey)
      (set! privatekey pkey)
    )
  )
  (if (not cert-loaded)
    (let ()
      (define cert (make-selfsigned-cert privatekey))
      (SSL_CTX_use_certificate master-context cert)
    )
  )

  (let ()
    (let ()
      (define port (string->number port-string))
      (define client-socket (socket AF_INET SOCK_STREAM 0))
      (connect client-socket AF_INET (car (hostent:addr-list (gethostbyname "localhost"))) port)
      (display "Client bound at: ") (display (vector-ref (getsockname client-socket) 2)) (newline)
      (let ()
        (define ssl (SSL_new master-context))
        (SSL_set_fd ssl (fileno client-socket))
        (display "CLIENT ABOUT TO CONNECT\n")
        (let ()
          (define ssl-accepted (SSL_connect ssl))
          (define sslp (SSL->port ssl))
          (display "SSL connected: ") (display ssl-accepted) (newline)
          (if (equal? ssl-accepted -1)
            (let ((error (SSL_get_error ssl ssl-accepted)))
              (format #t "CLIENT CONNECT ERROR: ~a\n" error)
            )
          )
          ; What can we learn about the server?
          (cond
            ((and (eq? (SSL_get_verify_result ssl) X509_V_OK) (SSL_get_peer_certificate ssl))
              (display "SERVER AUTHENTICATED WITH A CERTIFICATE\n")
            )
            ((eq? (SSL_get_verify_result ssl) X509_V_OK)
              (display "SERVER DID NOT PRESENT A CERTIFICATE\n")
            )
            (else (let ()
              (define peer (SSL_get_peer_certificate ssl))
              (define peername "......................................")
              (display "SERVER MUST HAVE BOTCHED THEIR CERT\n")
              (display peer) (newline)
              (display (X509_get_subject_name peer)) (newline)
              (display (X509_NAME_get_text_by_NID (X509_get_subject_name peer) NID_commonName peername #f)) (newline)
              (display "SERVER SUBJECT NAME IS: ") (display peername) (newline)
            ))
          )

          (if (null? rest-args)
            ; we are operating in interactive mode
            (let helper ()
              (define local-readval (read (current-input-port)))
              (cond
                ((eof-object? local-readval)
                )
                (else
                  (write local-readval sslp)
                  (display " " sslp)
                  (let ()
                    (define remote-readval (read sslp))
                    (display remote-readval) (newline) 
                  )
                  (helper)
                )
              )
            )
            ; else we have a form on the cmdline
            (begin
              (write (car rest-args) sslp)
              (display " " sslp)
              (display (read sslp)) (newline)
            )
          )
        )
      )
    )
  )
))
