#!/bin/bash
exec guile -e main -s "$0" "$@"
!#

;
; This is an example REPL server for Guile that demonstrates the use of the OpenSSL module
;

(use-modules (openssl))

(display "SSL_ERROR_NONE: ") (display SSL_ERROR_NONE) (newline)
(display "SSL_ERROR_ZERO_RETURN: ") (display SSL_ERROR_ZERO_RETURN) (newline)
(display "SSL_ERROR_WANT_READ: ") (display SSL_ERROR_WANT_READ) (newline)
(display "SSL_ERROR_WANT_WRITE: ") (display SSL_ERROR_WANT_WRITE) (newline)
(display "SSL_ERROR_SSL: ") (display SSL_ERROR_SSL) (newline)

(load "lib.scm")

(define die (lambda (message)
  (display message (current-error-port)) (newline (current-error-port))
  (exit 1)
))

(define main (lambda (args)
  (define privatekey-loaded #f)
  (define privatekey #f)
  (define ca-loaded #f)
  (define cert-loaded #f)
  (define port-string #f)
  (define rest-args '())

  (define master-context (or (SSL_CTX_new (SSLv23_method)) (die "unable to allocate SSL context")))
  ; >128 bits, require authentication, specifically include RC4-SHA, sort by strength
  (SSL_CTX_set_cipher_list master-context "HIGH:!ADH:RC4-SHA:@STRENGTH")

  (let helper ((cursor (cdr args)))
    (if (not (null? cursor))
      (cond
        ((string=? "ca" (car cursor))
          (SSL_CTX_load_verify_locations master-context (cadr cursor) (caddr cursor))
          (set! ca-loaded #t)
          (helper (cdddr cursor))
        )
        ((string=? "pkey" (car cursor))
          (SSL_CTX_use_PrivateKey_file master-context (cadr cursor) SSL_FILETYPE_PEM) 
          (set! privatekey-loaded #t)
          (helper (cddr cursor))
        )
        ((string=? "cert" (car cursor))
          (SSL_CTX_use_certificate_file master-context (cadr cursor) SSL_FILETYPE_PEM)
          (set! cert-loaded #t)
          (helper (cddr cursor))
        )
        ((string=? "port" (car cursor))
          (set! port-string (cadr cursor))
          (helper (cddr cursor))
        )
        ((string=? "verify" (car cursor))
          (SSL_CTX_set_verify master-context SSL_VERIFY_PEER)
          (helper (cdr cursor))
        )
        (else
          (display "UNKNOWN ARGUMENT\n")
          (set! rest-args cursor) 
        )
      )
    )
  )

  (if (not privatekey-loaded)
    (let ()
      (define pkey (EVP_PKEY_new))
      (define rsakey (RSA_generate_key 1024 #x10001))
      (EVP_PKEY_set1_RSA pkey rsakey)
      (SSL_CTX_use_PrivateKey master-context pkey)
      (set! privatekey pkey)
    )
  )
  (if (not cert-loaded)
    (let ()
      (define cert (make-selfsigned-cert privatekey))
      (SSL_CTX_use_certificate master-context cert)
    )
  )

  (let ()
    (define server-socket (socket AF_INET SOCK_STREAM 0))
    (define port (if port-string (string->number port-string) 0))

    (if (not server-socket) (die "unable to create server socket"))
    (bind server-socket AF_INET INADDR_ANY port)
    (display "Bound at: ") (display (vector-ref (getsockname server-socket) 2)) (newline)
    (listen server-socket 1)
    (let accept-helper ()
      (define connection-socket-pair (accept server-socket))
      (let ()
        (define connection-socket (car connection-socket-pair))
        (display "Accepted ") (display connection-socket) (newline)
        (display "FD IS: ") (display (fileno connection-socket)) (newline)
        (let ()
          (define ssl (SSL_new master-context))
          (SSL_set_fd ssl (fileno connection-socket))
          (let ()
            (define ssl-accepted (SSL_accept ssl))
            (define sslp (SSL->port ssl))
            (define read-failure (cons 'read 'failure))
            (display "SSL accepted: ") (display ssl-accepted) (newline)
            (if (equal? ssl-accepted -1)
              (let ((error (SSL_get_error ssl ssl-accepted)))
              (format #t "ACCEPT FAILS WITH ~a\n" (case error
                ((SSL_ERROR_NONE) "SSL_ERROR_NONE")
                ((SSL_ERROR_ZERO_RETURN) "SSL_ERROR_ZERO_RETURN")
                ((SSL_ERROR_WANT_READ) "SSL_ERROR_WANT_READ")
                ((SSL_ERROR_WANT_WRITE) "SSL_ERROR_WANT_WRITE")
                ((SSL_ERROR_WANT_CONNECT) "SSL_ERROR_WANT_CONNECT")
                ((SSL_ERROR_WANT_ACCEPT) "SSL_ERROR_WANT_ACCEPT")
                ((SSL_ERROR_WANT_X509_LOOKUP) "SSL_ERROR_WANT_X509_LOOKUP")
                ((SSL_ERROR_SYSCALL) "SSL_ERROR_SYSCALL")
                ((SSL_ERROR_SSL) "SSL_ERROR_SSL")
                (else (string-append "UNKNOWN: " (object->string error)))
              ))
              )
            )

            ; What can we learn about the client?
            (cond
              ((and (eq? (SSL_get_verify_result ssl) X509_V_OK) (SSL_get_peer_certificate ssl))
                (display "CLIENT AUTHENTICATED WITH A CERTIFICATE\n")
                (let ()
                  (define peer (SSL_get_peer_certificate ssl))
                  (define peername "......................................")
                  (display "CLIENT CERT:\n")
                  (display peer) (newline)
                  (display (X509_get_subject_name peer)) (newline)
                  (display (X509_NAME_get_text_by_NID (X509_get_subject_name peer) NID_commonName peername #f)) (newline)
                  (display "CLIENT SUBJECT NAME IS: ") (display peername) (newline)
                )
              )
              ((eq? (SSL_get_verify_result ssl) X509_V_OK)
                (display "CLIENT DID NOT PRESENT A CERTIFICATE\n")
              )
              (else (let ()
                (define peer (SSL_get_peer_certificate ssl))
                (define peername "......................................")
                (display "CLIENT MUST HAVE BOTCHED THEIR CERT\n")
                (display peer) (newline)
                (display (X509_get_subject_name peer)) (newline)
                (display (X509_NAME_get_text_by_NID (X509_get_subject_name peer) NID_commonName peername #f)) (newline)
                (display "CLIENT SUBJECT NAME IS: ") (display peername) (newline)
              ))
            )

            (let () 
              (define http-server-socket (socket AF_INET SOCK_STREAM 0))
              (define p2s #f)
              (define s2p #f)

              (display "CONNECT TO THE REAL SERVER SAYS: ") (display (connect http-server-socket AF_INET (car (hostent:addr-list (gethostbyname "localhost"))) 2021)) (newline)
              (display "HTTP SERVER SOCKET: ") (display http-server-socket) (newline)
 
              (let proxy-helper ()
                (define selectval (SSL-select
                  (delete #f (list (and (not p2s) sslp)
                                   (and (not s2p) http-server-socket)))
                  (delete #f (list (and p2s http-server-socket)
                                   (and s2p sslp)))
                  '()
                ))

                ;(display "PROXY SELECTVAL: ") (display selectval) (newline)

                (cond
                  ((member sslp (cadr selectval))
                    (write-char s2p sslp) (force-output sslp)
                    (set! s2p #f)
                    (proxy-helper)
                  )
                  ((member http-server-socket (cadr selectval))
                    (write-char p2s http-server-socket) (force-output sslp)
                    (set! p2s #f)
                    (proxy-helper)
                  )
                  ((member sslp (car selectval))
                    (set! p2s (read-char sslp))
                    ;(format #t "P2S: ~a\n" p2s)
                    (if (not (eof-object? p2s))
                      (proxy-helper)
                    )
                  )
                  ((member http-server-socket (car selectval))
                    (set! s2p (read-char http-server-socket))
                    ;(format #t "S2P: ~a\n" s2p)
                    (if (not (eof-object? s2p))
                      (proxy-helper)
                    )
                  )
                  (else
                    (display "NOBODY READY!\n")
                    (sleep 1)
                    (proxy-helper)
                  )
                )
              )

              (display "PROXY LOOP DONE!\n")
              (close http-server-socket)
              (close sslp)
              (close connection-socket)
              (accept-helper)
       
            )

          )
        )
      )
    )
  )
))
