;
; benchmarking routines
;
(library (proverb bench)
  (export
    elapsed
    benchmark-scheme-thunk-quiet
    benchmark-scheme-thunk
    svg-bench-leak
    svg-bench-all
  )
  (import
    (rnrs)
    (proverb doc)
    (proverb xml)
    (proverb xhtml)
    (proverb dialect)
    (ice-9 format)
    (only (guile)
      gettimeofday
      gc
      gc-stats
      make-list
      delete
    )
  )

(define elapsed (with-docstring
  "run <thunk> and return the number of elapsed microseconds"
  (lambda (thunk)
  (let ()
    (define start (gettimeofday))
    (thunk)
    (let ()
      (define end (gettimeofday))
      (let ()
        (define usecs (- (+ (* 1000000 (car end)) (cdr end))
          (+ (* 1000000 (car start)) (cdr start))))
        usecs
      )
    )
  )
  )
))

(define benchmark-scheme-thunk-quiet (with-docstring
  "run <thunk> <reps> times and return the a list of performance statistics"
  (lambda (thunk reps)
    ; get us as clean as possible
    (gc)
    (let ()
      (define stats (gc-stats))
      (let ()
        (define usecs (elapsed (lambda () (let helper ((c reps))
          (if (eq? c 0)
            #t
            (begin
              (thunk)
              (helper (- c 1))
            )
          )
        ))))
        (let ()
          (define new-stats (gc-stats))
          (let ()
            (define mean-usecs (/ usecs reps))
            (define mean-cells (/ (- (cdr (assoc 'total-cells-allocated new-stats)) (cdr (assoc 'total-cells-allocated stats))) reps))
            (define gc-count (- (cdr (assoc 'gc-times new-stats)) (cdr (assoc 'gc-times stats))))
            ; we caused the top GC ourselves
            (list mean-usecs mean-cells gc-count)
          )
        )
      )
    )
  )
))

(define benchmark-scheme-thunk (with-docstring
  "run <thunk> <reps> times and print a report to the current output port. description text <desc> is included in the report."
  (lambda (thunk reps desc)
  (format #t "Benchmarking a thunk (run ~a times) for ~a\n" reps desc)
  (let ()
    (define res (benchmark-scheme-thunk-quiet thunk reps))
    (format #t ".. ~a microseconds (mean), cells (mean) ~a across ~a gcs\n"
                (exact->inexact (list-ref res 0))
                (exact->inexact (list-ref res 1))
                (list-ref res 2)
    )
  )
  )
))

(define svg-bench-leak (with-docstring
  "(svg-bench-leak thunk inner-reps outer-reps desc) -- run THUNK for leak
checking. A system measurement is taken OUTER-REPS times. thunk is run INNER-REPS times during each measurement interval to better amortize the impact of the testing apparatus. The results are returned as an SVG."
  (lambda (thunk inner-reps outer-reps desc)
  (define results (map (lambda (v)
    (benchmark-scheme-thunk-quiet thunk inner-reps)
  ) (make-list outer-reps)))
  (let ()
    (define max-usecs (exact->inexact (apply max (map car results))))
    (define min-usecs (exact->inexact (apply min (map car results))))
    (define max-cells (exact->inexact (apply max (map cadr results))))
    (define min-cells (exact->inexact (apply min (map cadr results))))
    (with-xml-tag "<div>"
      (make-text (format #f "Usecs: ~a to ~a Cells: ~a to ~a\n" min-usecs max-usecs min-cells max-cells))
      (with-xml-tag "<br>")
      (with-xml-tag "<div style='position: relative; height: 400px; left: 0px; top; 0px'>"
        (with-xml-tag "<div style='position: absolute; left: 0px; top 0px'>"
          (svgwrap (svgboxchart (map exact->inexact (map car results))))
        )
        (with-xml-tag "<div style='position: absolute; left: 0px; top 0px'>"
          (svgwrap (svglinechart (map exact->inexact (map cadr results))))
        )
        (with-xml-tag "<div style='position: absolute; left: 0px; top 0px'>"
          (svgwrap (svgdotchart outer-reps (delete #f (index-map (lambda (i v) (if (> v 0) i #f)) (map caddr results)))))
        )
      )
    )
  )
  )
))

(define svg-bench-all (with-docstring
  "a varargs function (reps . thunks) that benchmarks each of thunks using reps inner repetitions for averaging"
  (lambda (reps . thunks)
  (define results (map (lambda (v)
    (benchmark-scheme-thunk-quiet v reps)
  ) thunks))
  (let ()
    (define max-usecs (exact->inexact (apply max (map car results))))
    (define min-usecs (exact->inexact (apply min (map car results))))
    (define max-cells (exact->inexact (apply max (map cadr results))))
    (define min-cells (exact->inexact (apply min (map cadr results))))
    (with-xml-tag "<div>"
      (make-text (format #f "Usecs: ~a to ~a Cells: ~a to ~a\n" min-usecs max-usecs min-cells max-cells))
      (with-xml-tag "<br>")
  (with-xml-tag "<div style='position: relative; height: 400px; left: 0px; top; 0px'>"
    (with-xml-tag "<div style='position: absolute; left: 0px; top 0px'>"
      (svgwrap (svgboxchart (map exact->inexact (map car results))))
    )
    (with-xml-tag "<div style='position: absolute; left: 0px; top 0px'>"
      (svgwrap (svglinechart (map exact->inexact (map cadr results))))
    )
  )
    )
  )
  )
))

)
