(library (proverb channels)
  (export
    make-channel channel?  channel-enqueue channel-enqueue-async
    channel-does-select?  channel-select-pre
    channel-select-post channel-open channel-close
    channel-get-desc make-fport-channel
    channel-make-blank
    channel-blank?
  )
  (import
    ; base Scheme
    (rnrs)
    ; Proverb specializations
    (proverb doc) (proverb help) (proverb debug)
    ; Non-portables
    (only (guile) port-closed? fileno close)
  )

;
; Documented in the channels article
;


(define channel-marker '(channel channel))

(define channel-blank-marker '(channel-blank channel-blank))

;
; a channel is an opaque object. use the accessors
;
(define make-channel (with-wiki-docstring
  "make a channel value. See the article on [[channels | channels]] for more info"
  (lambda (enqueue has-queue? select-pre select-post open close desc)
    (vector channel-marker enqueue has-queue? select-pre select-post open close desc)
  )
))

(define channel? (with-wiki-docstring
  "Is value a channel value? See the article on [[channels | channels]] for more info"
  (lambda (value)
    (and (vector? value) (eq? (vector-length value) 8) (eq? (vector-ref value 0) channel-marker))
  )
))

(define channel-enqueue (with-wiki-docstring
  "Enqueue value in channel with the [[tire|TIRE mux]] mux. See the article on [[channels | channels]] for more info. value will be actually transmitted at some indeterminate time."
  (lambda (channel mux value)
    ((vector-ref channel 1) mux value #f)
  )
))

(define channel-enqueue-async (with-wiki-docstring
  "Enqueue value in channel with the [[tire|TIRE mux]] mux. See the article on [[channels | channels]] for more info. continuation, a thunk, will be called when value is actually transmitted."
  (lambda (channel mux value continuation)
    ((vector-ref channel 1) mux value continuation)
  )
))

(define channel-does-select? (with-docstring
  "return #t if <channel> needs to participate in select multiplexing"
  (lambda (channel)
    (and (vector-ref channel 4) #t)
  )
))

(define channel-select-pre (with-wiki-docstring
  "run the run-select-loop select-pre function from the channel <channel> for TIRE mux <mux>. See [[channels]] for more information about channels and [[[(help run-select-loop)|here]]] for more info about run-select-loop."
  (let ()
    (define empty-response (list '() '() '()))
    (lambda (channel mux)
      (define handler (vector-ref channel 3))
      (if handler
        (handler mux)
        empty-response
      )
    )
  )
))

(define channel-select-post (with-wiki-docstring
   "run the run-select-loop select-post function from the channel <channel> for TIRE mux <mux>. See [[channels]] for more information about channels and [[[(help run-select-loop)|here]]] for more info about run-select-loop."
  (lambda (channel mux rfds wfds xfds)
    (define handler (vector-ref channel 4))
    (if handler
      (handler mux rfds wfds xfds)
      #t
    )
  )
))

(define channel-open (with-wiki-docstring
  "initialize the channel <channel>. See [[channels]] for more information."
  (lambda (channel)
    (define handler (vector-ref channel 5))
    (if handler
      (handler)
      #t
    )
  )
))

(define channel-close (with-wiki-docstring
  "finalize the channel <channel>. See [[channels]] for more information."
  (lambda (channel)
    (define handler (vector-ref channel 6))
    (if handler
      (handler)
      #t
    )
  )
))

(define channel-get-desc (with-docstring
  "return the description for channel <channel>"
  (lambda (channel)
    (or
      (vector-ref channel 7)
      "an anonymouns channel"
    )
  )
))

;
; Some example channel types
;

(define please-just-close (lambda (p)
  (catch #t (lambda () (close p)) (lambda args #f))
  ; close on guile can fail with errors if there is unflushed data
))

(define make-fport-channel (with-docstring
  "create a channel backed by the input port in-port, the output port out-port, reader function reader and writer function writer. Enqueued Forms will be buffered if they cannot immediately be written."
  (lambda (in-port out-port reader writer enqueue)
    (define buffer '())    ; this is where we read from
    (define buffer-end #f) ; this is where we write to
    (define write-helper (lambda (wfds)
      (if (and (not (null? buffer)) 
               (not (port-closed? out-port))
               (member out-port wfds))
        (let ()
          (define r (writer (caar buffer) out-port)) 
          (if (eof-object? r)
            (begin
              (please-just-close in-port)
              (please-just-close out-port)
              #f
            )
            (let ((form-continuation (cdar buffer)))
              (and form-continuation (form-continuation))
              (if (eq? buffer buffer-end) ; this is only true when we have just written the last form in the buffer
                (begin
                  (set! buffer '())
                  (set! buffer-end #f)
                )
                (set! buffer (cdr buffer))
              )
              #t
            )
          )
        )
        #t
      )
    ))
    (let ()
      (define read-helper (lambda (rfds mux)
        (if (and
              (not (port-closed? in-port))
              (member in-port rfds))
          (let ((next-char (peek-char in-port)))
            (if (eof-object? next-char)
              (begin
                (please-just-close in-port)
                (please-just-close out-port)
                #f
              )
              ;prevent whitespace from gumming up the works
              ; is this kosher? debugging with nc is useful.
              ; if this breaks other stuff it could be an
              ; enabled flag. --oabarne
              ; should TIRE have strong opinions about whitespace?  
              ;(if (or (eq? #\space next-char)
              ;        (eq? #\newline next-char)
              ;        (eq? #\return next-char)
              ;        (eq? #\tab next-char))
              ;  (read-char in-port)
              (let ()
                (define r (reader in-port))
                (if (eof-object? r)
                  (begin
                    (please-just-close in-port)
                    #f
                  )
                  (begin
                    (debugmsg "BOGUS ENQUEUE?!? in fport channel read helper\n")
                    (if (channel-blank? r)
                      (critmsg "SKIPPING BLANK!\n")
                      (enqueue mux c r)
                    )
                    (if (port-closed? in-port)
                      (begin
                        (please-just-close in-port)
                        #f
                      )
                      (if (char-ready? in-port)
                        (read-helper rfds mux)
                        #t
                      )
                    )
                  )
                )
              )
              ;)
            )
          )
          (if (or (port-closed? in-port) (port-closed? out-port))
            (begin
              (critmsg "channel is partly shut down -- probably asynchronously!")
              (please-just-close in-port)
              (please-just-close out-port)
              #f
            )
            ; we had nothing to do
            #t
          )
        )
      ))
      (define il (list in-port))
      (define ol (list out-port))
      (define c (make-channel
        ; enqueue
        (lambda (mux value continuation)
          (define c (cons (cons value continuation) '()))
          (if buffer-end
            (set-cdr! buffer-end c)
            (set! buffer c)
          )
          (set! buffer-end c)
        )
        ; has-queue
        (lambda ()
          (if (null? buffer)
            #f
            #t
          )
        )
        ; select-pre
        (lambda (mux)
          (if (and (port-closed? in-port) (port-closed? out-port))
            ; this channel is over
            #f
            ; else
            (begin
              (if (or (port-closed? in-port) (port-closed? out-port))
                (critmsg "about to select-pre for an fport channel on a closed port in-port closed?: ~a out-port closed?: ~a!\n" (port-closed? in-port) (port-closed? out-port))
              )
              (list il
                    (if (null? buffer) '() ol)
                    '()
              )
            )
          )
        )
        ; select-post
        (lambda (mux rfds wfds xfds)
          (and
            (write-helper wfds)
            (read-helper rfds mux)
          )
        )
        ; open
        #f
        ; close
        (lambda ()
          (please-just-close in-port)
          (please-just-close out-port)
          #t
        )
        ; description
        "scheme fport"
      ))

      c
    )
  )
))

;
; another very simple channel (write only)
; 
(define make-local-channel (with-docstring
  "create and return a channel that dispatches enqueued forms to <writer>. The new channel does not buffer forms, participate in select multiplexing, or provide initialization or finalization."
  (lambda (writer)
    (make-channel
      ; enqueue
      (lambda (mux value continuation)
        (writer value)
        (and continuation (continuation))
      ) 
      ; has-queue
      #f
      ; select-pre
      #f
      ; select-post
      #f
      ; open 
      #f
      ; close
      #f
      ; desc
      "local channel"
    )
  )
))

(define channel-make-blank (lambda ()
  channel-blank-marker
))

(define channel-blank? (lambda (v)
  (eq? v channel-blank-marker)
))

)
