
(library (proverb concurrent)
  (export
    make-concurrent-with-talkback
    make-concurrent
    concurrent-apply
    concurrent-apply-with-talkback
    run-inferior-process
    prefork-hook
    postfork-hook
    proverb-fork)
  (import
    (rnrs)
    (proverb types)
    (proverb dialect)
    (proverb doc)
    (proverb debug)
    (guile)
    (ice-9 format)
  )

(define prefork-hook (make-hook 0))
(define postfork-hook (make-hook 1))

(define proverb-fork (with-docstring
  "Wrap the system fork. execute prefork-hook before forking and postfork-hook after forking. postfork-hook gets called with with either 'parent or 'child or #f (if fork fails). This call returns the value returned by system fork."
  (lambda ()
    (run-hook prefork-hook)
    (let ((v (primitive-fork)))
      (if (equal? v 0)
        (run-hook postfork-hook 'child)
        (if (> v 0)
          (run-hook postfork-hook 'parent)
          (run-hook postfork-hook #f)
        )
      )
      v
    )
  )
))

(define make-concurrent-with-talkback (with-docstring
  "accept <sub> and return a new procedure that will cause <sub> to be run in a fresh process on each invocation. The new procedure will additionally substitute a talkback port as an argument to <sub> for any argument that is eq? equal to <talkback-token>. For this reason, <talkback-token> should probably be a non-immediate value. The parent caller is not provided with the result of the evaluation of <sub>. Instead, the caller receives one end of a full-duplex pipe connected to the child's talkback port. This routine is most commonly used to start long-running REPL procedures in a child process. The child process will not generally have access to the parent's ports unless passed as arguments to this function. The child process will terminate upon completion of <sub>."
  (lambda-w-types ((sub procedure?) (talkback-token #t))
    ; one end of sp will be open in parent and child
    (define sp (socketpair AF_UNIX SOCK_STREAM 0))

    ; our ultimate return value
    (define new-lambda (lambda args
      ; let's not burden our child with our garbage.
      (gc)
      (run-hook after-gc-hook)
      ; or with our pending I/O
      ; flush-all-ports has complained in the past about flushing
      ; closed ports
      (catch #t flush-all-ports always-false)
      (let ()
        (define child-pid (proverb-fork))
        (cond 
          ((eq? child-pid 0) (let ()
            (define saved-module (current-module))
 	    ; do our talkback-token substitution 
            (define args-copy (map (lambda (v)
              (if (eq? v talkback-token) (cons (cdr sp) (cdr sp)) v)
            ) args))
	    ; Throw away our ports. We ditch most ports except our talkback port and any ports
	    ; passed as arguments.

            (close (car sp)) ; -- this should be captured by the below
	    (let ()
              ; note that (cdr sp) is explicitly added to the list- the port pair is filtered from the args-copy
              (define protected-ports (append (delete #f (map (lambda (v) (and (port? v) v)) args-copy)) (list (cdr sp) (current-output-port))))
              (define protected-descriptors '(0 1 2))
              (port-for-each (lambda (p)
                ; guile 1.8 can deliver an environment here in some cases... Strange.
                ; this observed behavior may have been the result of guile heap corruption.
                (and
                  (port? p) (file-port? p) (not (member p protected-ports)) (not (port-closed? p))
                  (catch ; fileno throws an error on the wrong type of ports
                    #t
                    (lambda ()
                      (define fd (fileno p))
                      (or
                        (member fd protected-descriptors)
                        (begin
                          (close-fdes fd)
                          (close p)
                        )
                      )
                    )
                    always-false
                  )
                )
              ))
            )

            (call-with-dynamic-root
              ; thunk
              (lambda ()
	        (set-current-module saved-module)
                (apply sub args-copy)
                (kill (getpid) SIGKILL)
              )
              ; handler
	      (lambda args
                (kill (getpid) SIGKILL)
              )
            )
          ))
          ; was there an error
          ((eq? child-pid -1) (error "bogo"))
          ; else I'm the parent
          (else
            ; This guardian stuff is a work-around for guile not closing ports on gc.
            (let ((port-pair (cons (car sp) (car sp))))
              (define magic (vector 'concurrent-endpoint port-pair child-pid (getpid)))
              (close (cdr sp))
              (connection-guardian magic)
              ; attach a reference to magic so that magic gets gc'd 
              ;  at the same time as the port (only need to reference (car sp) since
              ;  guile ports are full-duplex 
              (set-object-property! (car sp) 'port-guardian magic)
              (debugmsg "Delivering the port: ~a properties=~a\n" port-pair (object-properties port-pair))
              port-pair
            )
          )
        )
      )
    ))
    (if (procedure-property sub 'args)
      (set-procedure-property! new-lambda 'args (procedure-property sub 'args))
    )
    new-lambda
  )
))

(define make-concurrent (with-docstring
  "accept <sub> and return a new procedure that will cause <sub> to be run in a fresh process on each invocation. The return value is unspecified. This routine is most commonly used to start a 'fire and forget' action."
  (lambda (sub)
    ; using a private non-immediate keeps us safe
    (define concurrent (make-concurrent-with-talkback sub (cons 1 1)))
    (let ()
      (define new-lambda (lambda args
        ; this returns a port pair which we ignore
        (apply concurrent args)
        ; and then we return nothing useful
        *unspecified*
      ))
      (if (procedure-property sub 'args)
        (set-procedure-property! new-lambda 'args (procedure-property sub 'args))
      )
      new-lambda
    )
  )
))

; a list of pairs of the form ( <parent-pid> . <child-pid> )
(define delinquent-children '())

; the connection guardian expects magic values
(define connection-guardian (with-docstring
  "a Scheme guardian used for private handles generated by make-concurrent-with-talkback."
  (make-guardian)))

(define concurrent-apply (with-docstring
  "Accept a subroutine <sub> run it with <args> in a child process"
  (lambda (sub args)
    (apply (make-concurrent sub) args)
  )
))

(define concurrent-apply-with-talkback (with-docstring
  "Accept a subroutine <sub> run it with <args> in a child process. substitute <talkback-token>, if in <args>, with a talkback port"
  (lambda (sub args talkback-token)
  (apply (make-concurrent-with-talkback sub talkback-token) args)
  )
))

(define run-inferior-process (with-docstring
  "Use the UNIX shell to run the command <shellfrag> in a child process. Return a full-duplex port connected to the standard input and standard output of the child."
  (lambda-w-types ((shellfrag string?))
  (define magic-talkback-val '(talkback))
  (let ()
    (define port-pair (concurrent-apply-with-talkback (lambda (iopair)
  	  (dup2 (port->fdes (car iopair)) 0)
	  (dup2 (port->fdes (cdr iopair)) 1)
          (execl "/bin/sh" "sh" "-c" shellfrag)
    ) (list magic-talkback-val) magic-talkback-val))
    port-pair
  )
  )
))

; non-define actions
(add-hook! after-gc-hook (let ()
  (define helper (lambda (children)
    (define val (connection-guardian))
    (if val
      (let ()
        (define port (car (vector-ref val 1)))
	(define child (vector-ref val 2))
	(define parent (vector-ref val 3))

        (debugmsg "deferred cleanup from make-concurrent")

	(if (eq? parent (getpid))
          (begin
            (debugmsg "MAGIC CLOSING A PORT OF OURS: ~a\n" port) 
            ; the port may have already been closed
            (catch #t (lambda () (shutdown port 2)) (lambda args #f))
	    (catch #t (lambda () (close port)) (lambda args #f))
            (debugmsg "MAGIC STILL HERE:\n") 
	    ; we may not have a child -- this connection may be to an inferior proverb.
            (if child
              (helper (cons child children))
              (helper children)
            )
          )
          (begin
            (debugmsg "MAGIC CLOSING A PORT SOMEONE ELSES: ~a\n" port) 
            ; we should close the port anyway (but not shutdown)...
            (close port)
            (helper children)
          )
        )
      )
      ; else
      (let ()
        (define subhelper (lambda (c tries)
          (if (and (not (null? c)) (< tries 3))
            (let ()
	      ; waitpid should never error here. We are only calling waitpid
	      ; in the case that we have a set of child pids we are looking
	      ; for... An error in waitpid means that the list is stale.
              (subhelper (delete #f (map (lambda (v)
                ;(critmsg "WAITING FOR PID ~a\n" v)
                (catch #t (lambda ()
                  (define waitval (waitpid v WNOHANG))
                  (debugmsg "WAITED FOR PID ~a: ~a\n" v waitval)
                  (if (eq? (car waitval) 0) v #f)
                ) (lambda args
                  (debugmsg "WAITPID FAILS")
                  #f
                ))
              ) c)) (+ tries 1))
            )
	    ; else
	    (set! delinquent-children (map (lambda (v) (cons (getpid) v)) c))
          )
        ))
        (if (null? children)
          (debugmsg "connection-guardian: no more children to reap\n" children)
          (debugmsg "connection-guardian: leaving with outstanding children: ~a\n" children)
        )
        (subhelper (append children (delete #f (map (lambda (v) (if (eq? (car v) (getpid)) (cdr v) #f)) delinquent-children))) 1)
        #f
      )
    )
  ))
  (lambda ()
    (helper '())
  )
))

)
