;
; Filesystem helpers...
;

(library (proverb filesystem)
  ;(export path-compose which path-expand path-decompose canonicalize-path build-path port->string filename->string support-dir)
  (export path-compose which path-expand path-decompose canonicalize-path build-path port->string filename->string)
  (import
    (rnrs files)
    (proverb doc)
    (proverb dialect)
    (ice-9 rdelim)
    (guile)
  )

;(define support-dir #f) ; we will set this later

(define path-compose (with-docstring
  "build a string path from a decomposed path"
  (lambda (dcpath)

  (define helper (lambda (dcp)
    (if (null? dcp)
      ""
      (if (string=? (car dcp) "/") 
        ; this an absolute path component
        (string-append (car dcp) (helper (cdr dcp)))
        ; this is a relative path component
        (if (null? (cdr dcp)) 
          ; avoid placing a trailing slash here
          (car dcp)
          (string-append (car dcp) "/" (helper (cdr dcp)))
        )
      )
    )
  ))
  (if (null? dcpath)
    "/"
    (helper dcpath)
  )
  )
))

(define path-decompose (with-docstring
  "build a decomposed path from a string path"
(lambda (spath)
    (define s (delete "" (studio-string-split spath #\/)))
    (if (and (> (string-length spath) 0) (eq? (string-ref spath 0) #\/))
     (cons "/" s) 
     s
    )  
  )
))

(define path-consolidate (with-docstring
  "accept a decomposed path and return a decomposed path devoid of extra slashes and dots"
  (lambda (dcpath)
    ; this routine is not tail recursive and could blow up the stack
    ; with a pathologically long path. No pun intended.
    (reverse (let helper ((dcp (reverse dcpath)))
      (if (null? dcp)
        dcp
        ; else
        (cond
          ; '.' is useless
          ((string=? (car dcp) ".")
            (helper (cdr dcp))
          )
          ; '..' skips a level 
          ((and (string=? (car dcp) "..") (> (length dcp) 1)) (let ()
            (cdr (helper (cdr dcp)))
          ))
          ((string=? (car dcp) "")
            (helper (cdr dcp))
          )
	  (else
            (cons (car dcp) (helper (cdr dcp)))
          )
        )
      )
    ))
  )
))

(define path-expand (with-docstring
  "accepts a decomposed path and returns a decomposed path after performing shell-like expansions for ~ ($HOME) and % (Proverb dir)"
  (lambda (dcpath)
    (cond
      ((null? dcpath) '())
      ((string=? (car dcpath) "~") (path-expand (append (path-decompose (getenv "HOME")) (cdr dcpath))))
      ((string=? (car dcpath) "%") (path-expand (append (path-decompose (getenv "STUDIO_DIRECTORY")) (cdr dcpath))))
      (else dcpath)
    )
  )
))

(define get-file-list (with-docstring 
  "Extract the list of files from the open directory <dir>"
(lambda (dir)

  (define helper (lambda (racc)
    (define val (readdir dir))
    (if (eof-object? val)
      (reverse racc)
      (helper (cons val racc))
    )
  ))
  (helper '())
)))

(define port->byte-list (with-docstring
  "return the ordered list of bytes contained in the input port specified by <port>"
  (lambda (port)

  (define helper (lambda (racc)
    (define char (read-char port))
    (if (eof-object? char)
      (begin
        (close port)
        (reverse racc)
      )
      (helper (cons char racc))
    )
  ))
  (helper '())
  )
))

(define port->string (with-docstring
  "return the contents of the file specified by <port> as a string"
  (lambda (path)
    (list->string (port->byte-list path))
  )
))


(define filename->byte-list (with-docstring
  "return the ordered list of bytes contained in the file specified by <path>"
  (lambda (path)
   
    (port->byte-list (open-input-file path))

  )
))

(define filename->string (with-docstring
  "return the contents of the file specified by <path> as a string"
  (lambda (path)
    (list->string (filename->byte-list path))
  )
))

(define which (with-docstring
  "accept str and return the first executable in PATH matching that specification. same rules as shell which"
  (lambda (str)
    (cond
      ((string=? str "") str)
      ((string-index str #\/) str)
      (else (let ()
        (define p (getenv "PATH"))
        (and p (let helper ((cursor (studio-string-split p #\:)))
          (if (null? cursor)
            #f
            (let ()
              (define trial (string-append (car cursor) "/" str))
              (if (access? trial X_OK)
                trial
                (helper (cdr cursor))
              )
            )
          )
        ))
      ))
    )
  )
))

(define mounts->tree (with-docstring
  "turn a list of system mounts into a tree"
  (lambda (mounts)
  (define sm (sort (delete #f (map (lambda (v)
    (if (string=? (list-ref v 2) "nfs")
      (list (delete "" (studio-string-split (list-ref v 1) #\/)) (list-ref v 2) (list-ref v 3) (list-ref v 0))
      (list (delete "" (studio-string-split (list-ref v 1) #\/)) (list-ref v 2) (list-ref v 3) (list-ref v 0)) 
    )
  ) mounts)) (lambda (v1 v2) (< (length v1) (length v2)))))

  (let helper ((cursor sm) (t (list "" #f #f '())))
    (if (null? cursor)
      t
      (let insert-helper ((c (caar cursor)) (st t))
        (if (null? c)
          (helper (cdr cursor) t)
          (let ()
            (define subtree (assoc (car c) (list-ref st 3)))
            (if subtree
              (insert-helper (cdr c) subtree)
              (let ()
                (define nst (if (null? (cdr c))
                  (list (car c) #t (car cursor) '())
                  (list (car c) #f #f '())
                ))
                (list-set! st 3 (cons nst (list-ref st 3)))
                (insert-helper (cdr c) nst)
              )
            )
          )
        )
      )
    )
  )
)))

(define resolve-mount (with-docstring
  "accept a mount tree and a canonicalized path list"
  (lambda (mt cpl)
  (let helper ((st mt) (cursor cpl) (best #f))
    (if (null? cursor)
      best
      (let ()
        (define v (member (car cursor) (map car (list-ref st 3))))
        (define v2 (assoc (car cursor) (list-ref st 3)))
        (if v
          (helper v2 (cdr cursor)  (list-ref v2 2))
          best
        )
      )
    )
  )
  )
))

(define path->mount (with-docstring
  "accept an absolute pathname and return the mount point that provides it"
  (lambda (path)
    (define mounts (read-mounts))
    (let ()
      (define mt (mounts->tree mounts))
      (let ()
        (define c (canonicalize-path path))
        (resolve-mount mt c)
      )
    )
  )
))

(define read-mounts (with-docstring
  "return the list of mounts on the system"
  (lambda ()
    ;
    ; sigh. /proc/mounts doesn't return ready for select.
    ;
    (define fname (string-append "/tmp/mount-results." (number->string (getpid))))
    (system (string-append "cp /proc/mounts " fname))
    (let ()
      (define f (open-input-file fname))
      (let helper ((racc '()))
        (define l (read-line f))
        (if (eof-object? l)
          (begin
            (close f)
            (delete-file fname)
            (reverse racc)
          )
          (helper (cons (studio-string-split l #\space) racc))
        )
      )
    )
  )
))

(define build-path (with-docstring
  "accept a list of path elements and assemble them into a pathname"
  (lambda (path-element-list)
    (cond
      ((null? path-element-list) "")
      ((equal? path-element-list '("/")) "/")
      ((equal? (car path-element-list) "/")
        (string-merge (cons "" (cdr path-element-list)) #\/)
      )
      (else
        (string-merge path-element-list #\/)
      )
    )
  )
))

(define canonicalize-path (with-docstring
  "accept a pathname <pathname> and return a canonical version as a list using readlink"
  (lambda (pathname)
    (define pl (path-decompose pathname))
    (let helper ((cursor (reverse pl)) (resolved '()))
      (if (null? cursor)
        resolved
        (let ()
          (define lv (catch #t (lambda () (readlink (build-path (reverse cursor)))) always-false))
          (if lv
            (if (eq? (string-ref lv 0) #\/)
              (cons "/" (helper (delete "" (reverse (studio-string-split lv #\/))) resolved))
              (helper (append (delete "" (reverse (studio-string-split lv #\/))) (cdr cursor)) resolved)
            )
            (helper (cdr cursor) (cons (car cursor) resolved))
          )
        )
      )
    )
  )
))

(define umount (with-docstring
  "like the C call. It's a pity that guile doesn't provide this."
  (lambda (dirname)
    (system (string-append "umount -f '" dirname "'"))
  )
))

)
