;
; Tracing and debugging forms for Proverb under Guile
;
;
(library (proverb debug)
  (export proverb-verbose-debugging? debug-port debugmsg critmsg trace! trace-args! trace-ret! trace-args-ret!! trace-depth!! trace-through!)
  (import (rnrs) (proverb doc) (guile))

(define proverb-verbose-debugging? #f)
(define emergency-port (car (fdes->ports 2)))
(define sneaky-emergency-port emergency-port)

(define debug-port (with-docstring
  "return a port suitable for writing debugging messages"
  (lambda ()
    emergency-port
  )
))

(define debugmsg (with-docstring
  "dump str with optional format arguments to the Proverb debugging log. message is suppressed if proverb-verbose-debugging? is false." 
  (lambda (str . args)
    (and proverb-verbose-debugging?
      (apply format sneaky-emergency-port (string-append "proverb-debug (pid=~a) (time=~a): " str) (getpid) (strftime "%Y/%m/%d-%H:%M:%S" (localtime (current-time))) args)
    )
  )
))

(define critmsg (lambda (str . args)
  "dump str with optional format arguments to the Proverb debugging log. message is sent to the log even if proverb-verbose-debugging? is false."
  (apply format sneaky-emergency-port (string-append "proverb-debug (pid=~a) (time=~a): " str) (getpid) (strftime "%Y/%m/%d-%H:%M:%S" (localtime (current-time))) args)
))

;
;
; foo -> resolves foo
; bar:foo -> resolves the public symbol foo in module bar
; bar::foo -> resolves the private symbol foo in module bar
(define trace-resolve (lambda (sym mod)
  (if (string? sym)
    (trace-resolve (string->symbol sym) mod)
  (catch 'unbound-variable (lambda ()
    (cons (lambda (newval) (eval `(,set! ,sym ,newval) mod)) (eval sym mod))
  ) (lambda args 
      (define s (string-split (symbol->string sym) #\:))
      (if (> (length s) 1)
        (let ()
          (define modname (with-input-from-string (car s) read))
          (if (null? (cdr s))
            (begin
              #f
            )
            (if (equal? (cadr s) "")
              (let ()
                (define modval (resolve-module modname))
                (let* ((n (string->symbol (string-join (cddr s) ":")))
                       (v (module-variable modval n)))
                    (cons
                      (lambda (newval)
                        (variable-set! v newval)
                      )
                      (variable-ref v)
                    )
                )
              )
              (let ()
                (define modval (resolve-interface modname))
                (let* ((n (string->symbol (string-join (cdr s) ":")))
                       (v (module-variable modval n)))
                  (cons
                    (lambda (newval)
                      (variable-set! v newval)
                    )
                    (variable-ref v)
                  )
                )
              )
            )
          )
        )
        #f
      ) 
    )
  )
  )
))

(define trace-update (lambda (sym mod val)
  (let ()
    (define v (trace-resolve sym mod))
    ((car v) val)
    val
  )
))

(define trace-wrap (lambda (proc)
  (define n (or (procedure-name proc) "anonymous"))
  (lambda args
    (define retcounter 0)
    (define old-debugval proverb-verbose-debugging?)
    (critmsg "TRACE |+ ~a\n" (procedure-name proc))
    (let ((r (apply proc args)))
      (set! retcounter (+ retcounter 1))
      (if (eq? retcounter 1)
        (critmsg "TRACE |- ~a\n" n) 
        (critmsg "TRACE |* ~a\n" n)
      )
      r
    )
  )
))

(define-macro (trace! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (,trace-update (quote ,proc) (,current-module) (,trace-wrap old-proc))
  )
)

(define-macro (trace-args! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (,trace-update (quote ,proc) (,current-module) (lambda args
      (define retcounter 0)
      (critmsg "TRACE |+ ~a: ~a\n" (procedure-name old-proc) args)
      (let ((r (apply old-proc args)))
        (set! retcounter (+ retcounter 1))
        (if (eq? retcounter 1)
          (critmsg "TRACE |- ~a\n" (procedure-name old-proc))
          (critmsg "TRACE |* ~a\n" (procedure-name old-proc))
        )
        r
      )
    ))
  )
)

(define-macro (trace-ret! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (,trace-update (quote ,proc) (,current-module) (lambda args
      (define retcounter 0)
      (critmsg "TRACE |+ ~a\n" (procedure-name old-proc))
      (dynamic-wind
        (lambda () (critmsg "TRACE |> ~a\n" (procedure-name old-proc)))
        (lambda ()
          (let ((r (apply old-proc args)))
            (set! retcounter (+ retcounter 1))
            (if (eq? retcounter 1)
              (critmsg "TRACE |- ~a: ~a\n" (procedure-name old-proc) r)
              (critmsg "TRACE |* ~a: ~a\n" (procedure-name old-proc) r)
            )
            r
          )
        )
        (lambda () (critmsg "TRACE |< ~a\n" (procedure-name old-proc)))
      )
    ))
  )
)

(define-macro (trace-through! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (,trace-update (quote ,proc) (,current-module) (lambda args
      (define retcounter 0)
      (critmsg "TRACE |+ ~a\n" (procedure-name old-proc))
      (let ((r (apply old-proc args)))
        (set! retcounter (+ retcounter 1))
        (if (eq? retcounter 1)
          (critmsg "TRACE |- ~a: ~a\n" (procedure-name old-proc) r)
          (critmsg "TRACE |* ~a: ~a\n" (procedure-name old-proc) r)
        )
        (if (procedure? r)
          (,trace-wrap r)
          (begin
            (critmsg "TRACE |! ~a: trace-through can't follow non-procedure retval\n" (procedure-name old-proc))
            r
          )
        )
      )
    ))
  )
)

(define-macro (trace-args-ret!! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (,trace-update (quote ,proc) (,current-module) (lambda args
      (define retcounter 0)
      (critmsg "TRACE |+ ~a: ~a\n" (procedure-name old-proc) args)
      (let ((r (apply old-proc args)))
        (set! retcounter (+ retcounter 1))
        (if (eq? retcounter 1)
          (critmsg "TRACE |- ~a: ~a\n" (procedure-name old-proc) r)
          (critmsg "TRACE |* ~a: ~a\n" (procedure-name old-proc) r)
        )
        r
      )
    ))
  )
)

(define-macro (trace-depth!! proc)
  `(let ()
    (define old-proc (,cdr (,trace-resolve (quote ,proc) (,current-module))))
    (define depth 0)
    (,trace-update (quote ,proc) (,current-module) (lambda args
      (define retcounter 0)
      (set! depth (+ depth 1))
      (critmsg "TRACE |+=~a ~a\n" depth (procedure-name old-proc))
      (let ((r (apply old-proc args)))
        (set! retcounter (+ retcounter 1))
        (if (eq? retcounter 1)
          (begin
            (critmsg "TRACE |-= ~a ~a: ~a\n" depth (procedure-name old-proc) r)
            (set! depth (- depth 1))
          )
          (critmsg "TRACE |*= ~a ~a: ~a\n" depth (procedure-name old-proc) r)
        )
        r
      )
    ))
  )
)

)
