(library (proverb unread)
  (export
    receive-value
    unread
  )
  (import
    (rnrs) (rnrs io simple)
    (proverb passval) (proverb doc) (proverb debug) (proverb dialect)
    (proverb channels)
    (guile modules)
    (only (guile)
      unspecified? record? variable? macro? promise?
      fluid? environment? closure? entity? stack? with-fluids*
      read-eval? read-hash-extend
      call-with-input-string struct? record-type?
      number->string
      force-output
      read-options read-enable read-disable
    )
    (ice-9 rw) (ice-9 rdelim)
  )

(define magic-write-predicate-pairs '())

(define binary-string-reader (let ()
  (define x-uvr! (lambda (v fd)
    (define len (string-length v))
    (let loop ((k 0))
      (if (>= k len)
        #t
        (let ()
          (define rv (read-string!/partial v fd k))
          (if (eof-object? rv)
            ; uniform-vector-read doesn't quite work in guile
            #f
            ; else
            (loop (+ k rv))
          )
        )
      )
    )
  ))
  (lambda (c port)
    (and (eq? (read-char port) #\#) (let ()
      (define k (read-delimited "#" port 'split))
      (if (eof-object? (cdr k))
        (cdr k)
        (let ()
          (define v (make-string (string->number (car k)) #\0))
          (x-uvr! v port)
          v
        )
      )
    ))
  )
))

; install the binary reader
(call-with-input-string "#b#3#foo" (lambda (ip)
  (catch #t (lambda ()
    (read ip)
  ) (lambda args
    (read-hash-extend #\b binary-string-reader)
  ))
  (close-input-port ip)
))

; for unspecified
(call-with-input-string "#U" (lambda (ip)
  (catch #t (lambda ()
    (read ip)
  ) (lambda args
    (read-hash-extend #\U (lambda args
      ; the Guile reader can't return actual unspecified
      '*unspecified*
    ))
  ))
))

;
; used for inline stringification
;
(read-hash-extend #\~ (lambda (c port)
  (define subordinate-form (read port))
  `(object->string ,subordinate-form)
))


;
; Pairs of the form ( <predicate> . <writer> )
;
(set! magic-write-predicate-pairs (append
  (list 
    ; things that are perfectly unreadable
    (cons null? write)
    (cons boolean? write)
    (cons symbol?  write)
    (cons unspecified? (lambda (val port) (display "#U " port)))
    (cons eof-object?  (lambda (val port) (display "#.the-eof-object " port)))
    (cons char? write)
    ; our special case for objects with associated serializers
    (cons passval? (lambda (val port) (passval-write val port inner-unread)))
    (cons global? (lambda (val port)
      (global-write val port inner-unread)
    ))
    ; and continue
    ;(cons string?  write)
    (cons string? (lambda (v p)
      (display (string-append "#b#" (number->string (string-length v) 10) "#") p)
      (display v p)
    ))
    (cons number?  write)
    ; meta-objects
    (cons list? (lambda (val port)
      (display "(" port)
      (inner-unread (car val) port)
      (for-each (lambda (v)
        (display " " port)
        (inner-unread v port)
      ) (cdr val))
      (display ")" port)
    ))
    (cons pair? (lambda (val port) (display "(" port) (inner-unread (car val) port) (display " . " port) (inner-unread (cdr val) port) (display ")" port)))
    (cons vector? (lambda (val port)
      (display " #(" port)
      (for-each (lambda (v)
        (display " " port) (inner-unread v port)
      ) (vector->list val)) 
      (display ")" port)
    ))
    ; relatively uncommon objects
    (cons record?  (lambda (val port) (write '*record* port)))
    (cons module?  (lambda (val port) (write '*module* port)))
    (cons variable?  (lambda (val port) (write '*variable* port)))
    (cons macro?  (lambda (val port) (write '*macro* port)))
    (cons promise?  (lambda (val port) (write '*promise* port)))
    (cons fluid? (lambda (val port) (write '*fluid* port)))
    (cons environment? (lambda (val port) (write '*environment* port)))
    (cons closure? (lambda (val port) (display "#." port) (inner-unread (procedure-and-environment-source val) port)))
    (cons procedure? (lambda (val port) (display "#." port) (display (procedure-name val) port) (display " " port)))
    (cons entity? (lambda (val port) (write '*entity* port)))
    (cons stack? (lambda (val port) (write '*stack* port)))
    (cons (lambda (v) #t) (lambda (val port) (write (string-append "bogo-" (object->string val)) port)))
  )
  magic-write-predicate-pairs
))

(define add-unreader (with-docstring
  "add an unreader function for values that match <predicate>"
  (lambda (predicate unreader)
    (set! magic-write-predicate-pairs (append magic-write-predicate-pairs (list (cons predicate unreader))))
  )
))

; I think the new unreader should go as far down the list as possible -- at
; the end or just before an unreader that would snarf its exemplar
(define add-unreader-with-exemplar (with-docstring
  "add an unreader function for values that match <predicate>. Ensure that this is the preferred unreader for <exemplar>"
(lambda (predicate unreader exemplar)
  (define new-pair (cons predicate unreader))
  (set! magic-write-predicate-pairs (let helper ((cursor magic-write-predicate-pairs) (racc '()))
    (if (null? cursor)
      (reverse (cons new-pair racc))
      ; else is there a conflict?
      (if ((caar cursor) exemplar)
        (append (reverse (cons new-pair racc)) cursor)
        ; else move on
        (helper (cdr cursor) (cons (car cursor) racc))
      )
    )
  ))
  )
))

(define inner-unread (with-docstring
  "(inner version) write bytes to port that will cause a read to return a reasonable representation of val."
  (let ()
    (define magic-write-find-writer (lambda (val)
      (let helper ((l magic-write-predicate-pairs))
        (if (null? l)
          #f
          (if ((caar l) val)
            (cdar l)
            (helper (cdr l))
          )  
        )
      )
    ))

    (lambda (val port)
      (define writer (magic-write-find-writer val))
      (if writer
        (writer val port)
        ; else
        (write val port)
      )
      ;(studio-flush-port port)
      (force-output port)
    )
  )
))

(define unread (with-docstring
  "write bytes to port that will cause a read to return a reasonable representation of val."
  (lambda (val port)
    (catch #t (lambda () (inner-unread val port)) (lambda args (eof-object)))
  )
))

(define eat-available-blanks (lambda (source)
  (let helper ((last #f))
    (if (char-ready? source)
      (case (peek-char source)
        ((#\space #\newline)
          (read-char source)
          (helper #t) 
        )
        (else last)
      )
      last
    )
  )
))

(define inner-receive-value (with-docstring
  "Obtain a scheme value from <source> using the Proverb remote protocol"
  (lambda (source)
    (with-fluids*  
      (list read-eval?) (list #t) 
      (lambda () 
        (define old-positions (member 'positions (read-options)))
        (read-disable 'positions)
        (let ((r (read source)))
          (if (char-ready? source) (debugmsg "RV: STILL READY\n")) 
          (if old-positions (read-enable 'positions))
          ; eat trailing whitespace
          (eat-available-blanks source)
          r
        )
      )
    )
  )
))

(define receive-value (with-docstring
  "Obtain a scheme value from <source> using the Proverb remote protocol"
  (lambda (source)
    (if (eat-available-blanks source)
      (if (char-ready? source)
        (inner-receive-value source)
        (channel-make-blank)
      )
      (inner-receive-value source)
    )
  )
))

)
