;
; Guile utilities (these should be functions that depend on guile)
;
(library (proverb guilefoo)
  (export
    module-get-symbol-list get-module-list procedure-ref delete-regexp run-select-loop-until procedure-and-environment-source predicate?
    try-use-modules
    eval-case
    process-use-modules
  )
  (import (guile) (ice-9 rdelim) (proverb dialect) (proverb debug) (proverb doc))

;(load "guile-debug.scm")

;
; Select all symbols from a module
;
(define module-get-symbol-list (with-docstring
  "return a list of the bindings in the module <module>. each element is of the form ( name . variable ) "
  (lambda (module)
    (define mo (module-obarray module))
    (hash-fold acons '() mo) 
  )
))

;
; this is quite tricky.
; 
(define procedure-ref (lambda (proc varname)

  "(procedure-ref <proc> <varname>) -- search the environment of <proc> for symbol <varname>"

  (let pr-helper ((el (procedure-environment proc)))
    (if (null? el)
      *unspecified*
      (if (pair? (car el))
        ; is this a define-style declaration?
        (if (list? (caar el))
          (let ()
            (define match (member varname (caar el)))
            (if match
              (let ()
                (define offset (- (length (caar el)) (length match)))
	        (list-ref (cdar el) offset)
              )
              (pr-helper (cdr el))
            )
          )
          ; else this is a single declaration
          (if (symbol? (caar el))
            (if (eq? varname (caar el))
              (cdar el)
              (pr-helper (cdr el))
            )
            ; else punt
            (pr-helper (cdr el))
          )
        )
        (if (procedure? (car el))
          (pr-helper (cdr el)) 
          ; otherwise just do the same thing?
          (pr-helper (cdr el))
        )
      )
    )
  )
))

;
; this is quite tricky. Same tricks as procedure-ref. Returns
; a source expression for the procedure and its environment.
; 

(define procedure-and-environment-source (with-docstring
  "(procedure-and-environment-source <proc>) -- returns a source expression for the procedure and its environment"
(lambda (proc)
  ; 
  ; Guile includes the eval closure in the environment
  ; (for closures defined in a module) and stores them
  ; newest first. We're creating a source
  ; expression and need newest last for label shadowing
  ; to work correctly.
  ;

  ; every environment element should be of the form
  ; ( <binding-name> . <binding-value> )
  ; or
  ; ( ( <binding-name-0> ... <binding-name-n> ) . ( <binding-value-0> ... <binding-value-n> ) )

  ; The two Scheme forms
  ; (let ((foo 5)) (lambda () foo))
  ; and
  ; (let () (define foo 5) (lambda () foo))
  ;
  ; should equivalent environments in Guile but do not quite.
  ; The first form would use ( foo . 5 ) to represent its environment
  ; and the second would choose ( ( foo ) . ( 5 ) ), printed as ( (foo) 5 ).

  (define env (procedure-environment proc))

  (let environment-helper ((el (reverse env)) (last-alias #f))
    (if (null? el)
      (if last-alias
        last-alias
        (procedure-source proc)
      )
      (if (pair? (car el))
        ; is this a define-style declaration?
        (if (list? (caar el))
          (if (null? (caar el))
            ; we can prune an unnecessary let
            (environment-helper (cdr el) last-alias)
            ; else there are bindings here
            (let frame-helper ((label-cursor (reverse (caar el)))
                               (value-cursor (reverse (cdar el)))
                               (la last-alias)
                               (acc '()))
              (if (null? label-cursor)
                `(let ()
                   ,@acc
                   ,(environment-helper (cdr el) la)
                )
                (if (eq? (car value-cursor) proc)
                  (frame-helper
                    (cdr label-cursor)
                    (cdr value-cursor)
                    (car label-cursor)
                    (cons (list 'define (car label-cursor) (if la
                      la
                      (procedure-source (car value-cursor))
                    )) acc)
                  )  
                  (frame-helper
                    (cdr label-cursor)
                    (cdr value-cursor)
                    la
                    (cons (list 'define (car label-cursor) (if (closure? (car value-cursor))
                      ; we could wind up with a cycle here
                      ;(procedure-and-environment-source (car value-cursor))
                      (procedure-source (car value-cursor))
                      (car value-cursor)
                    )) acc)
                  )  
                )
              )
            )
          )
          ; else this is a single declaration
          (if (symbol? (caar el))
            (if (eq? (cdar el) proc)
              (if last-alias
                `(let ()
                  ,@(list (list 'define (caar el) last-alias) (environment-helper (cdr el) (caar el)))
                )
                `(let ()
                  ,@(list (list 'define (caar el) (procedure-source proc)) (environment-helper (cdr el) (caar el)))
                )
              )
              (if (closure? (cdar el))
                `(let ()
                  ,@(list (list 'define (caar el) (list 'begin (format #f "bogus---symbol ~a is a closure" (caar el)) (procedure-source (cdar el)))) (environment-helper (cdr el) last-alias))
                )
                `(let ()
                  ,@(list (list 'define (caar el) (cdar el)) (environment-helper (cdr el) last-alias))
                )
              )
            )
            ; else 
            (environment-helper (cdr el) last-alias)
          )
        )
        ; else... we ignore (eval closures)
        (environment-helper (cdr el) last-alias)
      )
    )
  )
  )
))

(define procedure-summary (lambda (proc)
  "summarize a procedure"
  (define docs (procedure-documentation proc))
  (if (and docs (< (string-length docs) 25))
    docs
    (object->string (procedure-name proc))
  )
))

;
; split this code out
;

(define predicate? (lambda (val)
  "is val a predicate procedure?"
  (and
    (procedure? val)
    (equal? (procedure-property val 'arity) '(1 0 #f))
  )
))

(define delete-regexp (lambda (re lst)
  "delete elements from the list of strings <lst> that match the regular expression <re>"
  (define fail (cons 1 1))
  (if (null? lst)
    lst
    (if (string? re)
      (delete-regexp (make-regexp re) lst)
      (let helper ((cursor lst) (racc '()))
        (if (null? cursor)
          (reverse racc)
          (if (regexp-exec re (car cursor))
            (helper (cdr cursor) racc)
            (helper (cdr cursor) (cons (car cursor) racc))
          )
        )
      )
    )
  )
))

(define run-select-loop-until (with-docstring
  "run multiple services around a single select loop. services is of the form ((state-1 select-pre select-post) ...). body is run on each iteration. services run until their select-post routine returns false. This function runs until every service has a false state or until the thunk <check> returns false."
  (lambda (services body check)
    (define portcheck (lambda (v mesg)
      (if (port-closed? v)
        (begin
          ;(critmsg "run-select-loop-until: select on closed port (~a) addr=~a creating-pid=~a heritage=~a\n" mesg (object-address v) (object-property v 'creating-pid) (object-property v 'heritage))
          #f
        )
        v
      )
    ))
    (define safe-select (lambda (r w x t)
      (catch 'system-error (lambda ()
        (studio-select
          (delete #f (map (lambda (v) (portcheck v "read fd")) r))
          (delete #f (map (lambda (v) (portcheck v "write fd")) w))
          (delete #f (map (lambda (v) (portcheck v "except fd")) x))
          t
        )
      ) (lambda args
        (critmsg "select system error. woken by async? args: ~a\n" args)
        (list '() '() '())
      ))
    ))
    (define states (map car services))
    (define pres (map cadr services))
    (define posts (map caddr services))
    (define prehelper (lambda (sc prec)
      (if (null? sc)
        ; no fds
        (list '() '() '())
        ; else
        (apply (car prec) (cons (car sc) (prehelper (cdr sc) (cdr prec))))
      )
    ))
    (define posthelper (lambda (prestates sc postc racc)
      ; we need one select call per select-post routine -- but these
      ; inner selects run with no timeout. Since they run with no timeout
      ; they don't actually need all the prestates and we could do a better
      ; job about eliminating select-posts that are superfluous here
      ;(critmsg "posthelper\n")
      (let ()
      (define results (apply safe-select (append prestates '(0))))
      (if (null? sc)
        ; the empty list of new states
        (reverse racc)
        ; else
        (posthelper prestates (cdr sc) (cdr postc) (cons
          (apply (car postc) (cons (car sc) results))
          racc
        ))
      )
      )
    ))
    (define thin (lambda (ball states)
      "remove pres and posts for loops with dead states"
      (delete #f (map (lambda (b s)
        (and s b)
      ) ball states))
    ))
    ;(critmsg "run-select-loop-until::::\n")
    (let helper ((ball (map list states pres posts)))
      (if (null? ball)
        #f ; we have no more states to iterate
        (let ()
          (define prestates (prehelper (map car ball) (map cadr ball))) 
          (let ()
            (define results (apply safe-select (append prestates '(#f)))) 
            (and body (body))
            ; sigh. one post-helper might eat the descriptor intended for another. no more than one post-helper per select call
            (let ()
              ;(define rargs (thin ball (posthelper prestates (map car ball) (map caddr ball) '())))
              (define rargs (thin ball (posthelper results (map car ball) (map caddr ball) '())))
              (if (if check (check) #t)
                (helper rargs)
                #f
              )
            )
          )
        )
      )
    )
  )
))

(define make-half-duplex-forwarding-service (with-docstring
  "construct a run-select-loop service that forwards io from the Scheme port <from> to the Scheme port <to>"
  (lambda (from to)
    (define half-duplex-state (cons 0 (make-string 10000)))
    (define half-duplex-select-pre (lambda (state rfds wfds xfds)
        (list
          (if (eq? (car state) 0)
            (cons from rfds)
            rfds
          )  
          (if (eq? (car state) 0)
            wfds
            (cons to wfds)
          ) 
          xfds
        )
    ))
    (define half-duplex-select-post (lambda (state rfds wfds xfds)
      (cond
        ((and (member to wfds) (> (car state) 0)) (let ()
          (define wcount (write-string/full (cdr state) to 0 (car state)))
          (if (equal? wcount (car state)) 
            (begin
              (force-output to)
              (set-car! state 0)
              state
            )
            (begin
              (close-output-port to)
              #f
            )
          )
        ))
        ((and (member from rfds) (eq? (car state) 0)) (let ()
          (define count (studio-read-string!/partial (cdr state) from))
          (if count
            (begin
              (set-car! state count)
              state
            )
            (begin
              (close-input-port from)
              #f
            )
          )
        ))
        (else 
          (if (or (port-closed? from) (port-closed? to))
            #f ; should we close the other port for good measure?
            state
          )
        )
      )
    ))
    (list
      half-duplex-state
      half-duplex-select-pre
      half-duplex-select-post
    )
  )
))   

(define-macro (tilde-wrap function)
  (define s (gensym))
  `(let ()
    (define ,s ,function)
    (lambda (arg)
      (if (and (string? arg) (> (string-length arg) 1) (eq? (string-ref arg 0) #\~))
        (if (eq? (string-ref arg 1) #\/)
          (,s (string-append (getenv "HOME") (substring arg 1)))
          (let ()
            (define i (string-index arg #\/))
            (if i
              (let ()
                (define pw (getpwnam (substring arg 1 i)))
                (if pw
                  (,s (string-append (vector-ref pw 5) (substring arg i)))
                  (,s arg)
                )
              )
              (,s arg)
            )
          )
        )
        (,s arg)
      )
    )
  )
)

;; Make unicode lambda work. All Hail Guy Steele.
(define-macro (λ . body) `(lambda ,@body))

(define get-module-list (with-docstring
  "return the list of modules which Guile knows about. These modules may not all be available in your environment at the moment."
  (lambda ()
    (define alist '())
    (define lhelper (lambda (l a)
      (if (eq? l '())
        a
        (if (assoc (car l) a)
          (lhelper (cdr l) a)
          ; else
          (lhelper (cdr l) (helper (car l) a))
        )
      )
    ))
    (define helper (lambda (m a)
      (if (assoc m a)
        a ; already here
        ; else
        (lhelper (module-uses m) (cons (cons m '()) a))
      )
    ))
    (if (defined? 'get-interpreter-modules)
      (map car (lhelper (cons (current-module) (get-interpreter-modules)) alist))
      (map car (helper (current-module) alist))
    )
  )
))

(define ref-by-addr (with-docstring
  "retreive the first element with address <addr> from the list <l>. throw an error if no such element is found."
  (lambda (addr lst)
    (let helper ((cursor lst))
      (if (null? cursor)
        (error "address not found")
        (if (eq? (object-address (car cursor)) addr)
          (car lst)
          (helper (cdr cursor))
        )
      )
    )
  )
))

;
; I'm not sure how I feel about this
;
(set-procedure-property! throw 'has-side-effect 'no)

(define-macro (try-use-modules modspec)
  `(eval-case ((load-toplevel) (catch #t (lambda () (process-use-modules (list (list (quote ,modspec))))) (lambda  args #f))) (else (error "use-modules can only be used at the top level")))
)

(load "funcdocs.scm")

)


