(library (proverb help)
  ; we export current-module from Guile so that we can bind wiki articles to their providing
  ; module
  (export
    make-filesystem-wiki-ns
    search-help-articles help-index register-wiki-namespace
    wiki-put
    wiki-get
    wiki-search
    wiki-index 
    wiki-namespaces
  )
  (import
    (rnrs)
    (proverb doc)
    (proverb xml)
    (proverb regexp)
    (proverb filesystem)
    (proverb debug)
    (only (proverb dialect) dir->filelist always-false)
    (only (proverb bootstrap) support-dir)
    (guile)
  )

(define help-articles '())

(define make-get-fs-article-index (lambda (path)
  (lambda ()
    (define help-files (dir->filelist path))
    (delete #f (map (lambda (v)
      (define sl (string-length v))
      (let ()
        (define wname (if (> sl 5)
          (if (string=? (substring v (- sl 5) sl) ".wiki")
            (substring v 0 (- sl 5))
            #f
          ) 
          #f
        ))
        (if wname
          (let ()
            (define wsym (string->symbol wname))
            (let ()
              (define f (string-append path "/" wname ".wiki")) 
              (if (and (access? f R_OK) #t)
                (cons wsym f)
                #f
              )
            )
          )
          #f
        )
      )
    ) help-files))
  )
))

(define make-proverb-search-help-articles (lambda (path)
  (define get-fs-article-index (make-get-fs-article-index path))
  (lambda (query-string)
    (define help-files (get-fs-article-index))
    (define re (make-regexp (regexp-escape-string query-string)))
    (append
      ; incore
      (delete #f (map (lambda (article)
        (if (search-article article re)
          (car article)
          #f
        )
      ) help-articles))
      ; out-of-core
      (delete #f (map (lambda (v)
        (define wsym (car v))
        (define f (cdr v))
        (if (assoc wsym help-articles)
          #f ; don't search it if we already matched the in-core
          (if (regexp-exec re (symbol->string wsym))
            wsym
            (if (and (access? f R_OK) #t)
              (let ()
                (define s (filename->string f))
                (if (regexp-exec re s)
                  wsym
                  #f
                )
              )
              #f
            )
          )
        )
      ) help-files))
    )
  )
))

(define check-key (lambda (key)
  (define l (string->list (symbol->string key)))
  (let helper ((cursor l) (dotcount 0))
    (if (null? cursor)
      (if (equal? dotcount 2)
        #f
        #t
      )
      (if (eq? (car cursor) #\.)
        (helper (cdr cursor) (+ dotcount 1))
        (if (or (eq? (car cursor) #\/) (eq? (car cursor) #\\))
          (if (equal? dotcount 2)
            #f
            (helper (cdr cursor) 0)
          )
          (begin
            (helper (cdr cursor) dotcount)
          )
        )
      )
    )
  )
))

(define utf8-bom (list->string (map integer->char '(#xef #xbb #xbf))))

(define make-make-wiki-article (lambda (path)
  (with-wiki-docstring
  "create a wiki article in Proverb with name KEY and body ARTCLE-BODY. ARTICLE-BODY is HTML escaped and cannot contain mixed wiki and HTML elements."
  (lambda (key article-body)
    (if (check-key key)
      (cond
        ((string? article-body)
          (if (and (>= (string-length article-body) 3) (equal? (substring article-body 0 3) utf8-bom))
            (set! article-body (substring article-body 3))
          )
          (set! help-articles (cons (list key article-body 'wiki) (delete (assoc key help-articles) help-articles)))
          (if path
            (let ()
              (define mfn (string-append path "/" (symbol->string key) ".wiki"))
              (define update (lambda ()
                (catch #t (lambda ()
                  (let ((p (open-output-file mfn)))
                    (display article-body p)
                    (close p)
                    #t
                  )
                ) always-false)
              ))
              (if (access? mfn W_OK)
                (let rename-helper ((counter 1))
                  (let ((fn (string-append path "/" (symbol->string key) ".wiki." (number->string counter))))
                    (if (access? fn R_OK)
                      (rename-helper (+ counter 1))
                      (if (access? fn W_OK)
                        #f
                        (begin
                          (rename-file mfn fn)
                          (update)
                        ) 
                      )
                    )
                  )
                ) 
                (update)
              )
            )
            ; else no path
            "adding article to private cache"
          )
        )
        (else (error "bad body type"))
      )
      (error (format #f "evil key ~a ~a" key (check-key key)))
    )
  )
  )
))

(define make-get-article 
  (lambda (path)
    (with-docstring
      "retrieve the document with the key 'key' from the online help system"
      (lambda (key)
        (define val (assoc key help-articles))
        (if val
          (cadr val)
          (let ((fn (string-append path "/" (symbol->string key) ".wiki")))
            (if (access? fn R_OK)
              (let ((s (filename->string fn)))
                (if (and (>= (string-length s) 3) (equal? (substring s 0 3) utf8-bom))
                  (set! s (substring s 3))
                )
                s
              )
              #f
            )
          )
        )    
      )
    )
  )
)

(define make-put-article (lambda (path)
  (define make-wiki-article (make-make-wiki-article path))
  (with-docstring
    "put the contents of the article with the key 'key' into the online help system"
    (lambda (key body)
      (make-wiki-article key body)
    )
  )
))

(define search-article (with-docstring
  "accept an article, a pair ('name . \"body\") and run the compiled regular expression <re> against it and return true if there is a match against either name or body"
  (lambda (article re)
  (define helper (lambda (v)
    (cond
      ((string? v) (and (regexp-exec re v) #t))
      ((pair? v) (or (helper (car v)) (helper (cdr v))))
      (else #f)
    )
  ))
  (cond
    ((regexp-exec re (symbol->string (car article))) (car article))
    (else (helper (cdr article)))
  )
  )
))

(define make-proverb-help-index (lambda (path)
  (define get-fs-article-index (make-get-fs-article-index path))
  (with-docstring
    "return the list of article names available inside Proverb"
    (lambda ()
      (define help-files (get-fs-article-index))
      (append
        (map car help-articles)
        (delete #f (map (lambda (v)
          (define name (car v))
          (if (assoc name help-articles)
            #f
            name
          )
        ) help-files))
      )
    )
  )
))

(define help-index (with-docstring
  "return a list of help articles available in the enumerable namespaces. each element is a pair of the form ( \"namespace\" . 'article )"
  (lambda ()
    (apply append (map (lambda (v)
      (define i (assoc 'index (cdr v)))
      (if i
        (map (lambda (w) (cons (car v) w)) ((cdr i)))
        '()
      )
    ) wiki-namespaces))
  )
))

(define wiki-namespaces '())

(define register-wiki-namespace (lambda (name nsobj)
  (set! wiki-namespaces (reverse (cons (cons name nsobj) (reverse wiki-namespaces))))
))

(define search-help-articles (with-docstring
  "search all of the loaded documentation namespaces for query-string. each result is of the form (\"namespace\" . 'article)"
  (lambda (query-string)
    (apply append (map (lambda (v)
      (define s (assoc 'search (cdr v)))
      (if s
        (let ()
          (define matches ((cdr s) query-string))
          (map (lambda (w) (cons (car v) w)) matches)
        )
        '()
      )
    ) wiki-namespaces))
  )
))

(define wiki-put (with-docstring
  "(wiki-put namespace article-name article-body) -- article 'article-name' in wiki namespace 'namespace' gets the contents 'article-body'. If the wiki namespace 'namespace' does not already exist, this function returns false. If the update fails for some reason, this function returns false. If the update succeeds, this function returns non-false."
  (lambda (namespace article-name article-body)
    (define ns (assoc namespace wiki-namespaces))
    (if ns
      (let ((p (assoc 'put (cdr ns))))
        (if p
          ((cdr p) article-name article-body)
          #f
        )
      )
      #f
    )
  )
))

(define wiki-get (with-docstring
  "(wiki-get namespace article-name) -- returns the contents of the article 'article-name' in the wiki namespace 'namespace' or false if the article does not exist."
  (lambda (namespace article-name)
    (define ns (assoc namespace wiki-namespaces))
    (if ns
      (let ()
        (define g (assoc 'get (cdr ns)))
        (if g
          ((cdr g) article-name)
          #f
        )
      )
      #f
    )
  )
))

(define wiki-search (lambda (namespace query-string)
  (define ns (assoc namespace wiki-namespaces))
  (if ns
    (let ()
      (define q (assoc 'search (cdr ns)))
      (if q
        ((cdr q) query-string)
        #f
      )
    )
    #f
  )
))

(define wiki-index (with-docstring
  "(wiki-index namespace) -- return a list of the articles in wiki namespace 'namespace' or false if the namespace does not exist or cannot be enumerated."
  (lambda (namespace)
    (define ns (assoc namespace wiki-namespaces))
    (if ns
      (let ()
        (define i (assoc 'index (cdr ns)))
        (if i
          ((cdr i))
          #f
        )
      )
      #f
    )
  )
))

(define make-filesystem-wiki-ns (lambda (path)
  `(
    (get . ,(make-get-article path))
    (put . ,(make-put-article path))
    (search . ,(make-proverb-search-help-articles path))
    (index . ,(make-proverb-help-index path))
  )
))

(register-wiki-namespace "Proverb" (make-filesystem-wiki-ns (string-append support-dir "/wiki")))

)
