;
; routines for cacheing TIRE-passable forms in the Proverb HTTP server
;
(library (proverb http-cache)
  (export
    make-http-cache-fragment
    make-ephemeral-http-cache-fragment
    make-caching-xml-tag
    with-caching-xml-tag
    http-tire-cache
    http-tire-cache-add
    http-tire-cache-add-once
  )
  (import
    (guile)
    (rnrs)
    (rnrs lists)
    (proverb uri)
    (proverb passval)
    (proverb xml)
    (proverb doc)
    (proverb debug)
    (proverb http)
    (proverb html)
    (proverb dialect)
  )

(define-global http-cache-fragment (cons 'special 'http-cache-fragment))

(define make-http-cache-fragment (with-docstring
  "accept <fragment-string>,<mime-type> and make it available to the browser through a special cacheing namespace."
  (lambda (fragment-string mime-type)
    (list http-cache-fragment fragment-string mime-type #t)
  )
))

(define make-ephemeral-http-cache-fragment (with-docstring
  "like make-http-cache-fragment. fragment disappears from cache once read."
  (lambda (fragment-string mime-type)
    (list http-cache-fragment fragment-string mime-type #f)
  )
))

(define make-caching-xml-tag (with-docstring
  "construct a string representing an XML open tag for <tagname>. <rest> is a list of matched attribute names as symbols and Scheme objects as attribute values. Ex: (make-xml-tag 'img 'src \"foo.jpg\" 'width 64) yields <img src='foo.jpg' width='64'>. This form is can be passed directly to with-xml-tag."
  (lambda (tagname . rest)
  (string-append (apply string-append "<" (symbol->string tagname) " " (let helper ((cursor rest) (racc '()))
    (if (null? cursor)
      (reverse racc)
      (helper (cddr cursor) (cons (string-append
                                    (object->string (car cursor))
                                    "='"
                                    (xml-quote-escape-string (cond
                                      ((string? (cadr cursor)) (cadr cursor))
                                      ((and (pair? (cadr cursor)) (eq? (car (cadr cursor)) http-cache-fragment)) (let ()
                                        (define key (symbol->string (gensym "cache-")))
                                        (if (cadddr (cadr cursor))
                                          (http-tire-cache-add key (cadr (cadr cursor)) (caddr (cadr cursor)))
                                          (http-tire-cache-add-once key (cadr (cadr cursor)) (caddr (cadr cursor)))
                                        )
                                        (string-append "/cache?key=" (url-encode-string key))

                                      ))
                                      (else (object->string (cadr cursor)))))
                                    "' ") racc))
    )
  )) ">")
  )
))

(define with-caching-xml-tag (with-docstring
 "like with-xml-tag bug the tag expression is constructed with make-caching-xml-tag."
  (lambda (tag-expression . body)
    (cond
      ((string? tag-expression)
        (if (and
              (> (string-length tag-expression) 2)
              (equal? (string-ref tag-expression 0) #\<)
              (equal? (string-ref tag-expression (- (string-length tag-expression) 1)) #\>)
            )
          (let ((t (studio-string-split (substring tag-expression 1 (- (string-length tag-expression) 1)) #\space)))
            (make-raw (apply string-append (cons tag-expression (cons "" (append (map xmlize-value body) `("" "</" ,(car t) ">"))))))
          )
         (error "bogus tag")
        )
      )
      ((symbol? tag-expression)
        (apply with-xml-tag (make-caching-xml-tag tag-expression) body)
      )
      ((list? tag-expression)
        (apply with-xml-tag (apply make-caching-xml-tag tag-expression) body)
      )
      (else (error "bogus tag"))
    )
  )
))


(define http-tire-cache-store '())

(define http-tire-cache-add (lambda (key value mime)
  (set! http-tire-cache-store (cons (list key value mime #t) http-tire-cache-store))
))

(define http-tire-cache-add-once (lambda (key value mime)
  (set! http-tire-cache-store (cons (list key value mime #f) http-tire-cache-store))
))

(define http-tire-cache-get (lambda (key)
  (define r (assoc key http-tire-cache-store))
  (if r
    (begin
      (if (not (cadddr r))
        (set! http-tire-cache-store (remove r http-tire-cache-store))
      )
      (cdr r)
    )
    #f
  )
))

(define http-tire-cache (with-docstring
  "same calling convention as http-tire-rpc. provides access to cached objects"
  (lambda (tire-state http-state http-continuation inport outport path args)
    ; pass in a false http state if the http server runs in its own thread
    (catch #t (lambda ()
      (define key (cdr (safe-assoc "key" args '(("key" . #f)))))
      (if key
        (let ()
          (define v (http-tire-cache-get key))
          (if v
            (begin
              ;(critmsg "GOT THE KEY\n")
              ;(critmsg "L: ~a\n" (list? v))
              ;(critmsg "STILL HERE!\n")
              (http-string-responder (car v) (cadr v) outport)
            )
            (begin
              (critmsg "NO SUCH VALUE FOR KEY!\n")
            )
          )
        )
        (format outport "YOU SUPPLIED NO KEY\n")
      )
    ) (lambda err-args
      (critmsg "TIRE HTTP RPC ERROR: (pid=~a) ~a\n" (getpid) err-args)
      ;(backtrace)
      ;(and outport (display (xmlize (cons "READ ERROR" err-args)) outport))
    ))
    (http-continuation)
  )
))

)
