;
; Rotines for TIRE over HTTP
;
; TIRE documentation is available in tire.scm
(library (proverb http-rpc)
  (export http-tire-rpc)
  (import
    (rnrs)
    (guile)
    (proverb doc)
    (proverb debug)
    (proverb xml)
    (proverb dialect)
    (proverb tire)
    (proverb http)
    (proverb channels)
    (proverb threads)
  )

;
; TIRE can be transported over HTTP quite naturally without any need
; for a complicated translation layer. This code adapts
; a channel-oriented TIRE mux to the uncertain and unreliable style of
; TIRE with a browser.
;
; Browsers make no guarantees about the lifespan of connections, about how
; many open connections they will maintain, or about how HTTP requests will
; be scheduled across open connections.
;
; The key to TIRE over HTTP is somehow routing messages to a browser-based
; mux that is sometimes disconnected, sometimes multiply connected, and
; sometimes undetectably dead.
;
; Concepts:
;   session: a session a series of commands and queued responses that share
;            a common session identifier.
;
;   session identifier: a random string value chosen by the client

; each key is a session id
; each value is a pair (channel . inserter)
(define http-channel-hash (studio-make-hash-table string-hash string=? 128))
; each element is a list of pairs ( port . waiter )
; waiter is actually (http-port . http-continuation)
(define http-waiter-hash (studio-make-hash-table string-hash string=? 128))
(define hwh-mutex (make-mutex))

;
; return the list of http waiters
;
(define session-get-waiters (with-docstring
  "(session-get-waiters SESSION) -- return a list of waiters associated with an HTTP-RPC session. Browsers may keep multiple connections to Proverb's web server active at any time."
  (lambda (session)
    (or (with-mutex hwh-mutex (hash-ref http-waiter-hash session)) '())
  )
))

;
; add a new http waiter
(define http-session-add-waiter! (with-docstring
  "for the HTTP RPC session <session> add the waiter <waiter>. A waiter is an HTTP request from the browser that blocks until some arbitrary future data is available."
  (lambda (session waiter)
    (with-mutex hwh-mutex (hash-set! http-waiter-hash session (append (or (hash-ref http-waiter-hash session) '()) (list waiter))))
  )
))

;
; replace the existing waiters
;
(define session-replace-waiter! (with-docstring
  "for the HTTP RPC session <session> replace all current waiters with the single waiter <waiter>. Old waiters will be shut down."
  (lambda (session waiter)
  (define s (session-get-waiters session))
  (for-each (lambda (v)
    (http-waiter-shutdown session v)
      (session-remove-waiter! session v)
    ) s)
  (http-session-add-waiter! session waiter)
  )
))

(define session-remove-waiter! (with-docstring
  "(session-remove-waiter! session waiter) -- accept a session key session and a waiter c and remove them from the session structure. This call does not close or shutdown any ports."
  (lambda (session waiter)
  (debugmsg "session-remove-waiter!: before: ~a waiters\n" (length (session-get-waiters session)))
  (with-mutex hwh-mutex
    (hash-set! http-waiter-hash session
               (delete waiter (or (hash-ref http-waiter-hash session) '()))
    )
  )
  (debugmsg "session-remove-waiter!: after: ~a waiters\n" (length (session-get-waiters session)))
  )
))

(define http-waiter-shutdown (with-docstring
  "accepts a session key and a waiter. remove the waiter from the HTTP RPC accounting structures and finalize the underlying descriptors"
  (lambda (session waiter)
  ; accepts a session key and a waiter
  (http-waiter-shutdown-front session waiter)
  (or
    (port-closed? (car waiter)) 
    (begin
      (debugmsg "http-waiter-shutdown: SENDING THE REDUNDANT MESSAGE\n")

      (http-string-raw-responder 204 "No Content" '() "<?xml version=\"1.0\" encoding=\"UTF-8\"?><error>REDUNDANT WAITER</error>" "text/plain" (car waiter))
      (debugmsg "http-waiter-shutdown: BACK FROM SENDING THE REDUNDANT MESSAGE\n")
    )
  )
  (http-waiter-shutdown-back session waiter)
  )
))

(define http-waiter-shutdown-front (with-docstring
  "accepts a session key and a waiter. shutdown front updates the HTTP RPC accounting structures but does not touch the underlying descriptors"
  (lambda (session waiter)
    (debugmsg "HTTP-WAITER-SHUTDOWN-FRONT: SHUTTING DOWN WAITER session=~a p=~a &(car p)=~a\n" session waiter (object-address (car waiter)))
    (session-remove-waiter! session waiter)
  )
))

(define http-waiter-shutdown-back (with-docstring
  "accepts a session key and a waiter. shutdown back finalizes the underlying descriptors but does not touch the HTTP RPC accounting structures"
  (lambda (session waiter)
    (debugmsg "HTTP-WAITER-SHUTDOWN-BACK: SHUTTING DOWN WAITER session=~a p=~a &(car p)=~a\n" session waiter (object-address (car waiter)))
    ;(if (port-closed? (car waiter))
    ;  (critmsg "HTTP-SHUTDOWN: THE PORT IS ALREADY CLOSED: ~a\n" (object-address (car waiter)))
    ;  ; nonetheless, this can still fail occasionally. shutdown errors if the socket is not in a connected
    ;  ; state but isn't yet known to be closed. Mostly peer initiated shutdowns.
    ;  (catch #t (lambda () (shutdown (car waiter) 2)) always-true)
    ;)
    ; call the HTTP continuation
    ((cdr waiter))
  )
))


; use the channel model of channels.scm
;
; There's some chancey stuff here... We actually invoke the
; http-select-pre and select-post as interior forms to guarantee
; that we can actually accept new HTTP listeners when we're flushing
; the mux. Without this, dribbles will pile up.
;
; This hack would disappear if the 'eval' commands were non-blocking.
;
;
(define make-http-rpc-channel (with-docstring
  "make a new TIRE channel for the session <session> in the HTTP server state <http-state>. <http-state> is #f if the backing HTTP server is run in another thread."
  (lambda (session http-state http-headers)
    ; outgoing forms
    (define out-buffer '())
    (define buffer-mutex (if http-state #f (make-mutex)))
    (define this-channel-enqueue (lambda (mux value tire-continuation)
      (with-maybe-mutex buffer-mutex
        (set! out-buffer (append out-buffer (list (cons value tire-continuation))))
      )
      #t
    )) 
    (define this-channel-has-queue (lambda ()
      (with-maybe-mutex buffer-mutex (not (null? out-buffer)))
    ))
    (define this-channel-select-pre (lambda (mux)
      (define waiters (session-get-waiters session))
      (for-each (lambda (v)
        (if (port-closed? v)
          (critmsg "WHAT THE HECK? closed session ~a\n" v)
        )
      ) (map car waiters))
      ; if this channel has queued forms and the HTTP server runs in this
      ; thread then we may need to accept new HTTP connections.
      (if (this-channel-has-queue)
        (if http-state
          (begin
            (http-server-select-pre http-state '() (if waiters (map car waiters) '()) '())
          )
          (list '() (if waiters (map car waiters) '()) '())
        )
        (list '() '() '())
      )
      ;(if (and http-state (this-channel-has-queue))
      ;  (http-server-select-pre http-state '() (if waiters (map car waiters) '()) '())
      ;  (list '() (if waiters (map car waiters) '()) '())
      ;)
      ;(if (and waiters (not (null? waiters)) (this-channel-has-queue))
      ;  ; we only mark the descriptors for select
      ;  ; if we have an out-buffer
      ;  (begin
      ;    (debugmsg "a http-rpc-channel: SELECT-PRE: THERE ARE: ~a HTTP CONTINUATIONS AND ~a QUEUED FORMS\n" (length waiters) (with-maybe-mutex buffer-mutex (length out-buffer)))
      ;    (if http-state
      ;      (http-server-select-pre http-state '() (map car waiters) '())
      ;      (list '() (map car waiters) '())
      ;    )
      ;  )
      ;  (if http-state
      ;    (http-server-select-pre http-state '() '() '())
      ;    (list '() '() '())
      ;  )
      ;)
    ))
  
      (let* ()
        (define c (make-channel
          this-channel-enqueue    ; enqueue
          this-channel-has-queue  ; has queue
          this-channel-select-pre ; select pre
          ; select post
          (lambda (mux rfds wfds xfds)
            (debugmsg "a http-rpc-channel: SELECT-POST\n")
            (or
              (if (with-maybe-mutex buffer-mutex (not (null? out-buffer)))
              (let ()
                (define waiters (session-get-waiters session))
                (if (and waiters (not (null? waiters)))
                  (with-maybe-mutex buffer-mutex (let helper ((wcursor wfds) (hcursor waiters) (ocursor out-buffer) (wrote #f))
                    (if (or
                          (null? ocursor)
                          (null? hcursor)
                          (null? wcursor)
                        )
                      (begin
                        (set! out-buffer ocursor)
                        wrote
                      )
                      (if (member (car wcursor) (map car hcursor))
                        (let ((s (car hcursor)))
                          (http-waiter-shutdown-front session s)
                          (threadable-let ((o (car ocursor)))
                            (let ((str (xmlize (car o))))
                              (debugmsg "a http-rpc-channel -- THERE ARE: ~a CONTINUATIONS\n" (length waiters))
                              ;(http-string-responder str "text/xml" (car s))
                              (http-string-responder-with-headers str "text/xml" (or http-headers '()) (car s))
                              (http-waiter-shutdown-back session s)
                              ; call the channel enqueue continuation
                              (if (cdr o)
                                ((cdr o))
                              )
                            )
                          )
                          ; we only act on one descriptor per trip through the
                          ; loop
                          (helper '() (cdr hcursor) (cdr ocursor) #t)
                        )
                        (helper (cdr wcursor) hcursor ocursor #f)
                      )
                    )
                  ))
                  #f
                )
              )
              #f
            )
            (if http-state
              (http-server-select-post http-state rfds wfds xfds)
            )
          )
          #t
        )
        ; open
        #f
        ; close
        #f
        ; desc
        "HTTP RPC"
      ))
      c
    )
  )
))

(define httprpc-get-session-channel! (with-docstring
  "look up the channel in the TIRE mux <mux> for session ID <id>. If no channel exists, create a new one from HTTP server state <http-state>. pass a null http-state if the web server is running in another thread."
  (lambda (mux id http-state headers)
    (or (hash-ref http-channel-hash id)
        (let ()
          (define newc (make-http-rpc-channel id http-state headers))
          (hash-set! http-channel-hash id newc)
          (tire-mux-add-channel! mux newc)
          newc
        )
    )
  )
))

(define http-tire-rpc (with-docstring
  "accept a TIRE mux <tire-state>, an HTTP server state <http-state>, an HTTP request continuation <http-continuation> and HTTP request ports <inport>, <outport>, an HTTP request <path> and optional HTTP request arguments <args>. This call is intentionally similar to the calling convention of request handlers for the Proverb HTTP server. The HTTP server state may be false if the HTTP server runs in another thread"
  (lambda (tire-state http-state http-continuation inport outport path args headers)
    ; pass in a false http-state if the http server runs in its own thread
    (catch #t (lambda ()
      (lazy-catch #t (lambda ()
        (define r (call-with-input-string args (lambda (ip)
          (define rr (read ip))
          (close-input-port ip)
          rr
        )))
        ; r is not a tire form -- it is wrapped with some sessionizing
        (if (and (list? r) (not (null? r)))
          (cond
            ((and (eq? (car r) 'http-rpc) (= (length r) 3))
              ; do the RPC, return a TIRE form if one is available
              (let* ((session (list-ref r 1))
                     (form (list-ref r 2))
                     (c (httprpc-get-session-channel! tire-state session http-state headers))
                    )
                (session-replace-waiter! session (cons outport http-continuation))
                (if http-state 
                  (tire-mux-enqueue tire-state c form)
                  (tire-mux-buffer-enqueue tire-state c form)
                )
              )
            )
            ((and (eq? (car r) 'http-wait) (= (length r) 2))
              ; stick around waiting for a tire form
              (let* ((session (list-ref r 1))
                     (c (httprpc-get-session-channel! tire-state session http-state headers))
                    )
                (session-replace-waiter! session (cons outport http-continuation))
              )
            )
            (else (display "BOGUS RPC FORM!\n"))
          )
          "COMPLETELY BOGUS RPC FORM!\n"
        )
      ) (lambda err-args
        (critmsg "TIRE HTTP RPC ERROR: (pid=~a) ~a\n" (getpid) err-args)
        (backtrace)
        (and outport (display (xmlize (cons "READ ERROR" err-args)) outport))
        (http-continuation)
      )))
      ; the outer handler
      always-false
    )
  )
))

)
