;
; Whip up a scheme interaction environment for HTML/HTTP
;
(library (proverb http)
  (export
    current-http-server
    http-string-responder
    http-string-responder-with-headers
    http-string-raw-responder
    http-build-protected-directory-responder
    http-build-directory-responder
    http-build-redir
    http-build-symlink
    http-build-file-responder
    http-setup-server
    http-insert-handlers-from-template!
    http-protect
    http-write-protect
    http-write-redir
    http-server-select-pre
    http-server-select-post
    http-forward-port!
    ; introspection
    http-server-get-listener
    http-server-get-vfs
    http-server-get-url
    http-server-get-url-alist
    http-server-get-keepalives
    http-server-add-keepalive!
    http-server-remove-keepalive!
    ; request objects
    current-http-request
    http-request-get-method
    http-request-get-host
    http-request-get-auth
    http-request-get-origin
    ; statistics
    http-forwarded-ports
    http-requests
    ; authentication
    http-auth-simple
  )
  (import
    (rnrs)
    (rnrs lists)
    (ice-9 rw)
    (ice-9 rdelim)
    (guile modules)
    (proverb dialect)
    (proverb doc) 
    (proverb debug)
    (proverb guilefoo)
    (proverb html)
    (proverb md5)
    (proverb patterns)
    (proverb dataconv)
    (only (guile networking)
      inet-aton
      AF_INET
      socket
      SOCK_STREAM
      setsockopt
      getproto
      SO_REUSEADDR
      SOL_SOCKET
      bind
      listen
      connect
      INADDR_ANY
      INADDR_LOOPBACK
      gethostname
      getsockname
      gethostbyname
      gethostbyaddr
      hostent:addr-list
      hostent:name
      hostent:aliases
    )
    (only (guile)
      ; I/O functions
      port-closed?
      force-output
      peek-char
      the-eof-object
      setvbuf
      _IOFBF
      fdopen
      ; threading
      call-with-new-thread
      ; other functions
      string-index
      string-join
      list-head
      make-fluid
      fluid-ref
      fluid-set!
      object-property
      set-object-property!
      cond-expand
      eval-case export call-with-deferred-observers
      sort
      getenv
      gettimeofday
      *unspecified*
      random
      call-with-output-string
      defined? ; for ssl
      getpid ; for ssl
      fileno ; for ssl
      file-port? ; for ssl
      ; for debugging
      sleep
    )
  )

  (define ssl-available #f)
  (catch #t (lambda () (process-use-modules '(((openssl))))) (lambda args #f))
  (if (defined? 'SSL_CTX_new)
    (set! ssl-available #t)
  )

  (if ssl-available
    (begin
      (define ssl-master-context (SSL_CTX_new (SSLv23_method)))
      ; >128 bits, require authentication, specifically include RC4-SHA, sort by strength
      (SSL_CTX_set_cipher_list ssl-master-context "HIGH:!ADH:RC4-SHA:@STRENGTH")
      (define ssl-privatekey #f)
      (let ()
        (define pkey (EVP_PKEY_new))
        (define rsakey (RSA_generate_key 1024 #x10001))
        (EVP_PKEY_set1_RSA pkey rsakey)
        (SSL_CTX_use_PrivateKey ssl-master-context pkey)
        (set! ssl-privatekey pkey)
      )
      (define make-selfsigned-cert (lambda (pkey)
        "Accept a private key and create a new self-signed X509 certificate"
        (define cert (X509_new))
        (let ()
          (define name (X509_NAME_new))
          (X509_NAME_add_entry_by_NID name NID_commonName MBSTRING_ASC "localhost" -1 -1 0)
          (let ()
            (define serial (ASN1_INTEGER_new))
            (ASN1_INTEGER_set serial (getpid))
            (X509_set_version cert 2)
            (X509_set_serialNumber cert serial)
            (X509_set_issuer_name cert name)
            (X509_set_subject_name cert name)
            (X509_gmtime_adj (X509_get_notBefore cert) 0)
            (X509_gmtime_adj (X509_get_notAfter cert) (* 60 60 24 90))
            (X509_set_pubkey cert pkey)
            (X509_sign cert pkey (EVP_sha1))
            cert
          )
        )
      ))
      (let ()
        (define cert (make-selfsigned-cert ssl-privatekey))
        (SSL_CTX_use_certificate ssl-master-context cert)
      )
    )
  )

;(cond-expand
;  (guile
;    (use-modules (ice-9 rw)))
;    (use-modules (proverb dialect))
;  (chicken
;    (use tcp)
;  )
;  (else
;    #t
;  )
;)

(define peersafe-read-char (lambda (inp)
  (catch 'system-error (lambda ()
    (read-char inp)
  ) (lambda args
    the-eof-object
  ))
))

(define sleazy-global-http-request #f)
(define current-http-request (with-docstring
  "Return the current value for the HTTP request. This result is only true during the dynamic extent of an HTTP VFS handler."
  (lambda ()
    (fluid-ref sleazy-global-http-request)
  )
))

(define sleazy-global-http-server #f)
(define current-http-server (with-docstring
  "Return the current value for the HTTP server. This result is only true during the dynamic extent of an HTTP VFS handler."
  (lambda ()
    sleazy-global-http-server
  )
))

(define http-read-a-line (with-docstring
  "Read a character at a time from port p upto newline and return a list of chars. returns accumulated list on EOF."
  (cond-expand
    (guile
      (lambda (port)
        (define res (%read-line port))
        (string->list (if (string? (car res))
          (car res)
          (if (eof-object? (car res))
            ""
            (begin
              (critmsg "unknown value from %read-line")
              ""
            )
          )
        ))
      )
    )
    (else
      (lambda (port)
        (define helper (lambda (racc)
          (define char (peersafe-read-char port))
          (if (eof-object? char)
            (reverse racc)
            (if (eq? char #\newline)
              (reverse racc)
              (helper (cons char racc))
            )
          )
        ))
        (helper '())
      )
    )
  )
))


(define http-read-a-target (with-docstring
  "Read a character at a time from port p upto space and return a string"
  (lambda (port)
    (define res (catch #t
      (lambda () (read-delimited " " port))
      (lambda args (critmsg "http-read-a-target fails\n") #f)
    ))
    (if (and (string? res) (string=? res ""))
      (http-read-a-target port)
      res
    )
  )
))

(define decode-args (with-docstring
  "Decode URL-encoded HTTP args"
  (lambda (arg-string)
  (define split-args
    (if (string=? arg-string "")
      '()
      (studio-string-split arg-string #\&))
  )

  (map (lambda (v)
    (define split-kv (studio-string-split v #\=))
    (if (= (length split-kv) 1)
      (cons (car split-kv) *unspecified*)
      (cons (car split-kv) (cadr split-kv))
    )
  ) split-args)
  )
))

; returns a new namespace. handler may be a procedure or a fs
(define http-insert-handler! (with-docstring
  "accept an HTTP VFS <fs>, a new path list <path-list> and a VFS handler <handler> and return a new HTTP VFS that will handle <path-list>. although this function returns a new VFS object, the argument <fs> may be transformed."
  (lambda (path-list fs handler)
  (if (null? fs)
    ; the filesystem is completely empty
    (if (or (null? path-list) (equal? path-list '("/")))
      (if (list? handler)
        handler
        (list (cons "/" handler))
      )
      (list (cons (car path-list) (http-insert-handler! (cdr path-list) '() handler)))
    )
    ; else fs is not empty
    (if (or (null? path-list) (equal? path-list '("/")))
      ; this is a root handler 
      (if (list? handler)
        (append handler fs)
        (cons (cons "/" handler) fs)
      )
      ; this is not a root handler
      (let ()
        ; look up the first level of this path-list in the 
        ; existing VFS
        (define aval (assoc (car path-list) fs))
        (if aval
          ; The VFS already defines the first level.
          ; recurse. fs is transformed
          (begin
            (if (list? (cdr aval))
              (let ((new (http-insert-handler! (cdr path-list) (cdr aval) handler)))
                (set-cdr! aval new)
              )
              (begin 
                (if (null? (cdr path-list))
                  (if (list? handler)
                    (begin
                      (critmsg "WHACKING A FILE WITH A DIRECTORY: ~a\n" handler)
                    )
                    ; whacking a file with a file
                    (set-cdr! aval handler)
                  )
                  (critmsg "WITH AN IMPLIED DIRECTORY ~a\n" (cdr path-list))
                )
              )
            )
            fs
          )
          ; This is a new component
	  (cons (cons (car path-list) (http-insert-handler! (cdr path-list) '() handler)) fs) 
        ) 
      )
    )
  ) 
  )
))

(define http-string-raw-responder (with-docstring
 "send an HTTP response for string <str> with mime type mime-type on port"
  (lambda (code name headers str mime-type port)
 (define l (string-length str))
 (if (port-closed? port)
   (begin
     (debugmsg "http-string-raw-responder: responding to closed port!\n")
     #f
   )
   (let ()
     (define header-string (string-append
       "HTTP/1.1 " (number->string code 10) " " name "\n"
       (let helper ((cursor (reverse headers)) (sacc ""))
         (if (null? cursor)
           sacc
          (helper (cdr cursor) (string-append (car cursor) "\n" sacc))
         )
       )
       "Content-Type: " mime-type "\n"
       "Content-Length: " (number->string (string-length str) 10) "\n"
       "\n"
     ))
     (let ()
       (define c (write-string/full header-string port 0 (string-length header-string)))
       (or
         (equal? c (string-length header-string))
         (critmsg "incomplete header write\n")
       )
     )
     (let ()
       (define c (write-string/full str port 0 l))
       (or
         (equal? c l)
         (critmsg "incomplete write\n")
       )
     )
     (or
       (port-closed? port)
       (force-output port)
     )
   )
 )
  )
))

(define http-string-responder-with-headers (with-docstring
 "send an HTTP response for string s with mime type mime-type and additional HTTP headers headers on port"
  (lambda (str mime-type headers port)
    (http-string-raw-responder 200 "OK" headers str mime-type port)
  )
))

(define http-string-responder (with-docstring
 "send an HTTP response for string s with mime type mime-type and additional HTTP headers headers on port"
  (lambda (str mime-type port)
    (http-string-responder-with-headers str mime-type '() port)
  )
))

(define http-build-file-responder-with-headers (with-docstring
  "build a handler for an HTTP VFS that will deliver the contents of <filename> with MIME type <mime-type> and additional HTTP headers from <headers>. The contents of <filename> will be read anew from the filesystem on each request."
  (lambda (filename mime-type headers)
  (lambda (continuation inport outport path args)
    (define method (http-request-get-method (fluid-ref sleazy-global-http-request)))
    (case method
      ((GET) (let ()
    (define fp (open-input-file filename))
    (define uv (make-string 2048 #\0))
    ; inport is only char-ready? if its closed
    (if (and fp (not (port-closed? outport)))
      (let ((cl (file-size filename)))
        (display (string-append
          "HTTP/1.1 200 OK\n"
          "Content-Type: "
          (if mime-type
            mime-type
            "application/octet-stream"
          )
          "\n"
        ) outport)
        (for-each (lambda (v)
          (display (string-append v "\n"))
        ) headers)
        (display "Content-Length: " outport) (display cl outport) (display "\n" outport)
        (display "\n" outport)
       
        (cond-expand 
          (guile ;for some reason, chicken has a problem with this destructive read (segfaults)
            (let ()
              (define write-helper (lambda (s p start finish)
                (catch #t (lambda ()
                  (write-string/full s p start finish)
                ) (lambda args (critmsg "WRITE ERROR\n") #f))
              ))
            (let helper ((count (studio-read-string!/partial uv fp)) (total 0))
              (if (and count (not (= count 0)))
                (let ((write-count (write-helper uv outport 0 count) ))
                  (if (equal? write-count count)
                    (helper (studio-read-string!/partial uv fp) (+ total count))
                    (critmsg "WRITE HUH? ~a\n" count)
                  )
                )
                (if (equal? count 0)
                  (begin (critmsg "ZERO READ HUH? ~a\n" count) #f)
                  (if (equal? total cl)
                    #f 
                    (critmsg "WROTE LESS THAN CONTENT LENGTH!\n")
                  )
                )
              )
            )
            )
          )
          (chicken 
            (let helper ((str (read-string #f fp)))
              (if (> (string-length str) 0)
                (begin
	          (write-string str #f outport)
                  (helper (read-string #f fp))
                )
                'else-nobody-really-cares
              )
            )
          )
          (chez
            (let ()
              (define buffer (make-string 4096))
              (let helper ((count (block-read fp buffer)))
                (if (and (number? count) (> count 0))
                  (begin
                    (block-write outport buffer count)
                    (helper (block-read fp buffer))
                  )
                  'else-nobody-really-cares
                )
              )
            ) 
          )
          (else
            (display "BLOCK WRITE IN BUILD FILE RESPONDER!\n")
          )
        )
        ; we never checked to see if the write succeeded. We probably should!
        (close-input-port fp)
      )
      (critmsg "file-responder: some trouble with the ports! ~a\n")
    )
    (continuation)
  )
  )
  ((PUT) (let ()
    (define fp (open-output-file filename))
    (if fp
      (begin
        (display args fp)
        (close-output-port fp)
        (if (not (port-closed? outport))
          (display (string-append "HTTP/1.1 200 OK\nContent-Length 0:\n\n") outport)
        )
      )
      (critmsg "file-responder: couldn't open output file ~a\n" filename)
    )
    (continuation)
  ))
  (else
    (critmsg "BAD METHOD ~a FOR FILE RESPONDER\n" method)
    (continuation)
  )
  )
  )
  )
))

(define http-build-file-responder (with-docstring
  "build a handler for an HTTP VFS that will deliver the contents of <filename> with MIME type <mime-type>. The contents of <filename> will be read anew from the filesystem on each request."
  (lambda (filename mime-type)
    (http-build-file-responder-with-headers filename mime-type '())
  )
))

(define http-build-protected-directory-responder (with-docstring
  "build a handler for an HTTP VFS that will deliver the document tree <entries>. string entries will be composed with <prefix> to build a file responder, list entries yield subdirectories, and pair entries of (name . mimetype) will build file responders with mimetype. If protector is non-false, it will wrap every generated file responder."
  (lambda (prefix entries protector)
  (map (lambda (v)
    (cond
      ((and (pair? v) (string? (cdr v)))
        ; we have an associated mime type
        (cons (car v) (let ((r (http-build-file-responder (string-append prefix "/" (car v)) (cdr v)))) (if protector (protector r) r)))
      )
      ((list? v)
        (cons (car v) (http-build-protected-directory-responder (string-append prefix "/" (car v)) (cdr v) protector))
      )
      (else (cons v (let ((r (http-build-file-responder (string-append prefix "/" v) #f))) (if protector (protector r) r))))
    )
  ) entries)
  )
))

(define http-build-directory-responder (with-docstring
  "build a handler for an HTTP VFS that will deliver the document tree <entries>. string entries will be composed with <prefix> to build a file responder, list entries yield subdirectories, and pair entries of (name . mimetype) will build file responders with mimetype."
  (lambda (prefix entries)
    (http-build-protected-directory-responder prefix entries #f)
  )
))

(define http-build-redir (with-docstring
  "construct an HTTP VFS handler that redirects to the URL <url>"
  (lambda (url)
    (lambda (continuation inport outport path args)
      (display "HTTP/1.1 303 See Other \n" outport)
      (display "Location: " outport)
      (display url outport)
      (display "\n" outport)
      (display "Content-Length: 0\n" outport)
      (display "\n" outport)
      (continuation)
    )
  )
))

(define http-build-symlink (with-docstring
  "construct an HTTP VFS handler that redirects to <val>"
  (lambda (val)
    (lambda (continuation inport outport path args)
      ; the 303 response provides a location URI. We emit a relative URI
      ; so that we can get it right even when we are behind an stunnel
      ; proxy or a dominating web server
      (display (string-append
        "HTTP/1.1 303 See Other\n"
        "Location: ./"
        (if (and (string? val) (> (string-length val) 0) (eq? (string-ref val 0) #\/))
          (substring val 1)
          val
        )
        "\n"
        "Content-Length: 0\n"
        "\n"
      ) outport)
      (continuation)
    )
  )
))

; returns a proc that accepts args and that bundles in
; the hidden portion of the path fragment
(define http-manipulate-handlers (with-docstring
  "This is really a private routine for the HTTP server. Accept a path-list from an HTTP request, an HTTP VFS <fs> and return a new handler for this request. If the best VFS match for the path list is for a parent of the full path specified in path list, then the new handler will call the underlying handler with the balance of the outstanding path."
  (lambda (path-list fs)
    (if (null? path-list)
      (let ((aval (assoc "/" fs)))
        (if aval
          ; found a root handler
          (lambda (continuation inp outp args)
            ((cdr aval) continuation inp outp '() args)
          )
          #f
        )
      )
      (let ((aval (assoc (car path-list) fs)))
        (if aval
          (cond
            ((procedure? (cdr aval))
              (lambda (continuation inport outport args)
                ((cdr aval) continuation inport outport (cdr path-list) args)
              )
            )
            ((list? (cdr aval)) (let ((dir (cdr aval)))
              (if (null? dir)
                  #f
                  (http-manipulate-handlers (cdr path-list) dir)
              )
            ))
            (else #f)
          )
          #f
        )
      )
    )
  )
))

(define http-insert-handlers-from-template! (with-docstring
  "accept an HTTP VFS and additional filesystem template tree and return a new VFS. <vfs> may be transformed. The filesystem template is a tree of nodes each of the form (\"name\" body rest ). if body is procedure-valued, it is inserted as a handler. If string valued, it names a file responder with mime type (car rest). If body is list valued, it names a subdirectory."
  (lambda (vfs fs-template)
    (let helper ((cursor fs-template) (racc vfs))
      (if (null? cursor)
        racc
        (let ()
          (define name (let ()
            (define l (studio-string-split (caar cursor) #\/))
            (if (null? l)
              ("")
              (if (equal? "" (car l))
                (cons "/" (remove "" (cdr l)))
                (remove "" l)
              )
            )
          ))
          (helper (cdr cursor) (cond
            ; file or directory handled by procedure
            ((procedure? (cadar cursor))
              (http-insert-handler! name racc (cadar cursor))
            )
            ; file handled by the filesystem
            ((string? (cadar cursor))
              (http-insert-handler! name racc (http-build-file-responder
                (cadar cursor)
                (caddar cursor)
              ))
            )
            ; directory
            ((list? (cadar cursor))
              (http-insert-handler! name racc (cadar cursor))
            )
            (else
              (display "BOGO: ") (display (cadar cursor)) (newline)
              racc
            )
          ))
        )
      )
    )
  )
))


(define http-find-and-run (with-docstring
  "find a handler in the HTTP server <state> and VFS <fs> for the contents of <path-list> and run it with <args>. furnish the handler with the continuation <continuation> and in and outports <inport>,<outport>"
  (lambda (state continuation inport outport fs path-list args)
    (define handler (http-manipulate-handlers path-list fs))

    (if handler
      (dynamic-wind
        (lambda () (set! sleazy-global-http-server state))
        (lambda () (handler continuation inport outport args))
        (lambda () (set! sleazy-global-http-server #f))
      )
      ; else
      (begin
        (display (string-append
          "HTTP/1.1 404 Not Found \n"
          "Content-Length: 0\n"
          "\n"
        ) outport)
        (continuation)
      )
    )
  )
))

(define http-methods-dictionary #f)

(define http-logging #t)
(define http-requests '())

(define http-forwarded-ports '())

(define http-forward-port! (with-docstring
  "add TCP port <portno> to the list of ports available for forwarding through HTTP CONNECT"
  (lambda (portno)
    (set! http-forwarded-ports (cons portno (remove portno http-forwarded-ports)))
  )
))

;(define do-keepalive #t)
(define do-keepalive #f)

(define req-set-done  (lambda (req)   (vector-set! req 0 #t)))
(define req-set-start (lambda (req v) (vector-set! req 1 v)))
(define req-set-end   (lambda (req v) (vector-set! req 2 v)))
(define http-request-get-method (lambda (req) (vector-ref req 3)))
(define req-set-method (lambda (req v) (vector-set! req 3 v)))
(define req-set-path   (lambda (req v) (vector-set! req 4 v)))
(define http-request-get-host (lambda (req) (vector-ref req 5)))
(define req-set-host   (lambda (req v) (vector-set! req 5 v)))
(define http-request-get-auth (lambda (req) (vector-ref req 6)))
(define req-set-auth   (lambda (req v) (vector-set! req 6 v)))
(define req-set-origin   (lambda (req v) (vector-set! req 7 v)))
(define http-request-get-origin (lambda (req) (vector-ref req 7)))
(define req-set-header (lambda (req k v) (vector-set! req 8 (cons (cons k v) (vector-ref req 8)))))
(define req-get-header (lambda (req k) (assoc k (vector-ref req 8))))

(define make-http-request (lambda ()
  (vector #f #f #f #f #f #f #f #f '())
))

;
; accept an HTTP request from port <p>
;
(define http-get-request (with-docstring
  "accept an HTTP request from port <inp> and respond on port <outp> using the HTTP state <state>"
  (let ()
    (define tokenize-integer (lambda (l)
      (let helper ((cursor l) (racc '()))
        (if (null? cursor)
          (if (null? racc)
            #f
            (string->number (list->string (reverse racc)))
          )
          (if (member (car cursor) '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
            (helper (cdr cursor) (cons (car cursor) racc))
            (helper '() racc)
          )
        )
      )
    ))
    (lambda (inp outp state)
      (define start-time (gettimeofday))
      (define reqvec (make-http-request))
      (define fs (http-server-get-vfs state))
      (define requested-path #f)
      (define default-get-continuation (lambda ()
        (if (and do-keepalive
                 (catch #t
                   (lambda ()
                     ;(display "\n\n" outp)
                     (force-output outp)
                     #t
                   )
                   (lambda args
                     (debugmsg "default HTTP GET request continuation failed to flush HTTP out port\n")
                     #f
                   )
                 )
             )
          (begin
            (debugmsg "default HTTP GET request continuation reusing connection pair. closed for input: ~a closed for output: ~a\n" (port-closed? inp) (port-closed? outp))
            (http-server-add-keepalive! state (cons inp outp))
          )
          ; else
          (begin
            ; close should flush these ports
            (catch #t (lambda () (close-input-port inp)) always-true)
            (catch #t (lambda () (close-output-port outp)) always-true)
          )
        )
        (let ((end-time (gettimeofday)))
          (req-set-end reqvec end-time)
          (req-set-done reqvec)
          ;(vector-set! reqvec 1 (- (+ (* 1000000 (car end-time)) (cdr end-time)) (+ (* 1000000 (car start-time)) (cdr start-time))))
        )
      ))
      (define default-post-continuation (lambda ()
        ; close should flush these ports
        (catch #t (lambda () (close-input-port inp)) always-true)
        (catch #t (lambda () (close-output-port outp)) always-true)
      ))
      (define default-error-continuation (lambda ()
        ; close should flush these ports
        (catch #t (lambda () (close-input-port inp)) always-true)
        (catch #t (lambda () (close-output-port outp)) always-true)
        (http-server-remove-keepalive! state (cons inp outp))
      ))
      (define read-http-method (lambda (inp)
        (let helper ((racc '()))
          (define c (peersafe-read-char inp))
          (if (or (eof-object? c) (not (member c http-methods-dictionary)))
            (if (null? racc)
              (if (eof-object? c)
                (begin
                  (debugmsg "EOF WHILE READING METHOD. METHOD WAS ~a\n" (apply string (reverse racc)))
                  c
                )
                (begin
                  (critmsg "BAD CHARACTER ~a IN METHOD: ~a\n" c (apply string (reverse (cons c racc))))
                  #f
                )
              )
              (string->symbol (apply string (reverse racc)))
            )
            (helper (cons c racc)) 
          )
        )
      ))
      (req-set-start reqvec start-time)
      (set! http-requests (cons reqvec http-requests))
      (let gr-helper ((request-state 'INITIAL))
        (case request-state
          ((INITIAL) (let ()
            (define token (read-http-method inp))
            ;(critmsg "GET REQUEST: TOKEN IS: ~a\n" token)
            (if (eof-object? token)
              (begin
                (critmsg "EOF on HTTP method read\n") 
                (default-error-continuation)
              )
              (case token
                ((GET) (req-set-method reqvec 'GET) (gr-helper 'GET))
                ((POST) (req-set-method reqvec 'POST) (gr-helper 'POST))
                ((PUT) (req-set-method reqvec 'PUT) (gr-helper 'PUT))
                ((CONNECT) (req-set-method reqvec 'CONNECT) (gr-helper 'CONNECT))
                ((OPTIONS) (req-set-method reqvec 'OPTIONS) (gr-helper 'OPTIONS))
                (else
                  (critmsg "UNKNOWN HTTP METHOD ~a\n" token)
                  (default-error-continuation)
                )
              )   
            )
          ))
          ((GET) (let ()
            (define token (http-read-a-target inp))
            (set! requested-path token)
            (fluid-set! sleazy-global-http-request reqvec)
            (req-set-host reqvec "")
            (req-set-auth reqvec "")
            (req-set-path reqvec requested-path)
    	    (let absorber ()
              (define line-val (http-read-a-line inp))
              ;(critmsg "HTTP GET GOT HEADER: ~a\n" (list->string line-val))
              (cond
  	      ((bogo-string-match "Host: " line-val) => (lambda (match)
                  (req-set-host reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                ))
                ((bogo-string-match "Authorization: " line-val) => (lambda (match)
                  (req-set-auth reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                ))
                ((bogo-string-match "Origin: " line-val) => (lambda (match)
                  (req-set-origin reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                ))
                (else
                  #t 
                )
              )
  	      (if (> (length line-val) 2)
                (absorber)
	        ;(display "DONE ABSORBING!\n")
	      )
            )
            (gr-helper 'GET-RESPOND)
          ))
          ((POST) (let ()
            (define token (http-read-a-target inp))
	
            (define content-length 0)
            (define eat-headers (lambda ()
              (define line (http-read-a-line inp))
              (debugmsg "HTTP POST GOT HEADER: ~a\n" (list->string line))
              (cond
                ((bogo-pattern-match '("HTTP/1.1\r" EOF) line) => (lambda (match)
                  (eat-headers)
                ))
                ((bogo-string-match "Content-Length: " line) => (lambda (match)
                  (set! content-length (tokenize-integer match))
                  (if (not content-length)
                    (critmsg "INVALID CONTENT LENGTH: '~a'\n" match)  
                  )
                  (eat-headers)
                ))
                ((bogo-string-match "Authorization: " line) => (lambda (match)
                  (req-set-auth reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  (eat-headers)
                ))
                ((bogo-string-match "Origin: " line) => (lambda (match)
                  (req-set-origin reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  (eat-headers)
                ))
                ((bogo-string-match "Content-Type: " line) => (lambda (match)
                  (req-set-header reqvec 'content-type (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  (eat-headers)
                ))
                (else
                  ;(display "ABSORBED: ") (write (list->string line)) (newline)
                  (if (member #\: line)
                    (eat-headers)
                    (let ()
                      (define payload (let ()
                        (define payload-buffer (make-string content-length))
                        (studio-read-string!/full payload-buffer inp)
                        payload-buffer
                      ))
                      (http-find-and-run
                        state
                        (lambda () (debugmsg "POST using GET continuation\n") (default-get-continuation))
                        inp outp fs (remove "" (studio-string-split token #\/)) payload
                      )
                    )  
                  )
                )
              )
            ))
            (set! requested-path token)
            (req-set-host reqvec "")
            (req-set-auth reqvec "")
            (req-set-path reqvec requested-path)
            (fluid-set! sleazy-global-http-request reqvec)
            (eat-headers)
          ))
          ((PUT) (let ()
            (define token (http-read-a-target inp))
	
            (define content-length 0)
            (define eat-headers (lambda ()
              (define line (http-read-a-line inp))
              (cond
                ((bogo-pattern-match '("HTTP/1.1\r" EOF) line) => (lambda (match)
                  (eat-headers)
                ))
                ((bogo-string-match "Content-Length: " line) => (lambda (match)
                  (set! content-length (tokenize-integer match))
                  (if (not content-length)
                    (critmsg "INVALID CONTENT LENGTH: '~a'\n" match)  
                  )
                  (eat-headers)
                ))
                ((bogo-string-match "Authorization: " line) => (lambda (match)
                  (req-set-auth reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  (eat-headers)
                ))
                ((bogo-string-match "Origin: " line) => (lambda (match)
                  (req-set-origin reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  (eat-headers)
                ))
                (else
                  ;(display "ABSORBED: ") (write (list->string line)) (newline)
                  (if (member #\: line)
                    (eat-headers)
                    (let ()
                      (define payload (let ()
                        (define payload-buffer (make-string content-length))
                        (studio-read-string!/full payload-buffer inp)
                        payload-buffer
                      ))
                    
                      (http-find-and-run
                        state
                        (lambda () (debugmsg "PUT using GET continuation\n") (default-get-continuation))
                        inp outp fs (remove "" (studio-string-split token #\/)) payload
                      )
                    )  
                  )
                )
              )
            ))
            (set! requested-path token)
            (req-set-host reqvec "")
            (req-set-auth reqvec "")
            (req-set-path reqvec requested-path)
            (fluid-set! sleazy-global-http-request reqvec)
            (eat-headers)
          ))
          ((OPTIONS) (let ()
            (define token (http-read-a-target inp))
            (set! requested-path token)
            (fluid-set! sleazy-global-http-request reqvec)
            (req-set-host reqvec "")
            (req-set-auth reqvec "")
            (req-set-path reqvec requested-path)
    	    (let absorber ()
              (define line-val (http-read-a-line inp))
              (critmsg "HTTP OPTIONS GOT HEADER: ~a\n" (list->string line-val))
              (cond
  	      ((bogo-string-match "Host: " line-val) => (lambda (match)
                  (req-set-host reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                ))
                ((bogo-string-match "Authorization: " line-val) => (lambda (match)
                  (let ()
                    (req-set-auth reqvec (if (> (length match) 1) (apply string (list-head match (- (length match) 1))) ""))
                  ) 
                ))
                (else
                  #t 
                )
              )
  	      (if (> (length line-val) 2)
                (absorber)
                (gr-helper 'OPTIONS-RESPOND)
	      )
            )
          ))
          ;((CONNECT) (let ()
          ;  (define connect-port #f)
          ;  (define connect-host #f)
          ;  (let helper ()
          ;    (let ((c (remove #\return (http-read-a-line inp))))
          ;      (if (null? c)
          ;        (critmsg "DONE READING CONNECT HEADERS!\n")
          ;        (let ()
          ;          (define p1 (string-split (list->string c) #\space))
          ;          (if (> (length p1) 1)
          ;            (let ()
          ;              (define p2 (string-split (car p1) #\:))
          ;              (if (eq? (length p2) 2)
          ;                (begin
          ;                  (set! connect-host (car p2))
          ;                  (set! connect-port (string->number (cadr p2)))
          ;                )
          ;                'whocares
          ;              )
          ;            )
          ;            'whocares
          ;          )
          ;          (helper)
          ;        )
          ;      )
          ;    )
          ;  )
          ;  (if (and connect-port connect-host (member connect-port http-forwarded-ports))
          ;    (begin
          ;      (display "HTTP/1.0 200 OK\n\n" outp) (force-output outp)
          ;      (let ()
          ;        (define newp (socket AF_INET SOCK_STREAM 0))
          ;        (define buffer (make-string 10000))
          ;        (connect newp AF_INET (inet-aton "127.0.0.1") connect-port)
          ;        (let ()
          ;          (define worker-thunk (lambda args
          ;            (run-select-loop-until
          ;              (list 
          ;                (make-half-duplex-forwarding-service inp newp)
          ;                (make-half-duplex-forwarding-service newp outp)
          ;              )
          ;              (lambda () #t)
          ;            )
          ;            (critmsg "DONE WITH SELECT LOOP!\n")
          ;            (close-input-port inp)
          ;            (close-output-port outp)
          ;            (close-input-port newp)
          ;            (close-output-port newp)
          ;          ))
          ;          (call-with-new-thread worker-thunk)
          ;        )
          ;      )
          ;    )
          ;    (critmsg "HTTP CONNECT WITHOUT WELL FORMED PORT")
          ;  )
          ;))
          ((GET-RESPOND) (let* 
            (
              (argind (string-index requested-path #\?))
              (path-component (if argind (substring requested-path 0 argind) requested-path))
              (arg-component (if argind (substring requested-path (+ 1 argind)) ""))
            )
            (http-find-and-run state default-get-continuation inp outp fs (remove "" (studio-string-split path-component #\/)) (decode-args arg-component))
          
          ))
          ((OPTIONS-RESPOND) (let* ()
            ;(define message "HTTP/1.1 200 OK\nAccess-Control-Allow-Origin: https://www.cqctworld.org\nAccess-Control-Max-Age: 5000000\nAccess-Control-Allow-Headers: Content-Type\nAccess-Control-Allow-Methods: GET, POST, OPTIONS\nContent-Length: 0\n\n")
            (define message "HTTP/1.1 200 OK\nAccess-Control-Allow-Origin: *\nAccess-Control-Max-Age: 5000000\nAccess-Control-Allow-Headers: Content-Type\nAccess-Control-Allow-Methods: GET, POST, OPTIONS\nContent-Length: 0\n\n")
            (critmsg "FOO -- only Safari preflights these!\n")
            (display message outp)
            (critmsg message)
            (critmsg "BAR!\n")
            (default-get-continuation)
          ))
          (else
            (critmsg "HTTP SERVER BOGUS STATE!\n")
          )
        )
      )
    )
  )
))

(define http-run-server (with-docstring
  "run the HTTP server for the accepted connection specified by the pair of Scheme ports <iopair> and the HTTP server state <state>"
  (lambda (iopair state new?)
    ;NOTE: iopair is (in-port . out-port) --oabarne
    (cond-expand
      (guile
        (if (file-port? (car iopair)) (setvbuf (car iopair) _IOFBF))
        (if (file-port? (car iopair)) (setvbuf (cdr iopair) _IOFBF))
      )
      (else #t)
    )
    (http-get-request (car iopair) (cdr iopair) state)
    #t
  )
))

(define http-server-get-vfs (with-docstring
  "return the HTTP VFS object stored in the HTTP server state <state>"
  (lambda (state)
    (vector-ref state 1)
  )
))

(define http-server-set-vfs! (with-docstring
  "set the HTTP VFS object stored in the HTTP server state <state>"
  (lambda (state vfs)
    (vector-set! state 1 vfs)
  )
))

(define http-server-get-url (with-docstring
  "get the URL for the HTTP server state <state>"
  (lambda (state)
    (define u (vector-ref state 2))
    (let ()
      (define method (assoc 'method u))
      (define hostname (assoc 'hostname u))
      (define port (assoc 'port u))
      (if (or (not method) (not hostname) (not port))
        #f
        (string-append (cdr method) "://" (cdr hostname) ":" (number->string (cdr port)))
      ) 
    )
  )
))

(define http-server-get-url-alist (with-docstring
  "get the components of the URL for the HTTP server state <state> as an alist. Example: ((method . \"http\") (hostname . \"localhost\") (port . 3030))"
  (lambda (state)
    (vector-ref state 2)
  )
))

(define http-server-get-listener (with-docstring
  "return the listening socket stored in the HTTP server state <state>"
  (lambda (state)
    (vector-ref state 0)
  )
))

(define http-server-prune-keepalives! (with-docstring
  "trim dead connections from HTTP server state <state>"
  (lambda (state)
    (vector-set! state 3 (remove #f (map (lambda (v)
      (if (or (port-closed? (car v)) (port-closed? (cdr v)))
        (begin
          (debugmsg "ejecting a keepalive\n")
          #f
        )
        v
      )
    ) (vector-ref state 3))))
    #t
  )
))

(define http-server-get-keepalives (with-docstring
  "return the alive sockets HTTP server state <state>"
  (lambda (state)
    (http-server-prune-keepalives! state)
    (vector-ref state 3)
  )
))

(define http-server-add-keepalive! (with-docstring
  "add the socket pair <sockpair> to the HTTP server state <state>"
  (lambda (state sockpair)
    (http-server-prune-keepalives! state)
    (vector-set! state 3 (cons sockpair (remove sockpair (vector-ref state 3))))
  )
))

(define http-server-remove-keepalive! (with-docstring
  "remove the socket <sockpair> to the HTTP server state <state>"
  (lambda (state sockpair)
    (http-server-prune-keepalives! state)
    (vector-set! state 3 (remove sockpair (vector-ref state 3)))
  )
))

(define get-fqdn (with-docstring
  "attempt to return a fully qualified domain name for the current machine. this routine will attempt to return first a fqdn, then a resolvable hostname, then localhost. this process is more art than science."
  (lambda ()
    (let ((h (gethostname)))
      (if h
        (let ((hbn (catch #t
                          (lambda () (gethostbyname h))
                          (lambda args #f)))
             )
          (if (and hbn (hostent:addr-list hbn) (not (null? (hostent:addr-list hbn))))
            (let ((hbas (remove #f (map (lambda (v) (catch #t (lambda () (gethostbyaddr v)) (lambda args #f))) (hostent:addr-list hbn)))))
              (let ((r (map (lambda (v)
                              (define n (hostent:name v))
                              (define a (hostent:aliases v))
                              (let ((l (if n (cons n (or a '())) (or a '()))))
                                (if (null? l)
                                  h
                                  (car (sort l (lambda (a b)
                                    (define as (studio-string-split a #\.))
                                    (define bs (studio-string-split b #\.))
                                    (cond
                                      ; see if one is more fully qualified
                                      ((> (length as) (length bs)) #t)
                                      ((< (length as) (length bs)) #f)
                                      ; see if they are equally qualified
                                      ((equal? (cdr as) (cdr bs))
                                        ; they are. pick the shortest of the equally qualified names
                                        (< (string-length (car as)) (string-length (car bs)))
                                      )
                                      (else (> (string-length a) (string-length b)))
                                    )
                                  )))
                                )
                              )
                       ) hbas)))
                (if (null? r)
                  h
                  (car r)
                )
              )
            )
            "localhost"
          )
        )
        "localhost"
      )
    )
  )
))

(define http-setup-server (with-docstring
  "create an HTTP server state pre-populated with the VFS <vfs> and miscellaneous other arguments"
  (lambda (vfs args)
  (define http-hostname #f)
  (define setup-server-socket (lambda ()
    (define portbase 0)
    (cond-expand 
      (guile
        (let ()
        ;(define server-socket (socket AF_INET SOCK_STREAM 0))
        (define server-socket #f)
    
        (let ()
          (define p (assoc 'port args))
          (if p (set! portbase (cdr p)))
        )
        (let ()
          (define k (assoc 'keepalive args))
          (if k (set! do-keepalive (cdr k)))
        )
        (let ()
          (define h (assoc 'hostname args))
          (if h (set! http-hostname (cdr h)))
        )
        (if (not http-hostname)
          (set! http-hostname (get-fqdn))
        )
        (let ()
          (define s (assoc 'socketfd args))
          (if s
            (let ((port (fdopen (cdr s) "w+")))
              (if port
                (set! server-socket port)
                (critmsg "NO PORT")
              )
            )
            (begin
              (set! server-socket (socket AF_INET SOCK_STREAM 0))
              (setsockopt server-socket SOL_SOCKET SO_REUSEADDR 1)
              (let ((l (assoc 'local args)))
                (if (and l (not (cdr l)))
	          (bind server-socket AF_INET INADDR_LOOPBACK portbase)
	          (bind server-socket AF_INET INADDR_ANY portbase)
                )
              )
	      (listen server-socket 1)
            )
          ) 
        )

        server-socket
        )
      )
      (chicken 
        (let* ((p (assoc 'port args))
	       (http-listener (tcp-listen (if p (cdr p) portbase) 1)))
	  (define server-socket (tcp-listener-fileno http-listener))
	  (setprop! server-socket 'tcp-listener http-listener)
          
	  server-socket
        )
      )
      (else
        (display "HTTP-SETUP-SERVER: UNSURE HOW TO SET UP TCP ON THIS IMPLEMENTATION!\n")
      )
    )
  ))
  (let* ((server-socket (setup-server-socket))
         (url `(
           (method . ,(if ssl-available "https" "http"))
           (hostname . ,http-hostname)
           (port . ,(vector-ref (getsockname server-socket) 2))
         )))

    (let ((state (vector server-socket vfs url '())))
      state
    )
  )
  )
))

(define http-write-redir (with-docstring
  "create an HTML redirection in the host filesystem for the HTTP server <state>. if <target> is a string, the HTML redirection will be written to the file named by <target>, otherwise the redirection will be written in /tmp."
  (lambda (state target)
    (define url (http-server-get-url state))
    (define user (or (getenv "USER") "anon"))
    (let ()
      (define redir-filename (cond-expand
        (guile
          (if (string? target) target (string-append "/tmp/proverb-" user ".html"))
        )
        (else
          (if (string? target) target (string-append "/tmp/proverb-anon.html"))
        )
      ))
      (with-output-to-file redir-filename (lambda () (display (string-append
"<html>
  <head>
    <meta http-equiv='refresh' content='1;url=\""url"\"'></meta>
  </head>
  <body>
    <a href='"url"'>click here</a>
  </body>
</html>
"
      ))))
      redir-filename
    )
  )
))

(define http-server-select-pre (with-docstring
  "do all of the select pre work associated with the HTTP server <state>. Accepts accumulated select-pre descriptors from other select-pre/post servers and returns a new list with HTTP-specific descriptor lists."
  (lambda (state rfds wfds xfds)
    (if (port-closed? (http-server-get-listener state))
      (critmsg "the http server socket is closed?!?\n")
    )
    (list
      (cons (http-server-get-listener state) (append
        (let ((k (map car (http-server-get-keepalives state))))
          (debugmsg "KEEPALIVES: ~a\n" k)
          k 
        )
        rfds
      ))
      wfds
      xfds
    ) 
  )
))

(define http-visited? (with-docstring
  "a diagnostic function indented to test for select on stale fds"
  (lambda (rfds wfds xfds)
    (let helper ((cursor (list rfds wfds xfds)))
      (if (null? cursor)
        #f
        (if (null? (car cursor))
          (helper (cdr cursor))
          (if (object-property (car cursor) 'http-visited)
            #t
            (begin 
              (set-object-property! (car cursor) 'http-visited #t)
              (helper (cdr cursor))
            )
          )
        )
      )
    )
  )
))

(define http-server-select-post (with-docstring
  "do all of the select post work associated with the HTTP server <state>. Accepts the system-wide descriptor lists. returns <state>." 
  (lambda (state rfds wfds xfds)
    (if (not (http-visited? rfds wfds xfds))
      (if (member (http-server-get-listener state) rfds)
        (begin
          (set-object-property! rfds 'http-visited #t)
          (set! sleazy-global-http-server state)
          (if ssl-available
            (let ()
              (define foowrap (lambda (iopair)
                (define ssl (SSL_new ssl-master-context))
                (SSL_set_fd ssl (fileno (car iopair)))
                (SSL_accept ssl)
                (let ()
                  (define sslp (SSL->port ssl))
                  (cons sslp sslp)
                )
              ))
              (http-run-server (foowrap (accept-io-socket (http-server-get-listener state))) state #t)
            )
            (http-run-server (accept-io-socket (http-server-get-listener state)) state #t)
          )
          (set! sleazy-global-http-server #f)
        )
        (let keepalive-helper ((cursor (http-server-get-keepalives state)))
          ;(critmsg "CHECKING KEEPALIVES: ~a\n" cursor)
          (if (null? cursor)
            #f
            (if (member (caar cursor) rfds)
              (begin
                ; we will put it back into the keepalive pol if successful
                (http-server-remove-keepalive! state (car cursor))
                (if (not (port-closed? (caar cursor)))
                  (begin
                    (debugmsg "RUNNING A CONNECTION FROM A KEEPALIVE!\n")
                    (http-run-server (car cursor) state #f)
                  )
                  (begin
                    (critmsg "YIPES! KEEPALIVE DEAD!\n")
                  )
                )
              )
              (keepalive-helper (cdr cursor))
            )
          )
        )
      )
      (debugmsg "GOING AROUND THIS MULBERRY BUSH AGAIN")
    )
    state
  )
))

(define inner-http-protect (with-wiki-docstring
  "(http-protect handler auth) -- accepts a handler of the type expected by the Proverb http server and an authentication function. Returns a new handler of the same type that invokes the passed handler when auth returns non-false. Auth is called as (auth username password) where username and password are supplied with the browser. If the browser doesn't present credentials, auth is called as (auth #f #f)."
  (lambda (handler realm auth authrequired?)
    (define fail (lambda (inport outport continuation)
      (define nonce (number->string (random (expt 2 31)) 16))
      (http-string-raw-responder
        401
        "UNAUTHORIZED"
        (list
          (string-append "WWW-Authenticate: Digest realm=\"" (html-quote-escape-string realm) "\",nonce=\"" nonce "\",algorithm=MD5,qop=\"auth\"")
        )
        "DOOD! YOU HAVE TO LOG IN!"
        "text/plain"
        outport
      )
      (continuation)
    ))
    (define process-basic-auth (lambda (s)
      (if (string? s)
        (let ()
          (define i (string-index s #\space))
          (if i
            (let ()
              (define creds (base64->string (substring s (+ i 1))))
              (if (string? creds)
                (let ()
                  (define subi (string-index creds #\:))
                  (and subi (let ()
                    (define username (substring creds 0 subi))
                    (define password (substring creds (+ 1 subi)))
                    (auth username password)
                  ))
                )
                (auth #f #f)
              )
            )
            (auth #f #f)
          )
        )
        (auth #f #f)
      )
    ))
    (define process-digest-auth (lambda (s)
      (if (and (string? s) (>= (string-length s) (string-length "Digest ")))
        (let ()
          (define dictstr (substring s (string-length "Digest ")))
          (let ()
            (define dictionary (map (lambda (v)
              (define l (studio-string-split v #\=))
              (if (and (list? l) (> (length l) 1))
                ;(list (car l) (list->string (remove #\" (string->list (cadr l)))))
                (list (car l) (list->string (remove #\" (string->list (string-join (cdr l) "=")))))
                l 
              )
            ) (studio-string-split (list->string (remove #\space (string->list dictstr))) #\,)))
            (let ()
              (define safe-assoc (lambda (k l) (or (assoc k l) (list k ""))))
              (let ()
                (define nonce (cadr (safe-assoc "nonce" dictionary)))
                (define nonce-count (cadr (safe-assoc "nc" dictionary)))
                (define cnonce (cadr (safe-assoc "cnonce" dictionary)))
                (define username (cadr (safe-assoc "username" dictionary)))
                (define realm (cadr (safe-assoc "realm" dictionary)))
                (define qop (cadr (safe-assoc "qop" dictionary)))
                (define uri (cadr (safe-assoc "uri" dictionary)))
                (define response (cadr (safe-assoc "response" dictionary)))
                (define m (http-request-get-method (fluid-ref sleazy-global-http-request)))

                (and
                  nonce nonce-count cnonce username realm qop uri response
                  (let ()
                    (define authres (auth username #f))
                    (if (and authres (string? authres))
                      (let ()
                        (define ha1 authres) ; username:realm:pass
                        (or
                          (let ()
                            (define ha2 (md5 (string-append (symbol->string m) ":" uri)))
                            (string=?
                              (md5 (string-append ha1 ":" nonce ":" nonce-count ":" cnonce ":" qop ":" ha2))
                              response 
                            )
                          )
                          ;(let ()
                          ;  (critmsg "AUTH FAILURE!\n")
                          ;  #f
                          ;)
                        )
                      )
                      #f
                    )
                  )
                )
              )
            )
          )
        )
        (auth #f #f)
      )
    ))
    (lambda (continuation inport outport path args)
      (if (or (not authrequired?) (authrequired?))
        (if auth
          (if (process-digest-auth (http-request-get-auth (current-http-request)))
            (handler continuation inport outport path args)
            (fail inport outport continuation)
          )
          (fail inport outport continuation)
        )
        (handler continuation inport outport path args)
      )
    )
  )
))

(define http-protect (with-wiki-docstring
  "(http-protect handler auth) -- accepts a handler of the type expected by the Proverb http server and an authentication function. Returns a new handler of the same type that invokes the passed handler when auth returns non-false. Auth is called as (auth username password) where username and password are supplied with the browser. If the browser doesn't present credentials, auth is called as (auth #f #f)."
  (lambda (handler realm auth)
    (inner-http-protect handler realm auth #f)
  )
))

(define http-write-protect (with-wiki-docstring
  "(http-protect handler auth) -- accepts a handler of the type expected by the Proverb http server and an authentication function. Returns a new handler of the same type that invokes the passed handler when auth returns non-false. Auth is called as (auth username password) where username and password are supplied with the browser. If the browser doesn't present credentials, auth is called as (auth #f #f)."
  (lambda (handler realm auth)
    (inner-http-protect handler realm auth (lambda ()
      (define m (http-request-get-method (fluid-ref sleazy-global-http-request)))
      (if (eq? m 'GET)
        #f
        #t
      )
    ))
  )
))


(define http-auth-simple (with-docstring
  "a degenerate HTTP authentication function. (http-auth-simple username cred) returns true."
  (lambda (username cred)
    #t
  )
))

(cond-expand
  (sisc
    (define make-fluid (lambda () (cons 'fluid #f)))
    (define fluid-set! set-cdr!)
    (define fluid-ref cdr)
  )
  (else
    #t
  )
)

; non-define actions
(set! sleazy-global-http-request (make-fluid))
(fluid-set! sleazy-global-http-request #f)

(set! http-methods-dictionary (make-character-dictionary
  "GET"
  "POST"
  "CONNECT"

  "PROPFIND"  ;; WebDAV
  "PROPPATCH" ;; WebDAV
  "MKCOL"     ;; WebDAV
  "COPY"      ;; WebDAV
  "MOVE"      ;; WebDAV
  "LOCK"      ;; WebDAV
  "UNLOCK"    ;; WebDAV

  "OPTIONS"  ;; added to support XSS
))


)
