;
; Support routines for interactive completion.
;
; These routines are designed to be used as part of
; the read and eval error paths for interpreters to
; support interactive error recovery.
;
; In particular, we want to allow users to interactively complete
; dangling lists, function calls, and quoted strings

(library (proverb interactive-completion)
  (export
    partial-read diagnose-eof fix-string-read-errors
    find-completions-in-module find-completions-in-module-list
    find-completions-from-module find-largest-root
    find-largest-common-completion tab-complete-string
    is-complete-form?
  )
  (import
    (rnrs)
    (proverb types)
    (proverb doc) (proverb dialect) 
    (proverb lexical) (proverb debug)
    (proverb guilefoo)
    (guile modules)
    (guile errors)
    (only (guile)
      struct-ref
      list-set! set-procedure-property!
      *unspecified*
      macro? delete close
      open-input-string
      object->string
      module-variable
      variable-bound?
      variable-ref
      call-with-input-string
      struct? %record-type-error
      fluid-set!
    )
  )

(define partial-read (with-docstring
  "read the first Scheme form from a string"
  (lambda (str)
    (define sp (open-input-string str))
    (debugmsg "opened an input string!\n")
    (let ()
      (define result (read sp))
      (close sp)
      result
    )
  )
))

(define diagnose-eof
  (let ()
    (define string-backtracker (with-docstring
      "find the dangling quote in a string with an incomplete Scheme form"
      (lambda (lst racc quote-depth)
        (if (null? lst)
          ; we couldn't find the problem
          (begin
            (display "BIG PROBLEM! COULDN'T FIND ERROR\n")
            *unspecified*
          )
          ; else
          (if (eq? (car lst) #\")
            ; we found a double quote
            (if (eq? quote-depth 0)
    	      ; it's the dangling quote
              (let ()
                (if (or (null? (cdr lst)) (not (eq? (cadr lst) #\\)))
                  (let ()
                    (define pr (partial-read (list->string racc)))
	            (list 'open-string pr (length lst))
                  )
                  ; else
	          (string-backtracker (cdr lst) (cons (car lst) racc) quote-depth)
                )
              )
              ; otherwise we've already seen this mate
	      (string-backtracker (cdr lst) (cons (car lst) racc) (- quote-depth 1))
            )
            (if (eq? (car lst) #\) )
              ; we've got a close quote
              (string-backtracker (cdr lst) (cons (car lst) racc) (+ quote-depth 1))
              ; ust a normal char
	      (string-backtracker (cdr lst) (cons (car lst) racc) quote-depth)
            )
          )
        )
      )
    ))

    (define list-backtracker (with-docstring
      "return one of *unspecified* `(dangling-call <name of call (symbol)> <preceeding material (string)> <args (list)> <trailing whitespace? (boolean)>))"
      (lambda (str)
        (let helper ((l (reverse (string->list str))) (racc '()) (d 0) (i 0))
          (if (null? l)
            ; we couldn't find the problem
            (begin
              (display "BIG PROBLEM! COULDN'T FIND ERROR\n")
              *unspecified*
            )
            ; else
            (if (eq? (car l) #\( )
              ; we found an open paren
              (if (eq? d 0)
      	        ; it's the dangling paren
                (let ()
                  (define pr (partial-read (list->string racc)))
                  ; pr2 is the read value of the args that would be passed
                  (define pr2 (let ((res (partial-read
                                           (string-append "(" (string-tail str (- (string-length str) i)) ")")
                                         )))
                    (if (null? res)
                      #f
                      (cdr res)
                    ) 
                  ))
                  ; the last arg is true if the last char is a space
	          (list 'dangling-call
                        pr
                        (list->string (reverse (cdr l)))
                        pr2
                        (eq? (string-ref str (- (string-length str) 1)) #\space)
                  )
                )
                ; otherwise we've already seen this mate
	        (helper (cdr l) (cons (car l) racc) (- d 1) (+ i 1))
              )
              (if (eq? (car l) #\) )
                ; we've got a close quote
                (helper (cdr l) (cons (car l) racc) (+ d 1) (+ i 1))
                ; just a normal char
                (helper (cdr l) (cons (car l) racc) d (+ i 1))
              )
            )
          )
        )
      )
    ))

    (lambda (string errorlist)
      "Accept a string <string> and list of errors <errorlist> as passed to a catch handler and return a structure for recovering from the error. return unspecified if we can't figure it out."
      ;
      ; main diagnose routine
      (if (eq? (length errorlist) 5)
        (let ()
          (cond
            ; a guile 1.6ism
            ((eq? (list-ref errorlist 0) 'misc-error) (let ()
              (cond
                ((string=? (list-ref errorlist 1) "list") (let ()
                  (define res (list-backtracker string))
                  res
                ))
  	      ((string=? (list-ref errorlist 2) "end of file in string constant")
  	        (string-backtracker (reverse (string->list string)) '() 0)
                )
                (else *unspecified*)
              )
            ))
            ; a guile 1.8ism
            ((eq? (list-ref errorlist 0) 'read-error) (let ()
              (define smesg "end of file in string constant")
              (define eofmsg "end of file")
              (cond
                ((string=? (list-ref errorlist 1) "list") (let ()
                  (define res (list-backtracker string))
                  res
                ))
  	        ((string=? (list-ref errorlist 1) "scm_i_lreadparen") (let ()
                  (define res (list-backtracker string))
                  res
                ))
	        ((and (string=? (list-ref errorlist 1) "scm_lreadr") (>= (string-length (list-ref errorlist 2)) (string-length smesg)) (string=? (substring (list-ref errorlist 2) (- (string-length (list-ref errorlist 2)) (string-length smesg))) smesg))
	          (string-backtracker (reverse (string->list string)) '() 0)
                )
                (else *unspecified*)
              )
            ))
            (else *unspecified*)
          )
        )
        ; else
        *unspecified*
      )
    )
  )
)

(define fix-string-read-errors (with-docstring
  "silently correct the read errors in <string>. returns #f if unfixed, otherwise returns the repaired string"
  (lambda (string args)

  ; why couldn't this string be read?
  (define newres (diagnose-eof string args))

  (if (pair? newres)
    (case (car newres)
      ((dangling-call) (let ()
	(string-append string ")")
      ))
      ((open-string) (let ()
	(string-append string "\"")
      ))
      (else #f)
    )
    ; else newres wasn't a pair -- which means that diagnose failed
    #f
  )
  )
))

(define find-completions-in-module (with-docstring
  "return a list of possible completions in module <module>"
  (lambda-w-types ((string string?) (module module?))
  (define symbols (module-get-symbol-list module))
  (let ()
    (define possible-matches (delete #f (map (lambda (v)
      (define symstring (symbol->string (car v)))
      (define comp-length (string-length string))
      (if (>= (string-length symstring) comp-length)
        (if (string=? string (substring symstring 0 comp-length))
          symstring
	  #f
        )
	#f
      )
    ) symbols)))

    possible-matches
  )
  )
))

(define find-completions-in-module-list (with-docstring
  "return a list of possible completions in the list of modules <modules>"
  (lambda-w-types ((string string?) (modules (lambda (v) (and (list? v) (all? (map module? v))))))
  (define helper (lambda (l racc)
    (if (null? l)
      (reverse racc)
      (helper (cdr l) (cons (find-completions-in-module string (car l)) racc))
    )
  ))
  (apply append (helper modules '()))
  )
))

(define find-completions-from-module (with-docstring
  "return a list of possible completions from module and the modules used by module"
  (lambda (string module)
  (find-completions-in-module-list string (cons module (module-uses module)))
  )
))

(define find-largest-root (with-docstring
  "find the largest shared root string between string1 and string2"
  (lambda-w-types ((string1 string?) (string2 string?))
  (define comp-len (let () 
    (define len1 (string-length string1))
    (define len2 (string-length string2))
    (if (> len1 len2)
      len2
      len1
    )
  ))

  (define helper (lambda (c)
    (if (eq? c 0)
      ""
      (if (string=? (substring string1 0 c) (substring string2 0 c))
        (substring string1 0 c)
        (helper (- c 1))
      )
    )
  ))

  (helper comp-len)
  )
))

(define find-largest-common-completion (with-docstring
  "accept a list of possible completions as strings and return the longest possible completion."
  (lambda-w-types ((completions (lambda (v) (and (list? v) (all? (map string? v))))))
  (define helper (lambda (root l)
    (if (null? l)
      root
      (helper (find-largest-root root (car l)) (cdr l))
    )
  ))

  (if (null? completions)
    #f
    (helper (car completions) (cdr completions))
  )
  )
))

(define tab-complete-string (with-docstring
  "return the tab-completed version of <str>"
  (lambda (str module)
    ; try to read the input string as a scheme form.
    ; in most cases, we can't
    (define args (catch #t
      (lambda ()
        (call-with-input-string str read)
        #f
      )
      (lambda args
        args
      )
    ))
    (let ((diagnosis (and args (diagnose-eof str args))))
      (if (pair? diagnosis)
        (case (car diagnosis)
          ((dangling-call)
            (if (symbol? (cadr diagnosis))
              (let ()
                (define sym (list-ref diagnosis 1))
                (define prefix (list-ref diagnosis 2))
                (define args (list-ref diagnosis 3))
                (define whitespace (list-ref diagnosis 4))
                (define complete-arg-name (lambda (index)
                  (define sv (let ((v (module-variable module sym))) (and v (variable-bound? v) (variable-ref v))))
                  (if (and sv (procedure? sv))
                    (let ()
                      ;(define c (consider-procedure sv (current-module)))
                      (define c (consider-procedure sv))
                      (if c
                        (let ()
                          (define n (procedure-get-arg-names c))
                          (if (and (list? n) (> (length n) index))
                            (list-ref n index)
                            (string-append "arg-" (number->string index))
                          )
                        )
                        (string-append "arg-" (number->string index))
                      )
                    )
                    (if (and sv (macro? sv))
                      (string-append "macro-arg-" (number->string index))
                      (string-append "arg-" (number->string index))
                    )
                  )
                ))
                ; what are we completing
                (cond
                  ((and (null? args) (not whitespace)) (let ()
                    ; there are no args and no trailing whitespace 
                    (define possible-matches (find-completions-from-module (symbol->string sym) module))
                    (if (null? possible-matches)
                      str ; return the original string
                      ; else
                      (string-append prefix "(" (find-largest-common-completion possible-matches))
                    )
                  ))
                  ((null? args) (let ()
                    ; there are no args but there is trailing whitespace
                    (string-append str (complete-arg-name 0))
                  ))
                  ((not whitespace) (let ((ra (reverse args)))
                    ; there are args and no trailing whitespace 
                    (define possible-matches (if (symbol? (car ra))
                      (find-completions-from-module (symbol->string (car ra)) module) 
                      '()
                    ))
                    (if (null? possible-matches)
                      str ; return the original string
                      ; else
                      (string-append prefix "(" (symbol->string sym) " " (let helper ((cursor args) (racc '()))
                        (if (null? (cdr cursor)) 
                          (apply string-append (reverse racc))
                          (helper (cdr cursor) (cons (string-append (object->string (car cursor)) " ") racc))
                        )
                      ) (find-largest-common-completion possible-matches))
                    )
                  ))
                  (whitespace (let ()
                    (string-append str (complete-arg-name (length args)))
                  ))
                  (else
                    ; I have no idea
                    str ; the original string
                  )
                )
              )
              ; else this is some type of dangling call that I don't understand
              str
            )
          )
          ((open-string)
            (string-append str "\"")
          )
          (else
            (object->string diagnosis)
          )
        )
        (let ()
          (define possible-matches (find-completions-from-module str module))
          (if (null? possible-matches)
            str ; return the original string
            (find-largest-common-completion possible-matches)
          )
        )
      )
    )
  )
))

(define is-complete-form? (with-docstring
  "return #t if a form can be read from <str> without error, #f otherwise"
  (lambda (str)
    (catch #t (lambda () (call-with-input-string str read) #t) (lambda args
      (diagnose-eof str args)
      #f
    ))
  )
))

)
