(library (proverb itree)
  (export
    make-itree
    make-itree-node
    itree-node-start
    itree-node-end
    itree-node-value
    itree-node-left
    itree-node-right
    itree-set-node-start!
    itree-set-node-end!
    itree-set-node-left!
    itree-set-node-right!
    itree-compare-nodes
    itree-insert-interval
    itree-search-point
    itree-prev
    itree-next
  )
  (import (rnrs) (proverb doc))

(define make-itree (with-docstring
  "create a new, empty interval tree"
  (lambda ()
    (cons 'itree '())
  )
))

(define make-itree-node (with-docstring
  "create a new interval node for interval trees"
  (lambda (istart iend value)
    (vector istart iend value #f #f)
  )
))

(define itree-node-start (with-docstring
  "return the start of the interval represented by node"
  (lambda (node)
    (vector-ref node 0)
  )
))

(define itree-node-end (with-docstring
  "return the end of the interval represented by node"
  (lambda (node)
    (vector-ref node 1)
  )
))

(define itree-node-value (with-docstring
  "return the value of the interval tree node"
  (lambda (node)
    (vector-ref node 2)
  )
))

(define itree-node-left (with-docstring
  "return the left branch of the itree rooted at node"
  (lambda (node)
    (vector-ref node 3)
  )
))

(define itree-node-right (with-docstring
  "return the right branch of the itree rooted at node"
  (lambda (node)
    (vector-ref node 4)
  )
))

(define itree-set-node-start! (with-docstring
  "set the end of the interval represented by node to value"
  (lambda (node value)
    (vector-set! node 0 value)
  )
))

(define itree-set-node-end! (with-docstring
  "set the end of the interval represented by node to value"
  (lambda (node value)
    (vector-set! node 1 value)
  )
))

(define itree-set-node-left! (with-docstring
  "set the left branch of the interval represented by node to value"
  (lambda (node value)
    (vector-set! node 3 value)
  )
))

(define itree-set-node-right! (with-docstring
  "set the right branch of the interval represented by node to value"
  (lambda (node value)
    (vector-set! node 4 value)
  )
))

;
; these trees are currently limited to non-negative integers and
; an open interval
;

(define itree-compare-nodes (with-wiki-docstring
"
accepts two interval tree nodes and returns on of the symbols '<','>','='

* < : node-a is less than node-b
* > : node-a is greater than node-b
* = : node-a intersects node-b

works for interval trees with open, non-negative, integer intervals
"
  (lambda (node-a node-b)
  (define astart (itree-node-start node-a))
  (define aend (itree-node-end node-b))
  (define bstart (itree-node-start node-b))
  (define bend (itree-node-end node-b))
  (cond
    ((and aend (> bstart aend))
      '<
    )
    ((and bend (> astart bend))
      '>	
    )
    (else
      '=
    )
  )
  )
))

(define itree-insert-interval (with-docstring
  "insert the interval bracketed by istart, iend into the interval tree itree with value value"
  (lambda (tree istart iend value)
  (if (null? (cdr tree))
    ; this is an empty tree
    (set-cdr! tree (make-itree-node istart iend value))
    (let helper ((node (cdr tree)))
      (cond
        ((and (itree-node-end node) (> istart (itree-node-end node))) (let ()
          (define r (itree-node-right node))
          (if r
            (helper r)
            (itree-set-node-right! node (make-itree-node istart iend value))
          )
        ))
        ((and iend (< iend (itree-node-start node))) (let ()
          (define l (itree-node-left node))
          (if l
            (helper l)
            (itree-set-node-left! node (make-itree-node istart iend value))
          )
        ))
        ;
        ; else these intervals overlap
        ;

        ; these can be consolidated
        ((eq? value (itree-node-value node))
          (and
            (itree-node-end node)
            (or (not iend) (> iend (itree-node-end node)))
            (itree-set-node-end! node iend)
          )
          (and
            (< istart (itree-node-start node))
            (itree-set-node-start! node istart)
          )
        )
        (else
          ; these overlap but have different values
          ; i have no idea
          #f
        )
      )
    )
  )
  )
))

(define itree-search-point (with-docstring
  "search the interval tree tree for the point point"
  (lambda (tree point)
  (if (null? (cdr tree))  
    #t
    (let helper ((node (cdr tree)))
      (cond
        ((< point (itree-node-start node)) (let ()
          (define l (itree-node-left node))
          (if l
            (helper l)  
	    #t
          )
        ))
        ; point overlaps an open node
        ((not (itree-node-end node))
          (itree-node-value node)
        )
        ((> point (itree-node-end node)) (let ()
          (define r (itree-node-right node))
          (if r
            (helper r)
            #t
          )
        ))
        (else (itree-node-value node))
      )
    )
  )
  )
))

(define itree-prev (with-docstring
  "return the value of the node whose end most closely preceeds value"
  (lambda (tree value)
    (define lcase (lambda (node carry)
      (define l (itree-node-left node))
      (if l
        (rcase l carry)
        (if carry
          (itree-node-value carry)
          (if (eq? (itree-node-start node) 0)
            #f
            #t
          )
        )
      )
    ))
    (define rcase (lambda (node carry)
      (cond
        ; this node isn't preceeding
        ((< value (itree-node-start node))
          (lcase node carry)
        )
        ; value overlaps an open node
        ((not (itree-node-end node))
          (lcase node carry)
        )
        ; this is a preceeding node
        ((> value (itree-node-end node)) (let ()
          (define r (itree-node-right node))
          (if r
            (rcase r node)
            (itree-node-value node)
          )
        ))
        ; value overlaps a closed node
        (else
          (lcase node carry)
        )
      )
    ))
    (rcase (cdr tree) #f)
  )
))

(define itree-next (with-docstring
  "return the value of the node whose end most closely succeeds value"
  (lambda (tree value)
  (let helper ((node (cdr tree)) (carry #t))
    (cond
      ; this node isn't preceeding
      ((< value (itree-node-start node)) (let ()
        (define l (itree-node-left node))
        (if l
          (helper l node)
          (itree-node-value node)
        )
      ))
      ; value overlaps an open node
      ((not (itree-node-end node))
        (itree-node-value node)
      )
      ; this is a preceeding node
      ((> value (itree-node-end node)) (let ()
        (define r (itree-node-right node))
        (if r
          (helper r carry)
          (or (eq? carry #t) (itree-node-value carry))
        )
      ))
      ; value overlaps a closed node
      (else
        (itree-node-value node)
      )
    )
  )
  )
))

)
