;
; routines for dealing with MIME
;
(library (proverb mime)
  (export
    filename->mime-type
    audio-guess-mime-type
    image-guess-mime-type
    data-guess-mime-type
  )
  (import
    (rnrs)
    (proverb doc)
    (only (proverb dialect) studio-string-split)
  )

(define filename->mime-type (with-docstring
  "accept a filename and return a possible mime type based on the extension"
  (lambda (filename)
    (define s (studio-string-split filename #\.))
    (let ()
      (define sl (length s))
      (if (< sl 2)
        "text/octet-stream" 
        (let* ((ext (list-ref s (- sl 1)))
               (dict '(
                 ; text types
                 ("html"    . "text/html")
                 ("xhtml"   . "application/xhtml+xml")
                 ("js"      . "text/javascript")
                 ("manifest" . "text/cache-manifest")
                 ("xml"     . "text/xml")
                 ("css"     . "text/css")
                 ; image types
                 ("jpg"     . "image/jpeg")
                 ("png"     . "image/png")
                 ("gif"     . "image/gif")
                 ; other types
                 ("pdf"     . "application/pdf")
                 ("svg"     . "image/svg+xml")
                 ("torrent" . "application/x-bittorrent")
                 ("swf"     . "application/x-shockwave-flash")
                 ("xul"     . "application/vnd.mozilla.xul+xml")
                 ("mov"     . "video/quicktime")
                 ("ogg"     . "audio/ogg")
               ))
               (aval (assoc ext dict)))

          (if aval
            (cdr aval)
            "text/octet-stream"
          )
        )
      )
    )
  )
))

(define quicktime-sig (list->string (map integer->char (list
  #x00 #x00 #x00 #x20 #x66 #x74 #x79 #x70 #x71 #x74 #x20 #x20
))))

(define jpeg-sig (list->string (map integer->char (list
  #xff #xd8
))))

(define image-guess-mime-type (with-docstring
  "accept a string representing a binary image and return a possible mime type based on the contents"
  (lambda (str)
    (cond
      ((and (>= (string-length str) 4) (string=? (substring str 1 4) "PNG"))
        "image/png"
      )
      ((and (>= (string-length str) 3) (string=? (substring str 0 3) "GIF"))
        "image/gif"
      )
      ((and (>= (string-length str) 2) (string=? (substring str 0 2) jpeg-sig))
        "image/jpeg"
      )
      ((and (>= (string-length str) 5) (string=? (substring str 0 5) "<?xml"))
        "image/svg+xml"
      )
      ((and (>= (string-length str) (string-length quicktime-sig)) (string=? (substring str 0 (string-length quicktime-sig)) quicktime-sig))
        "video/quicktime"
      )
      (else #f)
    )
  )
))

(define audio-guess-mime-type (with-docstring
  "accept a string representing binary audio and return a possible mime type based on the contents"
  (lambda (str)
    (cond
      ((and (>= (string-length str) 3) (string=? (substring str 1 3) "ID3"))
        "audio/mpeg"
      )
      ((and (>= (string-length str) 2) (string=? (substring str 0 2) (string #\377 #\373)))
        "audio/mpeg"
      )
      ((and (>= (string-length str) 4) (string=? (substring str 0 4) (string #\M #\T #\h #\d)))
        "audio/midi"
      )
      (else #f)
    )
  )
))

(define data-guess-mime-type (with-docstring
  "attempt to guess a mime type for <str>. falls through to data/octet-stream when n
o type can be guessed."
  (lambda (str)
    (or
      (image-guess-mime-type str)
      (audio-guess-mime-type str)
      (cond
        ((and (>= (string-length str) 15) (string=? (substring str 1 3) "CACHE MANIFEST\n"))
          "text/cache-manifest"
        )
        (else "data/octet-stream")
      )
    )
  ) 
))

)
