;
; Code for managing inferior Schemes and forked children
;

(library (proverb remote)
  (export
    fork-add multi-fork-add popen-add
    remote-eval
    fork-eval
    fork-eval-w-callback
    fork-delay*
    fork-begin multi-fork-begin
    fork-delete!
    cookie-create-proverb-channel generate-local-cookie dribble
    current-tire-mux
    tire-eval-yield
    bogo-resume
    sleazy-tire-responder
    emergency-tire-responder
    popen-fports
    url-add
    loopback-add
    ; rarer forms
    tire-make-connect-back-url
    tire-open-back-func
    tire-open-back-system
    tire-slave-connect-back
    my-rank my-size
    snatch-responder
    snatch-embeddable-responder
    fork-add-helper*
    tire-eval
    tire-eval-general
  )
  (import 
    (rnrs)
    (rnrs lists)
    (rnrs io simple)
    (proverb debug) (proverb tire) (proverb doc)
    (proverb types)
    (proverb unread) (proverb dialect) (proverb concurrent)
    (proverb threads)
    (proverb wiki)
    (ice-9 syncase)
    (proverb cond-expand)
    (proverb guilefoo)
    (hashtable)
    (guile modules)
    (guile errors)
    (only (guile networking) getsockname gethostname)
    (only (guile filesystem) access? R_OK) ; only used to define proc-based getpid. Should move from remote.
    (only (guile) 
      port-closed?
      with-error-to-port
      with-output-to-port
      make-soft-port
      set-object-property! list-set! set-procedure-property!
      getpid gensym define-macro eval-case
      defmacro:transformer readlink
      format call-with-input-string set-current-error-port
      set-current-output-port
      fluid-set! debug-options
      force-output assert-defmacro?! set-defmacro-transformer! debug-options-interface *unspecified*
      kill SIGKILL
      system dup2 port->fdes
      ; hash routines
      hash-set! hashq-set! hash hash-ref
      ; threads
      system-async-mark
      call-with-new-thread
      join-thread
    )
  )

;global bindings for the mux environment
(define sleazy-root-mux #f)
(define sleazy-replacement-escape #f)
(define sleazy-tag #f)
(define current-tag (with-docstring
  "sleazy function to return the tag associated with the current command in a TIRE evaluation environment. Utterly vile."
  (lambda () sleazy-tag)
))
(define current-tire-mux (with-docstring
  "sleazy function to return the TIRE mux associated with the current TIRE evaluation environment. Less vile than current-tag."
  (lambda ()
    sleazy-root-mux
  )
))
(define sleazy-tire-responder #f)
;; there is a sleazy-tire-responder help article
(define emergency-tire-responder #f)
;; this doesn't have a default value. You can hook it if
;; you want to allow the dribble port to have a TIRE I/O
;; path of last resort

(define sleazy-continuation-snatched #f)
(define snatch-continuation (with-wiki-docstring "
snatch-continuation is called with no arguments and
delivers a function that must be called with the 
response to the [[tire]] command then in effect.

This function is not a continuation-valued object
and will return immediately. The evaluation will
proceed normally and the return value will be
ignored.

The waiting tire form will block until the returned function is passed a value.

The returned function may be called once.

This function is a wrapper around [[[(help snatch-responder)|snatch-responder]]].
"
  (lambda ()
    (define c (snatch-responder))
    (lambda (v)
      (c 'r v)
    )
  )
))
(define snatch-responder (with-wiki-docstring "

snatch-responder is called with no arguments and
delivers a function that must be called with the 
response to the [[tire]] command then in effect.

This function is not a continuation-valued object
and will return immediately. The evaluation will
proceed normally and the return value will be
ignored.

The waiting tire form will block until this function is passed a value.

The function may be called once.
"
  (lambda ()
    (define snatch-called #f)
    (define packaged-continuation sleazy-tire-responder)
    (set! sleazy-continuation-snatched #t)
    (lambda (c . v)
      (if snatch-called
        (error "calling a dead continuation!")
        (begin 
          (set! snatch-called #f)
          (apply packaged-continuation c v)
        )
      )
    )
  )
))

(define snatch-embeddable-responder (with-wiki-docstring "

snatch-embeddable-responder is a form of [[[(help snatch-responder)]]]
that is safe to call from threads other than the main mux
thread.

"
  (lambda ()
    (define snatch-called #f)
    (let* ((tag (current-tag))
           (mux (current-tire-mux))
           (from (tire-request-get-src (tire-mux-get-newest-request-for-tag mux tag)))
           (packaged-continuation (lambda (rcommand . rrest)
             (tire-mux-buffer-enqueue-sync mux from (cons rcommand (cons tag rrest))))))

      (set! sleazy-continuation-snatched #t)
      (lambda (c . v)
        (if snatch-called
          (error "calling a dead continuation!")
          (begin
            (set! snatch-called #f)
            (apply packaged-continuation c v)
          )
        )
      )
    )
  )
))

(define dribble (with-wiki-docstring "


dribble is called with a value and an optional token and 
issues an i message for the currently evaluation tire 
expression.  The i message will have the symbol 'dribble 
in the position after the tag, followed by obj.  If token 
is specified, 'dribble is replaced with the value of 
token.

If dribble called outside the extent of a tire expression, 
obj will be displayed to the port returned by 
current-error-port.  To properly issue i messages outside 
the extent of a tire expression, see [[snatch-responder]].
"
  (lambda (obj . token)
    (if sleazy-tire-responder
      (sleazy-tire-responder 'i
                             (if (null? token)
                               'dribble
                               (car token))
                             obj)
      (write obj (debug-port))
    )
  )
))

(define make-dribble-port (with-docstring
  "return a new port object that will turn writes into TIRE i actions with the command <tag>"
  (lambda (tag)
    (define buffer '())
    (define flusher (lambda (b)
      (define s (apply string-append (reverse b)))
      (set! buffer '())
      (if sleazy-tire-responder
        (dribble s tag)
        (if emergency-tire-responder
          (emergency-tire-responder 'i 'async s)
          (display s (debug-port))
        )
      )
    ))
    (define saver (lambda (b)
      (set! buffer b)
    ))
    (define flush-method (lambda ()
      (or (null? buffer)
          (flusher buffer)
      )
    ))
(cond-expand
(guile
  (let ()
    (define p (make-soft-port
      (vector 
        ; character output
        (lambda (c)
          ((if (eq? c #\newline)
            flusher
            saver
          ) (cons (string c) buffer))
        )
        ; string output
        (lambda (str) 
          ((if (member #\newline (string->list str))
            flusher
            saver
          ) (cons str buffer)) 
        )
        ; flush output
        flush-method
        ; read a char
        #f
        ; close a port
        (lambda ()
          (critmsg "CLOSING A DRIBBLE PORT?!?")
        )
        ; buffer len
        #f
      )
    "w"
    ))
    (set-object-property! p 'extra-flush flush-method)
    p
  )
)
(sisc
  (let ()
    (define p (make-custom-character-output-port
      ; character output
      (lambda (port char)
          ((if (equal? char #\newline)
            flusher
            saver
          ) (cons (string char) buffer))
      )
      ; string output
      (lambda (port string offset count)
          ((if (member #\newline (string->list string))
            flusher
            saver
          ) (cons str buffer)) 
      )
      ; flush
      (lambda (port)
        (flush-method)
      )
      ; close
      (lambda (port)
        #t
      )
    ))
    p
  )
)
(else (display "MAKE DRIBBLE PORT NOT SUPPORTED!\n"))
)
)
))

;(define dribble-output-port (make-dribble-port 'output))
;(define dribble-error-port (make-dribble-port 'error))
(define dribble-output-port #f)
(define dribble-error-port #f)

(define dskr-yield-alist '())

(define tire-eval-yield (with-wiki-docstring
"
tire-eval-yield is called with a
key and blocks evaluation of the enclosing 
expression (without blocking the mux) until a [[tire-eval-resume]] 
call is made.  If key is not specified, the value of ([[current-tag]]) is used.

When a corresponding [[tire-eval-resume]] is called from another
tire expression or context, the tire-eval-yield call will return
with the argument passed into tire-eval-resume
"
  (lambda (key)
  ; this call must only be executed inside a TIRE eval command
  (if (not key)
    (tire-eval-yield (current-tag))
    (let ()
      (define this-continuation-snatched sleazy-continuation-snatched)
      ;(set! sleazy-continuation-snatched #t)
      (let ()
        (define res (call-with-current-continuation (lambda (cc)
           (set! dskr-yield-alist (cons (list key cc) dskr-yield-alist))
           (sleazy-escape 'bogo)
        )))
        ; res should be a pair ( replacement-continuation . return-value)
        (set! sleazy-continuation-snatched this-continuation-snatched)
        ;(critmsg "PAST THE SLEAZY ESCAPE! RES=~a\n" res)
        (set! sleazy-replacement-escape (car res))
        (cdr res)
      )
    )
  )
  ) 
))

;"
;tire-eval-resume is called to resume execution of a form 
;interrupted by [[tire-eval-yield]].  tire-eval-resume is called
;with the tag of the yielded form and a variable number of arguments.
;The procedure passed to tire-eval-yield will be executed these args
;in the same [[Wikipedia:Variable#Scope_and_extent | dynamic extent]]  
;that tire-eval-yield was called.  The number of arguments passed to
;tire-eval-resume must be valid for the resumed procedure.
;
;"
(define bogo-resume (lambda (key arg)
  ; who to resume
  (define aval (assoc key dskr-yield-alist))
  ; they can be resumed no more
  (set! dskr-yield-alist (remove aval dskr-yield-alist))
  ;(set! sleazy-continuation-snatched #t)
  ((cadr aval) (cons sleazy-escape arg))
))

(define displaying-i-handler 
  (with-wiki-docstring "
    The displays '''output''' and '''error''' i-messages received in 
 dependent TIRE forms.  Other tire forms are discarded.
    See also: 
* [[tire-eval]]
* [[rewriting-i-handler]]
  "
  (lambda (form)
    (case (caddr form)
      ((output)
        (display (cadddr form)))
      ((error)
        (display (cadddr form) (current-error-port)))))))

(define rewriting-i-handler 
  (with-wiki-docstring "
    The default handler for i-messages received in dependent TIRE forms.

    See also: 
* [[tire-eval]]
* [[displaying-i-handler]]
  "
  (lambda (form)
    ;re-write tag
    (let ((p (tire-mux-get-oldest-request-for-tag (current-tire-mux) (cadr form))))
      (apply sleazy-tire-responder 'i (cddr form) )))))

;"
;
;tire-eval is called with a tire form and an optional i-handler procedure
;of one argument and enqueues the tire form into the current TIRE mux --
;as returned by (current-tire-mux).  
;This form blocks evaluation of the enclosing expression until a response 
;is received, but does not block the mux, so the tire form can be 
;addressed to the local mux or a remote mux. If i-handler is defined, it 
;will be called (with the complete i message) each time an i messages is 
;received, otherwise i messages are passed to [[displaying-i-handler]]. 
;i-handler is called in the same 
;[[Wikipedia:Variable#Scope_and_extent | dynamic extent]] that tire-eval
;was called in.  
;
;tire-eval returns the value of the tire response (the [[caddr]] of the r 
;message).
;
;"
(define tire-eval (lambda-w-types ((form list?) . (i-handler procedure?))
  (define mux (current-tire-mux))
  (define outer-responder sleazy-tire-responder)
  (define resume-key (gensym "resume-key-"))
  ;(tire-mux-buffer-enqueue
  (tire-mux-enqueue
    mux
    (make-tire-procedure-channel (lambda (response-form)
      (if (list? response-form)
        (cond
          ((and (>= (length response-form) 3) (eq? (car response-form) 'r))
            ;(critmsg "TIRE EVAL COMPLETED!\n")
            (bogo-resume resume-key (list-ref response-form 2))
          )
          ((and (>= (length response-form) 3) (eq? (car response-form) 'i))
            (if (null? i-handler)
              (begin
                (apply outer-responder 'i (cddr response-form))
              )
              ((car i-handler) response-form)
            )
          )
          ((and (>= (length response-form) 2) (eq? (car response-form) 'e))
            (critmsg "TIRE-EVAL: RESPONSE IS AN E!\n")
          )
          (else
            (critmsg "TIRE-EVAL: UNKNOWN TYPE OF RESPONSE FORM!\n")
          ) 
        )
        (critmsg "TIRE-EVAL: RESPONSE FORM IS NOT A LIST!\n")
      )
    ))
    form
  )
  ;(critmsg "TIRE-EVAL BACK FROM BUFFER ENQUEUE!\n")
  ;(critmsg "YIELDING WITH KEY ~a\n" resume-key)
  (let ()
    (define retval (tire-eval-yield resume-key))
    ;(critmsg "BACK FROM YIELD!\n")
    retval
  )
))


(define tire-eval-general (with-docstring
"tire-eval-general provides several features for eval'ing a form on
a remote host. These include (1) the ability to block until the remote
host has returned the evaluation, (2) the ability to check and see if the
remote host has returned the evaluation, and (3) the ability to register
a callback that will be called when the remote host has finished the
evaluation. To call tire-eval-general:
(tire-eval-general handle code proc)
Where:
handle is a handle for a remote mux, code is a list of the code to be eval'ed, 
and proc is a function that will be called with the results as arguments 
when results of eval'ing code are returned to this mux.  
It returns a pair (blocking-ftn, checking-ftn) where running 
(blocking-ftn) will block until the remote mux returns the results, and
return the results, and (checking-ftn) will tell if the remote mux has
returned the results yet (#t if so, #f if not)."
  (lambda-w-types ((handle string?) (code list?) (callback procedure?))
  (define mux (current-tire-mux))
  (define returned #f) ; has the forked process finished?
  (define waiting #f)  ; is this process blocking on the forked process?
  (define result #f)   ; the result of evaluating form on fork proc (note, could be #f)
  (define outer-responder sleazy-tire-responder)
  (define resume-key (gensym "forking-key"))
  (tire-mux-enqueue
    mux
    (make-tire-procedure-channel (lambda (response-form)
      (if (list? response-form)
        (cond
          ((and (>= (length response-form) 3) (eq? (car response-form) 'r))
	   (begin
	     (set! returned #t)
	     (set! result (list-ref response-form 2))
	     (if waiting (bogo-resume resume-key (list-ref response-form 2)) #t)
	     (callback result))
          )
          ((and (>= (length response-form) 3) (eq? (car response-form) 'i))
	   (apply outer-responder 'i (cddr response-form))
          )
          ((and (>= (length response-form) 2) (eq? (car response-form) 'e))
            (critmsg "TIRE-EVAL: RESPONSE IS AN E!\n")
          )
          (else
            (critmsg "TIRE-EVAL: UNKNOWN TYPE OF RESPONSE FORM!\n")
          ) 
        )
        (critmsg "TIRE-EVAL: RESPONSE FORM IS NOT A LIST!\n")
      )
    ))
    `(t ,(gensym "tire-eval-") ,handle eval ,code) 
  )
  (cons
    (lambda () 
      (if returned result
	(begin
	  (set! waiting #t)
	  (tire-eval-yield resume-key))))
    (lambda () returned))
)))

(define wiki-error-message "ERROR. type or click [[[(backtrace)]]] for a trace. You could also [[[(xhtml-show-stack (fluid-ref the-last-stack))|try]]] the newfangled stack widget.")

(define tire-applier* (with-docstring
  "tire-applier is a low-level function that implements the guts of TIRE eval and force commands. This form sets up several of the sleazy globals available in TIRE eval environments."
  (lambda (mux tag cookie responder thunk)
    ;save environment
    (define saved-continuation-snatched sleazy-continuation-snatched)
    (define saved-sleazy-escape sleazy-escape)
    (define result #f)
    (define xeval-stack #f)
    (cond-expand
      (guile
        (define xeval-handler (lambda args
          (define stack (make-stack #t xeval-handler thunk 0))
          (set! xeval-stack stack)
          (fluid-set! the-last-stack stack)
          (dribble (make-wiki wiki-error-message))
          (if (member 'backtrace (debug-options))
            (backtrace)
            (dribble (make-wiki "type or click [[[(debug-enable 'backtrace)]]] to automatically print backtraces in the future\n"))
          )
          (display (apply format #f (list-ref args 2) (list-ref args 3)) 
                   (current-error-port))
          (newline (current-error-port))
        ))
      )
      (chicken 
        (define xeval-handler (lambda args
          (display (format #f "ERROR. ~a\n" args) (current-error-port))
          #t
        ))
      )
      (else
        (define xeval-handler (lambda args
          #t
        ))
      )
    )
    ;(set! sleazy-continuation-snatched #f)
    (set! sleazy-replacement-escape #f)

    (call-with-current-continuation (lambda (cc)

      (set! sleazy-escape (with-wiki-docstring "a thoroughly disgusting function documented [[sleazy-escape | here]]" cc))

      (let ()
        (define last-result 'no-last-result)
        (define saved-output-port #f)
        (define saved-error-port #f)
        (define saved-sleazy-tag #f)
        (define saved-mux #f)
        (define saved-sleazy-continuation #f)
        ;(define saved-sleazy-escape #f)
        (define warpin-count 0)
        (define warpout-count 0)
        (dynamic-wind
          (lambda ()
            (set! warpin-count (+ warpin-count 1))
            ;(if (> warpin-count 1)
            ;   (critmsg "WARPIN MULTIPLE: COUNT ~a ~a ~a\n" warpin-count last-result result)
            ;)
            (set! saved-error-port (set-current-error-port dribble-error-port))
            (set! saved-output-port (set-current-output-port dribble-output-port))
            (set! saved-sleazy-tag sleazy-tag) (set! sleazy-tag tag)
            (set! saved-mux sleazy-root-mux) (set! sleazy-root-mux mux)
            ; there is an article about sleazy-tire-responder
            (set! saved-sleazy-continuation sleazy-tire-responder) (set! sleazy-tire-responder responder)
            (set! sleazy-continuation-snatched #f)
          )
          (lambda ()
            (set! result (catch #t (lambda ()
              (lazy-catch #t
                thunk
  	        xeval-handler
              )
            ) (lambda args (debugmsg "UNHANDLED EVAL ERROR: ~a\n" args) #f)))
          )
          (lambda ()
            (set! warpout-count (+ warpout-count 1))
            (set-current-output-port saved-output-port)
            (set-current-error-port saved-error-port)
            ;(if (> warpout-count 1)
            ;  (critmsg "WARPOUT MULTIPLE: COUNT ~a ~a ~a\n" warpout-count last-result result)
            ;)
            (set! last-result result)
            (set! sleazy-tag saved-sleazy-tag)
            (set! sleazy-root-mux saved-mux)
            (set! sleazy-tire-responder saved-sleazy-continuation)
	    (studio-flush-port dribble-error-port)
	    (studio-flush-port dribble-output-port)
            ;(set! sleazy-escape saved-sleazy-escape)
          )
        )
      )

      (if (not sleazy-continuation-snatched)
        (begin
          ;(critmsg "CALLING RESPONDER!\n")
          (responder 'r result)
        )
        (begin
          (debugmsg "NOT USING THE RESPONDER FOR TAG ~a\n" tag)
        )
      )

      (let ()
        (if sleazy-replacement-escape
          (let ()
            (define s sleazy-replacement-escape)
            (set! sleazy-replacement-escape #f)
            ;(critmsg "SOMEONE HAS REPLACED SLEAZY ESCAPE!\n")
            (s 'whoa-dude)
          )
        )
        (set! sleazy-continuation-snatched saved-continuation-snatched)
        (set! sleazy-escape saved-sleazy-escape)
        (set! sleazy-replacement-escape #f)
      )
    ))
  )
))

(define make-tire-evaluator (let ()
  (define tire-evaluator* (lambda (mux tag cookie responder body module)
    (define xeval-runner (lambda ()
      (cond-expand
        (guile
          ; macro expansion blows up on some biggish forms
          (eval (list "noexpand" body) (or module (current-module)))
        )
        (chicken
          (eval body (or module (current-module)))
        )
        (else
          (critmsg "XEVAL RUNNER IS NON-PORTABLE AND NOT IMPLEMENTED HERE!\n")
        )
      )
    ))
    (tire-applier* mux tag cookie responder xeval-runner)
  ))

  (with-docstring
    "make a TIRE mux cookie command handler that will use the Scheme evaluator to evaluate forms in the environment <module>"
    (lambda (module)
      (lambda (mux tag cookie responder body)
        (tire-evaluator* mux tag cookie responder body module)
      )
    )
  )
))

(define coroute (with-docstring
  "route cookie1 on the same channel as cookie2. This call depends on (current-tire-mux)."
  (lambda (cookie1 cookie2)
    (define m (current-tire-mux))
    (tire-mux-hash-cookie! m cookie1 (tire-mux-lookup-cookie m cookie2))
    *unspecified*
  )
)) 

;
; return the pair cookie, channel
;
(define cookie-magic (cons 'cookie 'cookie))
(define normal-exit-token (cons 'normal-exit 'normal-exit))

(define cookie-create-channel (with-wiki-docstring
  "create a new cookie, channel pair in the [[tire | TIRE]] mux 'mux'. The created cookie is anonymous."
  (lambda (mux)
  ; our cookie structure
  (define l (list
    cookie-magic ; our sentinel
    '()          ; our commands
  ))
  (cons
    ; our structure
    l
    ; our channel
    (let () (define c (make-tire-procedure-channel (lambda (form)
      (case (car form)
        ((t) (let ()
          (define cmd (assoc (list-ref form 3) (list-ref l 1)))
          (if cmd
            (let ()
              (define result (apply (cdr cmd)
                mux
                ; channel
                c
                ; escape procedure
                #f
                ; t,i,r,e
                (car form)
                ; tag
                (cadr form)
                ; cookie
                (caddr form)
                ; command
                (cadddr form)
                ; rest
                (cddddr form)
              ))
              ;#t
              normal-exit-token
            )
            ; perhaps only the mux issues error responses
            (tire-mux-enqueue mux c `(r ,(list-ref form 1) no-such-command))
          )
        ))
        ((i)
          (debugmsg "PROCEDURE COOKIE I NOT YET VALID!\n")
        )
        ((r)
          (debugmsg "PROCEDURE COOKIE R NOT YET VALID!\n")
        )
        ((e)
          (debugmsg "PROCEDURE COOKIE E NOT YET VALID: ~a\n" form)
        )
        ((s)
          (critmsg "PROCEDURE COOKIE GOT CANCEL: LAST THREAD: ~a" sleazy-last-seval_t-command)
          (if sleazy-last-seval_t-command
            (system-async-mark (lambda ()
              (debugmsg "THREAD CANCELLATION!\n")
              (scm-error (quote signal) #f "pseudo-ctrl-c via TIRE" #f (list 2))
            ) sleazy-last-seval_t-command)
          )
        )
        (else (display "BOGUS FORM!\n"))
      )
    ))) c)
  )
)))

(define cookie-add-raw-command! (with-wiki-docstring
  "add a new [[tire | TIRE]] raw command to a cookie. The handler will be called as (mux chan esc tform tag cookie command body). This function is often wrapped."
  (lambda (cookie command handler)
    (list-set! cookie 1 (cons (cons command handler) (list-ref cookie 1)))
    #t
  )
))

(define cookie-add-command! (with-wiki-docstring
  "add a new [[tire | TIRE]] command to a cookie. The handler will be called as (handler mux tag cookie responder body)."
  (lambda (cookie command handler)
    (cookie-add-raw-command! cookie command (lambda (mux chan esc tform tag cookie command . rest)
      (apply handler
        mux
        tag
        cookie
        ; responder. The responder may safely be called multiple times.
        (lambda (rcommand . rrest)
          (tire-mux-enqueue mux chan (cons rcommand (cons tag rrest)))
        )
        ; body
        rest
      )
    ))
  )
))

(define cookie-add-sync-command! (with-wiki-docstring
  "add a new synchronous [[tire | TIRE]] command to a cookie. Synchronous here means that responses generated by the cookie handler are enqueued into the cookie's mux synchronously. The handler will be called as (handler mux tag cookie responder body)."
  (lambda (cookie command handler)
    (cookie-add-raw-command! cookie command (lambda (mux chan esc tform tag cookie command . rest)
      (apply handler
        mux
        tag
        cookie
        ; responder. The responder may safely be called multiple times.
        (lambda (rcommand . rrest)
          (tire-mux-enqueue-sync mux chan (cons rcommand (cons tag rrest)))
        )
        ; body
        rest
      )
    ))
  )
))

(define sleazy-last-seval_t-command #f)

(define cookie-create-proverb-channel (with-docstring
  "create a new channel in mymux that provides supports Proverb evaluation endpoints. <module-or-thunk> is the evaluation environment"
  (lambda (mymux module-or-thunk)
  (define cpair (cookie-create-channel mymux))
  (define rerr (cons 'foo 'bar))
  (define te (make-tire-evaluator module-or-thunk))
  (define forced? #f)
  (define forced-result #f)

  (if (procedure? module-or-thunk)
    (cookie-add-sync-command! (car cpair) 'force (lambda (mux tag cookie responder . rest)
      (if forced?
        (responder 'r forced-result)
        (tire-applier* mux tag cookie responder (lambda ()
          (set! forced-result (module-or-thunk))
          (set! forced? #t)
          forced-result
        ))
      )
    ))
  )
   
  (cookie-add-sync-command! (car cpair) 'eval te)

  (cookie-add-sync-command! (car cpair) 'seval (lambda (mux tag cookie responder body)
    (define schemebody (call-with-input-string body (lambda (ip)
      (define r (catch #t
        (lambda () (read ip))
        (lambda args rerr)
      ))
      (close-input-port ip)
      r
    )))
    (if (eq? rerr schemebody)
      (responder 'r "Scheme parse error...")
      (te mux tag cookie responder schemebody)
    )
  ))

  (cookie-add-sync-command! (car cpair) 'seval_t (lambda (mux tag cookie responder body)
    (define schemebody (call-with-input-string body (lambda (ip)
      (define r (catch #t
        (lambda () (read ip))
        (lambda args rerr)
      ))  
      (close-input-port ip)
      r
    )))
    (if (eq? rerr schemebody)
      (responder 'r "Scheme parse error...")
      (set! sleazy-last-seval_t-command 
        (call-with-new-thread (lambda ()
          (te mux tag cookie responder schemebody)
        ))
      )
    )
  ))

  (cookie-add-sync-command! (car cpair) 'complete (lambda (mux tag cookie responder body)
    (if (string? body)
      (te mux tag cookie responder `(tab-complete-string ,body (current-module)))
      (responder 'r "complete command takes a string")
    )
  ))

  (cookie-add-command! (car cpair) 'listen (with-docstring
    "accept a TCP port number and return the port actually bound"
    (lambda (mux tag cookie responder portno)
      (define p (tire-mux-add-tcp-listener! mux portno))
      (responder 'r p)
    )
  ))

  (cookie-add-command! (car cpair) 'help (lambda (mux tag cookie responder . rest)
    (responder 'r (map car (list-ref (car cpair) 1)))
  ))

  (cookie-add-command! (car cpair) 'fork (lambda (mux tag cookie responder . rest)
    (define cookie (if (null? rest)
      (fork-add-helper* mux responder (or module-or-thunk (current-module)) #f)
      (fork-add-helper* mux responder (or module-or-thunk (current-module)) (car rest))
    ))
    (responder 'r cookie)
  ))

  (cookie-add-raw-command! (car cpair) 'forward (lambda (mux chan esc tform tag cookie command route subcommand . rest)
    (define from (tire-request-get-src (tire-mux-get-newest-request-for-tag mux tag)))
    (if (null? route)
      (tire-mux-enqueue mux from (cons 't (cons tag (cons cookie (cons subcommand rest)))))
      (tire-mux-enqueue mux from (cons 't (cons tag (cons (car route)
        (if (null? (cdr route))
          (cons subcommand rest)
          (cons 'forward (cons (cdr route) (cons subcommand rest)))
        )
      ))))
    )
  ))

  (cookie-add-sync-command! (car cpair) 'kill (lambda (mux tag cookie responder . rest)
    (responder 'i 'unroute cookie)
    (responder 'r 'ok)
    (kill (getpid) SIGKILL)
  ))

  cpair
  )
))

;; there is a remote-tire-connections article

(define loopback-add (with-docstring "like (fork-add), but the new cookie exists in the current process"
  (lambda optional-module
    (define cname (gensym (string-append "loopback-cookie-" (gethostname) "-" (number->string (getpid) 10) "-")))
    (define c (cookie-create-proverb-channel (current-tire-mux) (if (null? optional-module) (current-module) (car optional-module))))
    (tire-mux-add-channel! (current-tire-mux) (cdr c))
    (tire-mux-hash-cookie! (current-tire-mux) cname (cdr c))
    (sleazy-tire-responder 'i 'route cname)
    cname
  )
))

;
; store the child handles here so they won't be garbage collected
;
(define fork-add-helper* (with-docstring
  "fork the current process and return a cookie name. This is a private helper function wrapped by fork-add."
  (lambda (mux responder module-or-thunk cookiename)
    (define child-blivet (concurrent-apply-with-talkback (lambda (parent-port-pair)
      (with-error-to-port (debug-port) (lambda ()
      (with-output-to-port (debug-port) (lambda ()
      (define mymux (tire-mux-setup '()))
      (debugmsg "CHILD BLIVET INITIALIZED\n")
      (let ()
        (define cpair (cookie-create-proverb-channel mymux module-or-thunk))
        (define cookie (generate-local-cookie))
        (tire-mux-setup-buffering! mymux)

        (tire-mux-add-channel! mymux (cdr cpair))
        ; we call ourselves local
        (tire-mux-hash-cookie! mymux 'local (cdr cpair))
        ; we also call ourselves our pid
        (tire-mux-hash-cookie! mymux cookie (cdr cpair))
        ; and an optional user-supplied alias
        (if cookiename
          (tire-mux-hash-cookie! mymux cookiename (cdr cpair))
        )
      )
      (let ()
        ; the child gets the channel added to the mux
        ; for select()ing but does not automatically get
        ; a route added to the parent
        (define parent-channel (make-tire-buffered-fport-channel (car parent-port-pair) (cdr parent-port-pair)))
        (tire-mux-add-channel! mymux parent-channel)
      )
      (run-select-loop-until
        (list (list mymux tire-mux-select-pre tire-mux-select-post))
        #f
        (lambda ()
          (not (port-closed? (car parent-port-pair)))
        )
      )
      (debugmsg "child leaving select loop and exiting")
      ; then we're killed
    ))
    ))
    ) '(parent-port-pair) 'parent-port-pair))
    (let ()
      (debugmsg "sending enum message")
      (display '(t - - enum) (cdr child-blivet))
      (force-output (cdr child-blivet))
      (let ()
        (define child-enum (receive-value (car child-blivet)))
        (define c (make-tire-buffered-fport-channel (car child-blivet) (cdr child-blivet)))
      
        (debugmsg "child enumerated ~a" child-enum)
        (tire-mux-add-channel! mux c)
        (if cookiename
          (begin
            (tire-mux-hash-cookie! mux cookiename c)
            (responder 'i 'route cookiename)
            cookiename
          )
          (begin
            (for-each (lambda (v)
              (tire-mux-hash-cookie! mux v c)
              (responder 'i 'route v)
            ) (list-ref child-enum 2))
            ; we hash all of the children, but only return the first.
            ; we should adjust the definition of this function at some point to
            ; allow the multiple return.
            (car (list-ref child-enum 2))
          )
        )
      )
    )
  )
))

(define module-fork-add (with-docstring
  "like fork-add. The environment for the subordiate Scheme evaluator is the module <module>"
  (lambda (module)
    (fork-add-helper* (current-tire-mux) sleazy-tire-responder module #f)
  )
))

(define fork-add (with-docstring
  "Fork the current Scheme interpreter and run a new TIRE mux in the child. A cookie naming the new endpoint is returned."
  (lambda ()
  (module-fork-add (current-module))
  ))
)

(define multi-module-fork-add (with-docstring
  "like (multi-fork-add <num>). The environment for the subordiate Scheme evaluator is the module <module>"
  (lambda (module num)
    (let helper ((c num) (acc '()))
      (if (equal? c 0)
        acc
        (helper (- c 1) (cons (module-fork-add module) acc))
      )
    )
  )
))

(define multi-fork-add (with-docstring
  "Fork the current Scheme interpreter <num> times and run a new TIRE mux in each child. A list of cookies naming the new endpoints is returned."
  (lambda (num)
    (let helper ((c num) (acc '()))
      (if (equal? c 0)
        acc
        (helper (- c 1) (cons (fork-add) acc))
      )
    )
  )
))

(define fork-eval-w-callback 
  (with-docstring 
    "evaluates form on forked proc, exiting the proc upon completion.
    Returns a function f, where (f 'finished?) will return true if the
    evaluation is complete and (f) will return the result of the eval,
    and will block until the eval has been completed.
    The function callback will be run in the current process with the 
    result of the eval when the eval returns from the forked proc."
    (lambda (form callback)
      (let ((handle (fork-add))
	    (remote-code `(let ((responder (snatch-responder)))
			    (begin
			      (responder 'r ,form)
			      (kill (getpid) SIGKILL))))
	    (callbk (lambda (arg) (begin (callback arg) (fork-delete! handle)))))
	(let ((ftns (tire-eval-general handle remote-code callbk)))
	  (lambda param (if (and (not (null? param)) (eq? (car param) 'finished?)) 
			  ((cdr ftns)) ((car ftns))))
	  )))))

(define fork-eval 
  (with-docstring 
    "evaluates form on forked proc, exiting the proc upon completion.
    Returns a function f, where (f 'finished?) will return true if the
    evaluation is complete and (f) will return the result of the eval,
    and will block until the eval has been completed."
    (lambda (form)
      (fork-eval-w-callback form (lambda x #t)))))

(define fork-delay* (with-docstring
  "helper function wrapped by the macro fork-delay"
  (lambda (thunk)
  (define cookie (fork-add-helper* (current-tire-mux) sleazy-tire-responder thunk #f))
  (delay (let ()
    (define result (tire-eval `(t ,(gensym "remote-force-") ,cookie force)))
    (fork-delete! cookie)
    result
  ))
  )
))

(define-syntax fork-delay (syntax-rules ()
  ((fork-delay value)
    (fork-delay* (lambda () value))
  )
))

(define-syntax fork-begin (syntax-rules ()
  ((fork-begin terms ...)
    (force (fork-delay* (lambda () terms ...)))
  )
))

(define my-rank 0)
(define my-size 0)
(cond-expand
(guile
;(define dummy-iterator-sym (gensym "dummy-iterator-"))
;(define dummy-size-sym (gensym "dummy-size"))
;(define-macro (multi-fork-begin num . terms)
;  `(let ()
;    (define ,dummy-size-sym ,num)
;    (map (lambda (,dummy-iterator-sym) (force (fork-delay* (lambda () (set! my-rank ,dummy-iterator-sym) (set! my-size ,dummy-size-sym) (let () ,@terms))))) (let helper ((c (- ,dummy-size-sym 1)) (racc '())) (if (equal? c 0) (cons c racc) (helper (- c 1) (cons c racc)))))
;  )
;)
(define-syntax multi-fork-begin (syntax-rules ()
  ((multi-fork-begin num terms ...)
    (map (lambda (v) (force (fork-delay* (lambda () (set! my-rank v) (set! my-size num) (begin terms ...))))) (let helper ((c (- num 1)) (racc '())) (if (equal? c 0) (cons c racc) (helper (- c 1) (cons c racc)))))
  )
))
)
(chez
  (extend-syntax (multi-fork-begin)
    [(multi-fork-begin num terms ...)
     (map (lambda (v) (force (fork-delay* (lambda () terms ...)))) (make-list num))
    ]
  )
)
(else
  (define-syntax multi-fork-begin (syntax-rules ()
    ((multi-fork-begin num terms ...)
      (display "YOU WISH!\n")
    )
  ))
)
)

(define fork-delete! (with-docstring
  "(fork-delete! COOKIE) -- kill a forked process by cookie. This will kill its children as well."
  (lambda (cookie)
    (define mux (current-tire-mux))
    (let ()
      (define channel (tire-mux-lookup-cookie mux cookie))
      (let ()
        (define cookies-on-channel (assq
          channel
          (aggregate-alist-for-assq (map (lambda (v)
            (cons (tire-mux-lookup-cookie mux v) v))
            (vector->list (hashtable-keys (tire-mux-get-cookie-hash mux)))))
        ))
        (if channel
          (if (and cookies-on-channel
                   (member cookie (cdr cookies-on-channel))
                   (eq? (length (cdr cookies-on-channel)) 1)
              )
            (begin
              (tire-mux-delete-channel! mux channel)
              (sleazy-tire-responder 'i 'unroute cookie)
              #t
            )
            ; else
            (begin
              (critmsg "Crikey! Would have deleted a shared channel!\n")
              (tire-mux-unhash-cookie! mux cookie)
              (sleazy-tire-responder 'i 'unroute cookie)
              #t
            )
          )
          ; else
          #f
        )
      )
    )
  )
))

(define url-add (with-docstring
"(url-add TIRE-URL)
accept a tire URL naming a cookie and create a channel and route for it in the local mux.
returns the cookie."
  (lambda (url)
    (define c (make-tire-url-channel url))
    (define p (tire-parse-url url))
    (if c
      (let ()
        (define cookie (list-ref p 3))
        (define m (current-tire-mux))
        (tire-mux-add-channel! m c)
        (tire-mux-hash-cookie! m cookie c)
        (sleazy-tire-responder 'i 'route cookie)
        cookie
      )
      (error "couldn't resolve")
    )
  )
))

(define popen-fports (with-docstring
"(popen-fports COMMAND-STRING) accept a command string and return a pair of ports"
(lambda (command-line)
  (define ports (concurrent-apply-with-talkback (lambda (s)
        (dup2 (port->fdes (car s)) 0)
        (dup2 (port->fdes (cdr s)) 1)
        (system command-line)
        (debugmsg "popen-connect: slave concluded")
  ) '(#t) #t))
  ports
)
))

(define popen-connect (with-docstring
"(popen-connect COMMAND-STRING) accept a command string and return a TIRE channel"
(lambda (command-line)
  (define ports (popen-fports command-line))
  (make-tire-buffered-fport-channel (car ports) (cdr ports))
)
))

(define generate-local-cookie (with-docstring
  "generate a globally-unique cookie name for local evaluation environments"
  (lambda ()
    (string->symbol (string-append (gethostname) "-" (number->string (getpid))))
  )
))

(define popen-add (with-docstring
"(popen-add COMMAND-STRING) accept a command string and create a channel and route for it in the local mux. returns the cookie. This command must be called from a tire eval environment."
(lambda (command-line)
  (define ports (popen-fports command-line))
  (display '(t - local eval (generate-local-cookie)) (cdr ports))
  (studio-flush-port (cdr ports))
  (let ()
    (define res (receive-value (car ports)))
    (let ()
      (define channel (make-tire-buffered-fport-channel (car ports) (cdr ports)))
      (define m (current-tire-mux))
      (tire-mux-add-channel! m channel)
      (tire-mux-hash-cookie! m (list-ref res 2) channel)
      (sleazy-tire-responder 'i 'route (list-ref res 2))
      (list-ref res 2)
    )
  )
)
))

(define remote-tire-do (with-wiki-docstring
  "(remote-tire-do dest command . rest) -- see [[proverb-parallel]] for more info"
  (lambda (dest command . rest)
  (define tag-base (string-append "remote-eval-" (gethostname) "-" (number->string (getpid)) "-"))
  (cond
    ((list? dest) (let ()
      (define yield-key (gensym "tire-yield-key-"))
      (define tags (map (lambda (v) (gensym (string-append "concurrent-" tag-base))) dest))
      ; pre-build our return value
      (define results (map (lambda (v) #f) dest))
      (define outstanding-rs (length dest))
      (define sc sleazy-tire-responder)
      ; this hash is going to have as its values the cons cells into which the
      ; results can be directly jammed.
      (define hashup (studio-make-hash-table hash eq? (length dest)))
      (let ()
        (define aggregating-channel (make-tire-procedure-channel (lambda (incoming-form)
          (debugmsg "AGGREGATING CHANNEL: ~a\n" incoming-form)
          (if (and (list? incoming-form) (not (null? incoming-form)))
            (cond
              ((eq? (car incoming-form) 'i)
                ; forward the i to the original requestor
                (apply sc 'i (cddr incoming-form))
              )
              ((eq? (car incoming-form) 'r)
                ; aggregate the result
                (set! outstanding-rs (- outstanding-rs 1))
                (set-car! (hash-ref hashup (list-ref incoming-form 1)) (list-ref incoming-form 2))
                (if (eq? outstanding-rs 0)
                  (begin
                    (debugmsg "ABOUT TO RESUME!\n")
                    ;(tire-eval-resume yield-key results)
                    (bogo-resume yield-key results)
                  )
                )
              )
              ((eq? (car incoming-form) 'e)
                (apply sc 'i 'forwarded-error (cddr incoming-form))
              )
              (else (error "bogo"))
            )
          )
        )))
        ; pre-weave our results in order
        (let hash-builder ((tag-cursor tags) (dest-cursor results))
          (if (not (null? dest-cursor))
            (begin
              (hash-set! hashup (car tag-cursor) dest-cursor)
              (hash-builder (cdr tag-cursor) (cdr dest-cursor))
            )
          )
        )
        (for-each (lambda (t d)
          (tire-mux-buffer-enqueue (current-tire-mux) aggregating-channel `(t ,t ,d ,command ,@rest))
        ) tags dest)
        (tire-eval-yield yield-key)
      )
    ))
    (else ; dest is a cookie
      (tire-eval `(t ,(gensym tag-base) ,dest ,command ,@rest))
    )
  )
 )
))

(define remote-eval (with-wiki-docstring
  "(remote-eval dest form) -- see [[proverb-parallel]] for more info"
  (lambda (dest form)
    (remote-tire-do dest 'eval form)
  )
))

(define remote-seval (with-wiki-docstring
  "(remote-eval dest form-str) -- see [[proverb-parallel]] for more info"
  (lambda (dest form-str)
    (remote-tire-do dest 'seval form-str)
  )
))

(define tire-make-connect-back (with-docstring
  "return a pair ( cookie . yielder ). accept count connections before yielding"
  (lambda (count)
  (define left count)
  (define children '())
  (define cpair (cookie-create-channel (current-tire-mux)))
  (define csym (gensym "callback-cookie"))
  (cookie-add-raw-command! (car cpair) 'connectback (lambda (mux chan esc tform tag cookie command body)
    (define p (tire-mux-get-oldest-request-for-tag mux tag))
    (if (> left 0)
      (begin
        (set! children (cons body children))
        (tire-mux-hash-cookie! mux body (tire-request-get-src p))
        (tire-mux-enqueue mux chan `(r ,tag ,(- count left)))
        (set! left (- left 1))
        (if (eq? left 0)
          (begin
            (tire-mux-unhash-cookie! mux csym)
            (bogo-resume csym (reverse children))
          )
        )
      )
      (begin
        (critmsg "TOO MANY CONNECTBACKS!\n")
        (tire-mux-enqueue mux chan `(r ,tag #f))
      )
    )
  ))
  (tire-mux-hash-cookie! (current-tire-mux) csym (cdr cpair))
  (cons csym  (lambda ()
    (critmsg "YIELDING!\n")
    (tire-eval-yield csym)
  )) 
  )
))

(define tire-open-back-func (with-docstring
  "accept <func> and <count>. <func> is called wit a TIRE connect-back URL and count. This
function blocks until the <count> clients have connected back."
  (lambda (func count)
    (define b (tire-make-connect-back-url count))
    (func (car b) count) 
    ((cdr b))
  )
))

(define tire-open-back-system (with-docstring
  "accept a format string <fmt-string> and <count>. A new string for system will be
  built from fmt-string and a connect back URL and <count>. This function blocks until <count> clients
 have connected back."
  (lambda (fmt-string count)
    (tire-open-back-func (lambda (url subcount)
      (system (format #f fmt-string url count))
    ) count)
  )
))

(define tire-make-connect-back-url (with-docstring
  "return a pair ( <URL> . <blocker> ). <URL> is a new TIRE URL that can be used by other processes to connect back to the current TIRE mux. <blocker> is a function that will block until <count> processes have connected back."
  (lambda (count)
    (define l (tire-mux-get-listeners (current-tire-mux)))
    (if (null? l)
      (let ()
        (define l (tire-mux-add-tcp-listener! (current-tire-mux) 0))  
        (tire-make-connect-back-url count)
      )
      (let ()
        (define c (tire-make-connect-back count))
        (cons
          (string-append "tire://" (gethostname) ":" (number->string (vector-ref (getsockname (car l)) 2)) "/" (symbol->string (car c)))
          (cdr c)
        )
      )
    )
  )
))

(define tire-slave-connect-back (lambda (mux url next)
  (define u (tire-parse-url url))
  (define s (gensym "tag-"))
  (define r (make-tire-procedure-channel (lambda (form)
    (if (and (list? form) (> (length form) 2) (eq? (car form) 'r) (eq? (cadr form) s))
      (tire-mux-unhash-cookie! mux (list-ref u 3))
    )
    (critmsg "GOT A RETURN FORM!\n")
    (next form)
  )))
  (let ()
    (define c ((list-ref u 0)))
    (tire-mux-add-channel! mux c) 
    (tire-mux-hash-cookie! mux (list-ref u 3) c)
    (tire-mux-enqueue mux r `(t ,s ,(list-ref u 3) connectback ,(generate-local-cookie)))
  )
))

;; there is a proverb-parallel article in the wiki

; non-define actions

(set! dribble-output-port (make-dribble-port 'output))
(set! dribble-error-port (make-dribble-port 'error))

;
; defeat libc pid caching
;
(if (access? "/proc/self" R_OK)
  (set! getpid (with-docstring "a non-caching variant of getpid()" (lambda ()
    (string->number (readlink "/proc/self"))
  )))
)

)
