;
; Scheme data conversion functions for guile studio
;
; these functions must be written in pure scheme (with the exception of lambda-w-types)
; -- in particular, these routines can't use *unspecified*
;

(library (proverb dataconv)
  (export
    ; data conversion from ports
    read-uint8
    read-uint16-be
    read-uint16-le
    ; data conversion from strings
    ;; unsigned conversions
    string->uint8
    string->uint16-be
    string->uint16-le
    string->uint32-le
    string->uint32-be
    ; data conversion from lists of string references
    ref->uint8
    ; data conversion among integer types
    uint8s->uint16
    uint8s->uint32
    base64->string
  )
  (import
    (rnrs)
    (rnrs arithmetic bitwise)
    (proverb doc)
    (proverb debug)
    (proverb types)
    (only (proverb dialect)
      studio-read-string!/full 
    )
    (only (guile) list-set! the-eof-object integer-expt make-list)
  )

;
; this file has very high redundancy
; a help article is available as proverb-data-conversion
;

(define read-uint8 (with-docstring
  "read 8 bits from port <port> and convert to an unsigned integer"
  (lambda (port)
  (define val (read-char port))
  (if (eof-object? val)
    val
    (char->integer val)
  )
  )
))

(define string->uint8 (with-docstring
  "turn the first character of the string <str> into an integer"
  (lambda (str)
    (char->integer (string-ref str 0))
  )
))

(define string->uint16-le (with-docstring
  "turn the first two characters of the string <str> into an unsigned integer using little endianness"
  (lambda (str)
    (+ (char->integer (string-ref str 0)) (* 256 (char->integer (string-ref str 1))))
  )
))

(define string->uint16-be (with-docstring
  "turn the first two characters of the string <str> into an unsigned integer using big endianness"
  (lambda (str)
    (+ (char->integer (string-ref str 1)) (* 256 (char->integer (string-ref str 0))))
  )
))

(define string->uint32-le (with-docstring
  "turn the first four characters of the string <str> into an unsigned integer using little endianness"
  (lambda (str)
    (+ (char->integer (string-ref str 0))
       (* #x100 (char->integer (string-ref str 1)))
       (* #x10000 (char->integer (string-ref str 2)))
       (* #x1000000 (char->integer (string-ref str 3)))
    )
  )
))

(define string->uint32-le (with-docstring
  "turn the first four characters of the string <str> into an unsigned integer using little endianness"
  (lambda (str)
    (+ (char->integer (string-ref str 0))
       (* #x100 (char->integer (string-ref str 1)))
       (* #x10000 (char->integer (string-ref str 2)))
       (* #x1000000 (char->integer (string-ref str 3)))
    )
  )
))

(define string->uint32-be (with-docstring
  "turn the first four characters of the string <str> into an unsigned integer using little endianness"
  (lambda (str)
    (+ (char->integer (string-ref str 3))
       (* #x100 (char->integer (string-ref str 2)))
       (* #x10000 (char->integer (string-ref str 1)))
       (* #x1000000 (char->integer (string-ref str 0)))
    )
  )
))

(define int32-be->string (with-docstring
  "convert <int> to a string as a 32 bit unsigned integer arranged MSB to LSB"
  (lambda-w-types ((int integer?))
    (define val1 (integer->char (bitwise-and int #xff)))
    (define val2 (integer->char (/ (bitwise-and int #xff00) #x100)))
    (define val3 (integer->char (/ (bitwise-and int #xff0000) #x10000)))
    (define val4 (integer->char (/ (bitwise-and int #xff000000) #x1000000)))

    (string val4 val3 val2 val1)
  )
))

(define int32-le->string (with-docstring
  "convert <int> to a string as a 32 bit unsigned integer arranged MSB to LSB"
  (lambda-w-types ((int integer?))
    (define val1 (integer->char (bitwise-and int #xff)))
    (define val2 (integer->char (/ (bitwise-and int #xff00) #x100)))
    (define val3 (integer->char (/ (bitwise-and int #xff0000) #x10000)))
    (define val4 (integer->char (/ (bitwise-and int #xff000000) #x1000000)))

    (string val1 val2 val3 val4)
  )
))

(define read-uint16-le (with-docstring
  "read 16 bits arranged LSB to MSB from <port> and convert to an unsigned integer"
  (lambda (port)
    (define buf (make-string 2))
    (let ()
      (define res (studio-read-string!/full buf port))
      (if res
        (string->uint16-le buf)
        the-eof-object
      )
    )
  )
))

(define read-uint16-be (with-docstring
  "read 16 bits arranged MSB to LSB from <port> and convert to an unsigned integer"
  (lambda (port)
    (define buf (make-string 2))
    (let ()
      (define res (studio-read-string!/full buf port))
      (if res
        (string->uint16-be buf)
        the-eof-object
      )
    )
  )
))

(define read-int32-le (with-docstring
  "read 32 bits arranged LSB to MSB from <port> and convert to an unsigned integer"
  (lambda (port)
    (define buf (make-string 4))
    (let ()
      (define res (studio-read-string!/full buf port))
      (if res
        (string->uint32-le buf)
        the-eof-object
      )
    )
  )
))

(define write-uint16-le (with-docstring
  "write <int> to <port> as a 16 bit unsigned integer arranged LSB to MSB"
  (lambda-w-types ((int integer?) (port output-port?))
  (define val1 (write-char (integer->char (bitwise-and int #xff)) port))
  (if (eof-object? val1)
    val1
    (let ()
      (define val2 (write-char (integer->char (/ (bitwise-and int #xff00) #x100)) port))
      (if (eof-object? val2)
        val2
        #t
      )
    )
  )
  )
))

(define write-int32-le (with-docstring
  "write <int> to <port> as a 32 bit unsigned integer arranged LSB to MSB"
  (lambda-w-types ((int integer?) (port output-port?))
  (define val1 (write-char (integer->char (bitwise-and int #xff)) port))
  (if (eof-object? val1)
    val1
    (let ()
      (define val2 (write-char (integer->char (/ (bitwise-and int #xff00) #x100)) port))
      (if (eof-object? val2)
        val2
        (let ()
          (define val3 (write-char (integer->char (/ (bitwise-and int #xff0000) #x10000)) port))
	  (if (eof-object? val3)
            val3
	    (let ()
	      (define val4 (write-char (integer->char (/ (bitwise-and int #xff000000) #x1000000)) port))
	      (if (eof-object? val4)
                val4
                #t
              )
            )
          )
        )
      )
    )
  )
  )
))

(define integer->lefloat32 (with-docstring
  "convert the integer <int> into a string representing a little-endian floating point value"
  (let ()
    ; find the highest order bit in i
    (define integer->shift (lambda (i)
      (let helper ((v 22))
        (if (eq? v 0)
          0
          (if (bitwise-and i (integer-expt 2 v))
            v
            (helper (- v 1))
          )
        )
      )
    ))

    (lambda (int)
      (define msb (integer->shift int))
      (list->string (if (eq? int 0)
        (list->string (map integer->char (list 0 0 0 0)))
        (list->string (map integer->char (list
          (bitwise-and #xff (quotient (* int (integer-expt 2 (- 23 msb))) (integer-expt 2 0)))
          (bitwise-and #xff (quotient (* int (integer-expt 2 (- 23 msb))) (integer-expt 2 8)))
          (bitwise-ior
            (bitwise-and #x7f (quotient (* int (integer-expt 2 (- 23 msb))) (integer-expt 2 16)))
            (* (integer-expt 2 7) (bitwise-and (+ 127 msb) 1))
          )
          (quotient (+ 127 msb) 2)
        )))
      ))
    )
  )
))

(define uint8s->uint32 (with-docstring
   "convert the unsigned 8-bit integers <lsb>,<b1>,<b2>,<msb> into an unsigned integer with the bytes ordered from least significant byte to most significant byte"
   (lambda (lsb b1 b2 msb)
  (+
    (* #x00000001 lsb)
    (* #x00000100 b1)
    (* #x00010000 b2)
    (* #x01000000 msb)
  )
  )
))

(define uint8s->uint16 (with-docstring
  "convert the unsigned 8-bit integers <lsb>,<msb> into an unsigned integer using <lsb> as the least significant byte and <msb> as the most significant byte"
  (lambda (lsb msb)
  (+
    (* #x0001 lsb)
    (* #x0100 msb)
  )
  )
))

(define bytes->uint32 (with-docstring
   "convert the bytes <lsb>,<b1>,<b2>,<msb> into an unsigned integer with the bytes ordered from least significant byte to most significant byte"
  (lambda (lsb b1 b2 msb)
 (uint8s->uint32
   (char->integer lsb)
   (char->integer b1)
   (char->integer b2)
   (char->integer msb)
 )
  )
))

(define bytes->uint16 (with-docstring
   "convert the bytes <lsb>,<msb> into an unsigned integer using <lsb> as the least significant byte and <msb> as the most significant byte"
  (lambda (lsb msb)
 (uint8s->uint16
   (char->integer lsb)
   (char->integer msb)
 )
  )
))

(define refs->uint32 (with-docstring
   "convert the bytes at indicies <lsr>,<r1>,<r2>,<msr> in string <str> into an unsigned integer with the indicies ordered from least significant byte to most significant byte"
  (lambda (str lsr r1 r2 msr)
    (apply bytes->uint32 (map (lambda (v) (string-ref str v)) (list lsr r1 r2 msr)))
  )
))

(define refs->uint16 (with-docstring
   "convert the bytes at indicies <lsr>,<msr> in string <str> into an unsigned integer using <str>[<lsr>] as the least significant byte and <str>[<msr>] as the most significant byte"
  (lambda (str lsr msr)
    (apply bytes->uint16 (map (lambda (v) (string-ref str v)) (list lsr msr)))
  )
))

(define ref->uint8 (with-docstring
  "convert the byte at index <index> in string <str> into an 8 bit unsigned integer"
  (lambda (str index)
    (char->integer (string-ref str index))
  )
))

(define leref->uint32 (with-docstring
  "convert the four bytes starting at index <msr> in string <str> in little-endian order into a 32 bit unsigned integer"
  (lambda (str lsr)
    (apply refs->uint32 (list str lsr (+ lsr 1) (+ lsr 2) (+ lsr 3)))
  )
))

(define leref->uint16 (with-docstring
  "convert the two bytes starting at index <msr> in string <str> in little-endian order into a 16 bit unsigned integer"
  (lambda (str lsr)
    (apply refs->uint16 (list str lsr (+ lsr 1)))
  )
))

(define beref->uint32 (with-docstring
  "convert the four bytes starting at index <msr> in string <str> in big-endian order into a 32 bit unsigned integer"
  (lambda (str msr)
    (refs->uint32 str (+ msr 3) (+ msr 2) (+ msr 1) msr)
  )
))

(define beref->uint16 (with-docstring
  "convert the two bytes starting at index <msr> in string <str> in big-endian order into a 16 bit unsigned integer"
  (lambda (str msr)
    (refs->uint16 str (+ msr 1) msr)
  )
))

(define base64rev #f)

(define base64->string (with-docstring
  "accept a string of base-64 encoded data and return the contents as a new string"
  (lambda (b64s)
    (define bits (let helper ((cursor (string->list b64s)) (racc '()))
      (if (null? cursor)
        (reverse racc)
        (let ()
          (define bits (assoc (car cursor) base64rev))
          (if bits
            (helper (cdr cursor) (cons (cdr bits) racc))
            (helper (cdr cursor) racc)
          )   
        )
      )
    ))
    (define all6   (lambda (v) v))
    (define low2   (lambda (v) (bitwise-and v #x03)))
    (define low4   (lambda (v) (bitwise-and v #x0f)))
    (define high2  (lambda (v) (/ (bitwise-and v #x30) (expt 2 4))))
    (define high4  (lambda (v) (/ (bitwise-and v #x3c) (expt 2 2))))
    (define shift2 (lambda (v) (* (expt 2 2) v)))
    (define shift4 (lambda (v) (* (expt 2 4) v)))
    (define shift6 (lambda (v) (* (expt 2 6) v)))

    (let helper ((cursor bits) (acc '()))
      (if (null? cursor)
        (list->string (map integer->char acc))
        (if (< (length cursor) 4)
          (helper (append cursor (make-list (- 4 (length cursor)) 0)) acc)
          (let ()
            (helper (list-tail cursor 4) (append acc (list
              (+ (shift2 (all6 (list-ref cursor 0))) (high2 (list-ref cursor 1)))
              (+ (shift4 (low4 (list-ref cursor 1))) (high4 (list-ref cursor 2)))
              (+ (shift6 (low2 (list-ref cursor 2))) (all6 (list-ref cursor 3)))
            )))
          )
        )
      )
    )
  )
))

(set! base64rev (let ()
  (define base64str "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/")
  (let helper ((cursor (string->list base64str)) (index 0) (racc '()))
    (if (null? cursor)
      (reverse racc)
      (helper (cdr cursor) (+ index 1) (cons (cons (car cursor) index) racc))
    )
  )
))

)
