;
; Scheme utility functions for guile studio
;
; these functions must be written in pure scheme (with the exception of lambda-w-types)
; -- in particular, these routines can't use *unspecified*
;

(define alist? (with-docstring
  "is <value> an alist?"
  (lambda (value)
    (define helper (lambda (hl)
      (if (null? hl)
        #t
        (if (pair? (car hl))
          (helper (cdr hl))
          #f
        )
      )
    ))
    (and (list? value) (helper value))
  )
))

(define value? (with-docstring
  "is <val> a Scheme object?"
  (lambda (val)
    #t
  )
))

(define aggregate-alist-with-equality-and-comparison-and-merge (with-docstring
  "accpet an alist with duplicated keys and return a new alist with unique keys and aggregated values. keys are sorted with the comparison function <comparison> and compared with <equality>. values are aggregated with <merge>. This function is usually called through a wrapper like aggregate-alist-for-assoc"
  (lambda (alist equality comparison merge)
    (define helper (lambda (sub carry-val carry-racc racc)
      (if (null? sub)
        (reverse (cons (cons carry-val (reverse carry-racc)) racc))
        (if (equality carry-val (caar sub))
          (helper (cdr sub) carry-val (merge (cdar sub) carry-racc) racc)
  	  ; else
          (helper (cdr sub) (caar sub) (list (cdar sub)) (cons (cons carry-val (reverse carry-racc)) racc))
        )
      )
    ))
    (if (null? alist)
      '()
      (let ()
        (define sorted-list (sort alist (lambda (a b) (comparison (car a) (car b)))))
        (helper sorted-list (caar sorted-list) '() '())
      )
    )
  )
))

(define aggregate-alist-with-equality-and-comparison (with-docstring
  "wrapped version of aggregate-alist-with-equality-and-comparison-and-merge that uses cons for merging"
  (lambda (alist equality comparison)
    (aggregate-alist-with-equality-and-comparison-and-merge alist equality comparison cons)
  )
))

(define with-property (lambda (prop propval val)
  (set-procedure-property! val prop propval)
  val
))

(define aggregate-alist-for-assoc (let ()
  (define crusty-comparator (lambda (a b)
    (if (equal? a b)
      #t
      (cond
        ((and (string? a) (string? b)) (string>=? a b))
        (else (>= (object-address a) (object-address b)))
      )
   )
  ))
  (with-docstring
    "wrapped version of aggregate-alist-with-equality-and-comparison-and-merge that uses the same equality test as assoc"
    (with-property 'reduce-merge
                   (lambda (args) 
                     (aggregate-alist-with-equality-and-comparison-and-merge (apply append args) equal? crusty-comparator append)
                   )
                   (lambda (alist)
                     (aggregate-alist-with-equality-and-comparison alist equal? crusty-comparator)
                   )
    )
  )
))

(define aggregate-alist-for-assq (with-docstring
  "wrapped version of aggregate-alist-with-equality-and-comparison-and-merge that uses the same equality test as assq"
  (lambda (alist)
    (define crusty-comparator (lambda (a b)
      (or (eq? a b) (>= (object-address a) (object-address b)))
    ))
    (aggregate-alist-with-equality-and-comparison alist eq? crusty-comparator)
  )
))

(define aggregate-alist-for-assv (with-docstring
  "wrapped version of aggregate-alist-with-equality-and-comparison-and-merge that uses the same equality test as assv"
  (lambda (alist)
    (define crusty-comparator (lambda (a b)
      (or (eqv? a b) (>= (object-address a) (object-address b)))
    ))
    (aggregate-alist-with-equality-and-comparison alist eqv? crusty-comparator)
  )
))

(define list-has-cycle? (with-docstring
  "does the list <lst> have a cycle?"
  (lambda-w-types ((lst list?))
  (define helper (lambda (hl)
    (if (eq? hl '())
      (list-has-cycle? (cdr lst))
      (if (eq? (cdr hl) lst)
        lst ; return the point where we found the cycle
        (helper (cdr hl))  
      )
    )
  ))
  (if (eq? lst '())
    #f
    (helper lst)
  )
  )
))

(define duplicate-alist (with-docstring
  "build a new copy of <alist> with new cons cells for each key, value pair"
  (lambda-w-types ((alist alist?))
  (define helper (lambda (l racc)
    (if (eq? l '())
      (reverse racc)
      (helper (cdr l) (cons (cons (car (car l)) (cdr (car l))) racc))
    )
  ))
  (helper alist '())
  )
))

(define merge-alists (with-docstring
  "build a new alist by merging <upper> and <lower> with key collisions arbitrated in favor of <upper>"
  (lambda-w-types ((upper alist?) (lower alist?))
  (let helper ((old lower) (new upper))
    (if (null? old)
      (reverse new)
      ; else
      (helper
        (cdr old)
        (if (assoc (caar old) new)
          new
          ; else
          (cons (car old) new)
        )
      )
    )
  )
  )
))

(define index-map (with-docstring
  "like map on a single list. <procedure> is called like (procedure index value)"
  (lambda-w-types ((procedure procedure?) (list list?))
  (define helper (lambda (nl c racc)
    (if (null? nl) 
      (reverse racc)
      (helper (cdr nl) (+ c 1) (cons (procedure c (car nl)) racc))
    )
  ))
  (helper list 0 '())
)))

(define vector-index-map (with-docstring
"Form new vector of <func:procedure of length vecs + 1 args> applied to index
  and indexth value in each position of vecs in turn
    where <vecs: list of vecs> are all of same length. index is supplied to func as argument 0."
  (lambda ( func . vecs)
  (if (> (length vecs) 0)
      (let ((newvec (make-vector (vector-length (car vecs)) #f)))
	(do ((i 0 (1+ i)))
	    ((>= i (vector-length (car vecs))) newvec)
	  (vector-set!
	   newvec i
	   (apply func i (map (lambda (vec) (vector-ref vec i)) vecs)))
      ) )
      (vector)
) ) ) )

(define vector-for-each (with-docstring
  "Apply <func:procedure of length vecs args> to each position of vecs in turn
    where <vecs: list of vecs> are all of same length"
  (lambda (func . vecs)
  (if (> (length vecs) 0)
      (do ((i 0 (1+ i)))
	  ((>= i (vector-length (car vecs))) #t)
	(apply func (map (lambda (vec) (vector-ref vec i)) vecs))
      )
      #t
) ) ) )

(define vector-index-for-each (with-docstring
  "Apply <func:procedure of length vecs args + 1> to index and indexth value
    in each position of vecs in turn
    where <vecs: list of vecs> are all of same length"
  (lambda (func . vecs)
  (if (> (length vecs) 0)
      (do ((i 0 (1+ i)))
	  ((>= i (vector-length (car vecs))) #t)
	(apply func i (map (lambda (vec) (vector-ref vec i)) vecs))
      )
      #t
) ) ) )

(define vector-head (with-docstring
  "like list-head, but for vectors"
  (lambda (vec keep-count)
    ;(if (or (< keep-count 0) (> keep-count (vector-length vec)) )
    ;  (throw 'out-of-range 'vector-head "Argument ~S out of range: ~S, should be ~a,~a" (list 2 keep-count 0 (vector-length vec)) #f)
    ;)
    (let ((newvec (make-vector keep-count #f)))
      (do ((i 0 (1+ i)))
 	  ((>= i keep-count) newvec)
        (vector-set! newvec i (vector-ref vec i))
      )
    )
  )
))

(define vector-tail (with-docstring
  "like list-tail, but for vectors. A new vector is created."
  (lambda (vec skip-count)
    ;(if (or (< skip-count 0) (> skip-count (vector-length vec)) )
    ;  (throw 'out-of-range 'vector-head "Argument ~S out of range: ~S, should be ~a,~a" (list 2 skip-count 0 (vector-length vec)) #f)
    ;)
    (let ()
      (define n-comp (- (vector-length vec) skip-count)) ; n complement
      (let ((newvec (make-vector n-comp #f)))
        (do ((i 0 (1+ i)))
   	  ((>= i n-comp) newvec)
          (vector-set! newvec i (vector-ref vec (+ i skip-count)))
        )
      )
    )
  )
))

(define all? (with-docstring
  "does list <l> have exclusively non-false members?"
  (lambda-w-types ((lst list?))
  (if (null? lst)
      #t
      (and (car lst) (all? (cdr lst)))
  )
  )
))

(define alist-with-key-type? (with-docstring
  "Is an object an alist with elements that each have keys that satisfy the predicate <proc>"
  (lambda-w-types ((lst list?) (proc procedure?))
  (define helper (lambda (hl)
    (if (null? hl)
      #t
      (if (and (pair? (car hl)) (proc (caar hl)))
        (helper (cdr hl))
        #f
      )
    )
  ))
  (and (list? lst) (helper lst))
  )
))

(define uniq (with-docstring
  "dedupe a sorted list <lst> using eq?"
  (lambda (lst)
  (if (null? lst)
    '()
    (let helper ((cursor (cdr lst)) (carry (car lst)) (racc (list (car lst))))
      (if (null? cursor)
        (reverse racc)
        (helper (cdr cursor) (car cursor) (if (eq? carry (car cursor))
          racc
          (cons (car cursor) racc)
        ))
      )
    )
  )
  )
))

(define unique (with-docstring
  "dedupe a sorted list using equal?"
  (lambda (lst)
  (if (null? lst)
    '()
    (let helper ((cursor (cdr lst)) (carry (car lst)) (racc (list (car lst))))
      (if (null? cursor)
        (reverse racc)
        (helper (cdr cursor) (car cursor) (if (equal? carry (car cursor))
          racc
          (cons (car cursor) racc)
        ))
      )
    )
  )
  )
))

(define bmsearch (with-docstring
 "backtracking binary monotonic #f to non-false inflection point search. returns last <cursor> that returns #f for (<predicate> <cursor>). divider accepts a span and a 'towards' direction #t/#f and yields a new span"
  (lambda (cursor beachhead span divider predicate transformer)
  (define inflection (predicate cursor))
  (if span
    ; we still have a ways to go.
    (if inflection
      (bmsearch (transformer beachhead span) beachhead (divider span #f) divider predicate transformer)
      (bmsearch (transformer cursor span) cursor (divider span #t) divider predicate transformer)
    )
    ; end of the road. we found the point or there is no point.
    (and inflection beachhead)
  )
  )
))

(define find-concave-inflections (with-docstring
  "return the list of indicices in the list <lst> that represent concave inflection points"
  (lambda (lst)
    ; index is the index of p
    (let helper ((cursor (cddr lst)) (index 1) (gp (car lst)) (p (cadr lst)) (racc '()))
      (if (null? cursor) 
        (reverse racc)
        (cond
          ((equal? (car cursor) p)
            (helper (cdr cursor) (+ index 1) gp (car cursor) racc)
          )
          ((and (< p gp) (< p (car cursor)))
            (helper (cdr cursor) (+ index 1) p (car cursor) (cons index racc))
          )
          ;; we only find bottoms
          ;((and (> p gp) (> p (car cursor)))
          ;  (helper (cdr cursor) (+ index 1) p (car cursor) (cons index racc))
          ;)
          (else
            (helper (cdr cursor) (+ index 1) p (car cursor) racc)
          )
        )
      )
    )
  )
))

(define find-substr-indicies (with-docstring
  "find every occurence of the substring seq in the string str"
  (lambda (str seq)
    (define seql (string->list seq))
    (define max-index (- (string-length str) 1))
    (let helper ((index 0) (carry 0) (match-cursor seql) (racc '()))
      (if (null? match-cursor)
        (helper index index seql (cons carry racc))
        (if (> index max-index)
          (reverse racc)
          (if (eq? (string-ref str index) (car match-cursor))
            (helper (+ index 1) carry (cdr match-cursor) racc)
            (if (eq? match-cursor seql)
              (helper (+ index 1) (+ index 1) seql racc)
              (helper index index seql racc)
            )
          )
        )
      )
    )
  )
))

(define-public string-merge (with-docstring
  "opposite of string-split. accepts a list of strings <lst> and returns a new string with a <char> between each element of <lst>"
  (lambda (lst char)
    (if (null? lst)
      ""
      (if (null? (cdr lst))
        (car lst)
        (let helper ((cursor (cdr lst)) (s (car lst)))
          (if (null? cursor)
            s
            (helper (cdr cursor) (string-append s (string char) (car cursor)))
          )
        )
      )
    )
  )
))

(define error-preserving-wrap (with-docstring
  "accept three thunks and run them in order. The first thunk is expected to set up some context, the middle thunk is expected to do some work, and the third thunk is expected to clean up this mess. If the middle thunk errors, the error will be caught, the cleanup thunk will be run, and the error will be regenerated. If in is #f, it is skipped."
  (lambda (in thunk out)
    (and in (in))
    (catch #t (lambda ()
      (define retval (thunk))
      (out)
      retval
    ) (lambda args (out) (apply scm-error args)))
  )
))

(define string-head (with-docstring
  "like list-head, but for strings"
  (lambda (string num)
    (substring string 0 num)
  )
))

(define string-tail (with-docstring
  "like list-tail, but for strings"
  (lambda (string num)
    (substring string num)
  )
))

(define decompose-list (with-docstring
  "return a sublist of <lst> for rank <rank> of size <size>"
  (lambda (lst rank size)
  (define ll (length lst))
  (let ()
    (define qu (quotient ll size))
    (define re (remainder ll size))
    (let ()
      (define my-size (+ qu (if (< rank (remainder ll size)) 1 0)))
      (define my-offset (+ (* qu rank) (if (> rank re) re rank)))
      (list-head (list-tail lst my-offset) my-size)
    )
  )
  )
))

(define string-prefix-match (with-docstring
  "does the string <prefix> match exactly match against the beginning of <str>?"
  (lambda (prefix str)
    (define pl (string-length prefix))
    (and (>= (string-length str) pl) (string=? (substring str 0 pl)))
  )
))

(define always-true (with-docstring
  "a function of arbitrary arguments that always returns true. use to avoid the overhead of creating a disposable lambda"
  (lambda args
    #t
  )
))

(define always-false (with-docstring
  "a function of arbitrary arguments that always returns false. use to avoid the overhead of creating a disposable lambda"
  (lambda args
    #f
  )
))

(define identify-scheme-type (with-docstring
  "return a colloquial string describing the underlying Scheme system"
  (lambda ()
    (cond-expand
      (guile (string-append "GNU Guile " (version)))
      (chicken "Chicken Scheme")
      (chez "Chez Scheme")
      (else "some wacky Scheme")
    )
  )
))

(define safe-assoc (with-docstring
  "(safe-assoc <tag> <alist> <backup>) -- search alist <alist> for <tag>. If <tag> Can't be found in <alist>, search <backup>. If <tag> can't be found in <backup>, synthesize a new pair (<tag> . *unspecified*)"
  (lambda (tag alist backup)
    (define result (assoc tag alist))
    (or
      result
      (let ()
        (define result (assoc tag backup))
        (if result
          result
          (cons tag *unspecified*)
        )
      )
    )
  )
))


(load "itree.scm")
