;
; Routines for dealing with tire protocol and tire muxes
;

(library (proverb tire)
  (export
    tire-connect-tcp tire-parse-url make-tire-url-channel make-tire-fport-channel
    make-tire-buffered-fport-channel make-tire-thread-channel
    make-tire-procedure-channel
    tire-mux-setup tire-mux-add-channel! tire-mux-delete-channel!
    tire-mux-get-channels tire-mux-get-buffer-enqueuer tire-mux-set-buffer-enqueuer!
    tire-mux-channels-recurse tire-mux-channels-fold tire-make-request
    tire-request-get-tag tire-request-get-src tire-request-get-dest
    tire-request-get-form tire-request-get-continuation tire-mux-add-request!
    tire-mux-delete-request!  tire-mux-get-requests tire-mux-get-oldest-request-for-tag
    tire-mux-get-newest-request-for-tag tire-mux-get-default-channel
    tire-mux-set-default-channel!  tire-mux-add-listener!
    tire-mux-delete-listener!  tire-mux-get-listeners tire-mux-get-cookie-hash
    tire-mux-set-select-fds!  tire-mux-get-read-fds tire-mux-get-write-fds
    tire-mux-get-exception-fds tire-mux-hash-cookie!  tire-mux-unhash-cookie!
    tire-mux-lookup-cookie tire-mux-enqueue-sync
    tire-mux-enqueue tire-mux-enqueue* tire-mux-setup-buffering!
    tire-mux-buffer-enqueue tire-mux-buffer-enqueue-sync tire-mux-service-channels
    tire-mux-add-tcp-listener!
    
    ; functions that I hope you don't need
    ;tire-mux-flush-until
    tire-mux-select-pre
    tire-mux-select-post
 
    ; I'm ashamed of this
    sleazy-escape
  )
  (import
    ; core Scheme
    (hashtable) (rnrs) (rnrs lists) (rnrs io simple)
    ; proverb specializations
    (proverb doc)
    (proverb help)
    (proverb dialect)
    (proverb channels)
    (proverb threads)
    (proverb cond-expand)
    (proverb unread)
    (proverb debug)
    (proverb html)
    ; non-portable
    (guile networking)
    (only (guile)
      set-object-property!
      call-with-output-string
      string-downcase
      error
      save-stack
      scm-error
      stack-saved?
      fluid-set!
      stack-saved?
      current-thread
      system-async-mark
      ; I/O
      setvbuf
      _IOFBF
      ; system
      getpid
      ; temporary for debugging
      current-thread
    )
  )

  ; this is the value of (current-thread) in the thread currently flushing
  ; the mux. false if no thread is currently flushing the mux.
  (define sleazy-flushing-thread #f)

  ;
  ; there is an article about sleazy-escape
  (define sleazy-escape #f)

(define tire-channel-enqueue-sync (with-wiki-docstring
  "Enqueue value in channel with the [[tire|TIRE mux]] mux. See the article on [[channels | channels]] for more info. The call will block until value is actually transmitted."
  (lambda (channel mux value)
    (define called-back #f)
    (define this (current-thread))
    (channel-enqueue-async channel mux value (lambda ()
      (debugmsg "channel-enqueue-sync -- called back\n")
      (set! called-back #t)
      (if sleazy-flushing-thread
        (begin
          (debugmsg "there is a flushing thread")
          (system-async-mark (lambda () #t) sleazy-flushing-thread)
        )
      )
    ))
    (or called-back (tire-mux-flush-until mux (lambda () called-back)))
  )
))

  (define *unspecified* (if #f #t))

(define tire-connect-tcp (with-docstring
  "create a channel connected to the TIRE mux at <host> on TCP port <port>"
  (lambda (host port)
    (cond-expand
      (guile
        (let ()
          (define s (socket AF_INET SOCK_STREAM 0))
          (connect s AF_INET (car (hostent:addr-list (gethostbyname host))) port)
          (make-tire-buffered-fport-channel s s)
        )
      )
      (chicken
        (let-values (((in out) (tcp-connect host port)))
          (make-tire-buffered-fport-channel in out)
        )
      )
      (else
        (display "TIRE-CONNECT-TCP UNAVAILABLE ON THIS SYSTEM!\n" (current-error-port))
      )
    ) 
  )
))

;; there is a tire-URLs article

(define tire-parse-url (with-wiki-docstring
  "accept an [[Wikipedia:Uniform_Resource_Locator | URL]] and return ( opener host port [[Proverb:cookie|cookie]] ). opener is a thunk that returns a [[Proverb:channel|channel]]."
(lambda (url)
  (define prefix "tire://")
  ;(define popen-prefix "tire+popen://")
  (define cookify (lambda (v)
    (if (string? v)
      (or (string->number v) (string->symbol v))
      v
    )
  ))
  (cond
    ; this at least starts like an URL
    ((and (> (string-length url) (string-length prefix)) (string=? prefix (string-downcase (substring url 0 (string-length prefix))))) (let ()
      (define slashsep (studio-string-split (substring url (string-length prefix)) #\/))
      (let ()
        (define s (studio-string-split (car slashsep) #\:))
        (if (eq? (length s) 2)
          (let ()
            (define host (car s))
            (define port (string->number (cadr s)))
            (list (lambda () (tire-connect-tcp host port)) host port (if (null? (cdr slashsep)) #f (cookify  (cadr slashsep))))
          )
          ; else
          #f
        )
      )
    ))
    ;; a popen URL. WICKEDLY UNSAFE!
    ;((and (> (string-length url) (string-length popen-prefix)) (string=? popen-prefix (string-downcase (substring url 0 (string-length popen-prefix))))) (let ()
    ;  (define slashsep (studio-string-split (substring url (string-length popen-prefix)) #\/))
    ;  (let ()
    ;    (define command (url-decode-string (car slashsep)))
    ;    (list (lambda () (popen-connect command)) (if (null? (cdr slashsep)) #f (cookify (cadr slashsep))))
    ;  )
    ;))
    (else (let ()
      (define s (studio-string-split url #\:))
      (if (eq? (length s) 2)
        (let ()
          (define host (car s))
          (define port (string->number (cadr s)))
          (let ()
            (define c (tire-connect-tcp host port))
            (make-tire-buffered-fport-channel c c)
            (list (lambda () (tire-connect-tcp host port)) host port #f)
          )
        )
        ; else
        #f
      )
    ))
  )
)))

(define make-tire-url-channel (with-docstring
  "accept the TIRE url <url> and return a channel that can reach the named mux and any specified endpoints. This routine creates a channel only and does not update any TIRE mux routing tables."
  (lambda (url)
    (define r (tire-parse-url url))
    (if r
      ((car r))
      #f
    )
  )
))

;
; think about making buffered fport channels
;
(define make-tire-fport-channel (with-docstring
  "make a new TIRE channel backed by Scheme ports <in-port> and <out-port>. Values may be serialized directly to the ports resulting possiby in very many system calls for complicated forms."
  (lambda (in-port out-port)
    (cond-expand
      (guile
        (setvbuf in-port _IOFBF)
        (setvbuf out-port _IOFBF)
        (set-object-property! in-port 'creating-pid (getpid))
        (set-object-property! out-port 'creating-pid (getpid))
        (set-object-property! in-port 'heritage "tire fport channel")
        (set-object-property! out-port 'heritage "tire fport channel")
      )
      (else #t)
    )
    (make-fport-channel
      in-port
      out-port
      receive-value
      unread
      tire-mux-enqueue
    )
  )
))

(define make-tire-buffered-fport-channel (with-docstring
  "make a new TIRE channel backed by Scheme ports <in-port> and <out-port>. Values will be serialized to local buffers prior to transmission. This can reduce the number of system calls required to transmit a form but may cause large string allocations in the tramsmitter for very large forms."
  (lambda (in-port out-port)
    (cond-expand
      (guile
        (setvbuf in-port _IOFBF)
        (set-object-property! in-port 'creating-pid (getpid))
        (set-object-property! out-port 'creating-pid (getpid))
        (set-object-property! in-port 'heritage "tire buffered fport channel")
        (set-object-property! out-port 'heritage "tire buffered fport channel")
      )
      (else #t)
    )
    (make-fport-channel
      in-port
      out-port
      receive-value
      (lambda (v p)
        (catch #t (lambda ()
        (display (call-with-output-string (lambda (sp) (unread v sp))) p)
        (studio-flush-port p)
        ) (lambda args (eof-object)))
      )
      tire-mux-enqueue
    )
  )
))

;
; this does not create a new thread.
;
(define make-tire-thread-channel (with-docstring
  "return a message queue represented by a pair ( enqueue-function . channel). channel will deliver forms enqueued through enqueue-function into a mux. Keep enqueue-function in a seperate thread and shoot forms into a mux with this facility"
  (lambda ()
  ; returns a enqueuer and a channel
  (define sleazy-pipe (studio-pipe))
  (define sleazy-mutex (make-mutex))
  (define out-buffer '())
  (cons
  (lambda (from form . sync?)
    ;(critmsg "TIRE THREAD CHANNEL ENQUEUEING!\n")
    (with-mutex sleazy-mutex (set! out-buffer (append out-buffer (list (cons from form)))))
    (write-char (if (or (null? sync?) (not sync?))
      #\a
      #\b
    ) (cdr sleazy-pipe))
    (studio-flush-port (cdr sleazy-pipe))
  )
  (make-channel
    ; enqueue -- nothing is ever addressed to this channel
    #f
    ; has-queue --
    #f
    ; select-pre
    (lambda (mux)
      (list (list (car sleazy-pipe)) '() '())
    )
    ; select-post
    (lambda (mux rfds wfds xfds)
      (if (member (car sleazy-pipe) rfds)
        (begin
          ;(critmsg "MAKE TIRE THREAD CHANNEL SELECT POST ~a ~a\n" (object-address rfds) rfds)
        (let ()
          (define h (with-mutex sleazy-mutex (let ()
            (define i (car out-buffer))
            (set! out-buffer (cdr out-buffer))
            i
          )))
          ;(critmsg "THREAD CHANNEL FPORT READABLE TOP OF OUT-BUFFER IS ~a!\n" (cdr h))
          (case (read-char (car sleazy-pipe))
            ((#\a)
              ;(critmsg "ENQUEUE NON-SYNC\n")
              (tire-mux-enqueue mux (car h) (cdr h))
            )
            ((#\b)
              ;(critmsg "ENQUEUE SYNC\n")
              (tire-mux-enqueue-sync mux (car h) (cdr h))
            )
            (else (critmsg "UNKNOWN MESSAGE ON THREAD PIPE!\n"))
          )
        )
        )
      )
    )
    ; open
    #f
    ; close
    #f
    ; desc
    "cross-thread signalling"
  )
  )
  )
))

(define make-tire-procedure-channel (with-docstring
  "make a TIRE channel backed by the function proc. proc will be called with each enqueued form. sleazy-escape is set."
  (lambda (proc)
  (define enter-counter 0)
  (define exit-counter 0)
  (make-channel
    ; enqueue
    (lambda (mux form callback)
      (define saved-escape sleazy-escape)
      (define retval
        (begin
          (set! enter-counter (+ enter-counter 1))
          ;support non-local flow control in enqueue procedure
          (call-with-current-continuation
            (lambda (cc)
              (set! sleazy-escape cc)
              (proc form)))))
      (set! exit-counter (+ exit-counter 1))
      (set! sleazy-escape saved-escape)

      ;this really should be called a 'callback'
      ;  elsewhere it is called 'continuation', 
      ;  but it is not allowed to be a real continuation 
      ;  (i.e. callback must return here)
      ;  XXX: is it ok for callbacks to be called multiple times?
      (if callback (callback))
      
      ; your return value is irrelevant
    )
    ; has-queue
    ; forms are dispatched immediately... no queue
    #f
    ; select-pre
    (lambda (mux)
      (list '() '() '())
    )
    ; select-post
    #f
    ; open
    #f
    ; close
    #f
    ; desc
    "procedure backed"
  )
  )
))

(define tire-mux-setup (with-docstring
  "create a new TIRE mux. Args is ignored."
  (lambda (args)
    ; each request is of the form (tag . ( port . continuation ) )
    ; ports, outstanding-tire-requests, outstanding-tire-responses
    ; last-cookie, last-tag, last-command
    (define mux (vector
      '()    ; channels
      '()    ; requests
      #f     ; default channel
      '()    ; listening ports
      (make-eq-hashtable)
      #f     ; select fds
      #f     ; add-channel-mutex
      #f     ; enqueue-mutex
      #f     ; buffer enqueuer
    ))
    (define mux-channel (make-tire-procedure-channel (lambda (form . rest)
      (if (and (list? form) (not (null? form)))
        (case (car form)
          ((t)
            (if (>= (length form) 4)
              (case (cadddr form)
                ((help) (let ()
                  (tire-mux-enqueue mux mux-channel (list 'r (cadr form) '(help enum)))
                ))
                ((enum) (let ()
                  (define h (tire-mux-get-cookie-hash mux))
                  (tire-mux-enqueue mux mux-channel (list 'r (cadr form)
                                                             (remove 'local (remove '- (vector->list (hashtable-keys h))))))

                ))
                (else (tire-mux-enqueue mux mux-channel (list 'r (cadr form) 'no-such-command)))
              )
              (critmsg "mux root cookie: t form is too short")
            )
          )
          (else
            (critmsg "mux root cookie: non-t form\n")
          )
        )
        (critmsg "mux root cookie: form is not a list\n")
      )
    )))
    (tire-mux-add-channel! mux mux-channel)
    (tire-mux-hash-cookie! mux '- mux-channel)

    mux
  )
))

(define tire-mux-add-channel! (with-wiki-docstring
  "(tire-mux-add-channel! mux channel) -- add [[Proverb:channel|channel]] to the mux. This allows the mux to
incorporate the channel into its select operations."
  (lambda (mux channel)
    (channel-open channel)
    (with-maybe-mutex (vector-ref mux 6)
      (if (not (memq channel (vector-ref mux 0)))
        (vector-set! mux 0 (cons channel (vector-ref mux 0)))
      )
    )
    *unspecified*
  )
))

(define tire-mux-delete-channel! (with-wiki-docstring
  "(tire-mux-delete-channel! mux channel) -- remove [[Proverb:channel||channel]] from the mux. Unroutes all cookies currently routed
through channel. If channel is currently the default channel, the mux default channel will be unset."
  (lambda (mux channel)
    (define h (tire-mux-get-cookie-hash mux))
    (channel-close channel)
    (with-maybe-mutex (vector-ref mux 6)
      (vector-set! mux 0 (remove channel (vector-ref mux 0)))
      (let ()
        (define all-cookies (vector->list (hashtable-keys h)))
        (let ()
          (define offending-cookies (remove #f (map (lambda (k v)
            (if (eq? v channel)
              k
              #f
            )
          ) all-cookies (map (lambda (k) (hashtable-ref h k #f)) all-cookies))))
          (for-each (lambda (v)
            (tire-mux-unhash-cookie! mux v)
          ) offending-cookies)
        )
      )
      ; unset the default if this was the default
      (and
        (eq? (tire-mux-get-default-channel mux) channel)
        (tire-mux-set-default-channel! mux #f)
      )
    )
    *unspecified*
  )
))

(define tire-mux-get-channels (with-wiki-docstring
  "(tire-mux-get-channels mux) -- return the list of channels held by the mux."
  (lambda (mux)
    (with-maybe-mutex (vector-ref mux 6) (vector-ref mux 0))
  )
))

(define tire-mux-get-buffer-enqueuer (with-docstring
  "(tire-mux-get-buffer-enqueuer mux) -- return the procedure for buffer enqueueing forms in the mux."
  (lambda (mux)
    (vector-ref mux 8)
  )
))

(define tire-mux-set-buffer-enqueuer! (with-docstring
  "(tire-mux-set-buffer-enqueuer! mux enqueuer) -- assign the procedure for buffer enqueueing forms in the mux."
  (lambda (mux enqueuer)
    (vector-set! mux 8 enqueuer)
  )
))

(define tire-mux-channels-recurse (with-docstring
  "a continuation-passing-style fold for the channels in TIRE mux <mux>. This routine will visit every channel present in the mux at the time of the call. Channels deleted during traversal will be skipped. <proc> is called as (<proc> channel cps-continuation)"
  (lambda (proc init mux)
  (define i (lambda (v) v))
  (define find-next-valid-channels (lambda (current old)
    (if (null? current)
      '() 
      (let helper ((cursor old) (delivery cdr))
        (if (null? cursor)
          '() 
          (let ()
            (define m (memq (car cursor) current))
            (if m
              (delivery m)
              (helper (cdr cursor) i)
            )
          )
        )
      )
    )
  ))
  (let helper ((h (tire-mux-get-channels mux)) (l init))
    (if (null? h)
      l
      (proc (car h) l (lambda (retval)
        (define nextchans (find-next-valid-channels (tire-mux-get-channels mux) h))
        (helper nextchans retval)
      ))
    )
  )
  )
))

(define tire-mux-channels-fold (with-docstring
  "fold the procedure <proc> over the channels in the TIRE mux <mux> with initial value <init>. (tire-mux-channels-fold cons '() mux) would return a list of the channels in mux."
  (lambda (proc init mux)
    (tire-mux-channels-recurse (lambda (val last next)
      (next (proc val last))
    ) init mux)
  )
))

(define tire-make-request (with-docstring
  "make a TIRE request value from the parameters"
  (lambda (tag src-channel dest-channel original-form continuation)
  (cons
    tag
    (vector
      src-channel
      dest-channel
      original-form
      continuation
    )
  )
  )
))

(define tire-request-get-tag (with-docstring
  "extract the tag from a TIRE request"
  (lambda (request)
    (car request)
  )
))

(define tire-request-get-src (with-docstring
  "extract the source channel from a TIRE request"
  (lambda (request)
    (vector-ref (cdr request) 0)
  )
))

(define tire-request-get-dest (with-docstring
  "extract the destination channel from a TIRE request"
  (lambda (request)
    (vector-ref (cdr request) 1)
  )
))

(define tire-request-get-form (with-docstring
  "extract the TIRE form from a TIRE request"
  (lambda (request)
    (vector-ref (cdr request) 2)
  )
))

(define tire-request-get-continuation (with-docstring
  "extract the continuation from a TIRE request"
  (lambda (request)
    (vector-ref (cdr request) 3)
  )
))

(define tire-mux-add-request! (with-docstring
  "<mux> is the tire mux and <request> is a tire-request"
(lambda (mux request)
  (vector-set! mux 1 (cons request (vector-ref mux 1)))
  *unspecified*
  )
))

(define tire-mux-delete-request! (with-docstring
  "delete the TIRE request <request> from the TIRE mux <mux>"
  (lambda (mux request)
    ;(critmsg "TIRE-MUX DELETE REQUEST: TAG=~a\n" (car request))
    (vector-set! mux 1 (remove request (vector-ref mux 1)))
    *unspecified*
  )
))

(define tire-mux-get-requests (with-docstring
  "return the list of outstanding TIRE requests in the TIRE mux <mux>"
  (lambda (mux)
    (vector-ref mux 1)
  )
))

(define tire-mux-get-oldest-request-for-tag (with-docstring
  "return the oldest outstanding TIRE request for the tag <tag> in the TIRE mux <mux>"
  (lambda (mux tag)
    (assoc tag (reverse (tire-mux-get-requests mux)))
  )
))
 
(define tire-mux-get-newest-request-for-tag (with-docstring
  "return the newest outstanding TIRE request for the tag <tag> in the TIRE mux <mux>"
  (lambda (mux tag)
    (assoc tag (tire-mux-get-requests mux))
  )
))

(define tire-mux-get-default-channel (with-docstring
  "return the default channel for the TIRE mux <mux>"
  (lambda (mux)
    (vector-ref mux 2)
  )
))

(define tire-mux-set-default-channel! (with-docstring
  "set the default channel for the TIRE mux <mux> to <channel>. If channel is #f, the mux operates without a default channel."
  (lambda (mux channel)
    (vector-set! mux 2 channel)
  )
))

(define tire-mux-add-listener! (with-docstring
  "add a listening socket <listener> to the TIRE mux <mux>"
  (lambda (mux listener)
    (vector-set! mux 3 (cons listener (vector-ref mux 3)))
    *unspecified*
  )
))

(define tire-mux-delete-listener! (with-docstring
  "remove the listening socket <listener> from the TIRE mux <mux>"
  (lambda (mux listener)
    (vector-set! mux 3 (remove listener (vector-ref mux 3)))
    *unspecified*
  )
))

(define tire-mux-get-listeners (with-docstring
  "return the list of listening sockets for the TIRE mux <mux>"
  (lambda (mux)
    (vector-ref mux 3)
  )
))

(define tire-mux-get-cookie-hash (with-docstring
  "return the hash table mapping cookies to channels for the TIRE mux <mux>"
  (lambda (mux)
    (vector-ref mux 4)
  )
))

(define tire-mux-set-select-fds! (with-docstring
  "set the descriptors for select in the TIRE mux <mux>. This whole thing sounds wrong."
  (lambda (mux fds)
    (vector-set! mux 5 fds)
  )
))

(define tire-mux-get-read-fds (with-docstring
  "return the list of read fds used for select in the TIRE mux <mux>. This whole thing sounds wrong."
  (lambda (mux)
    (define fds (vector-ref mux 5))
    (if fds (car fds) '())
  )
))

(define tire-mux-get-write-fds (with-docstring
  "return the list of write fds used for select in the TIRE mux <mux>. This whole thing sounds wrong."
  (lambda (mux)
    (define fds (vector-ref mux 5))
    (if fds (cadr fds) '())
  )
))

(define tire-mux-get-exception-fds (with-docstring
  "return the list of exception fds used for select in the TIRE mux <mux>. This whole thing sounds wrong."
  (lambda (mux)
    (define fds (vector-ref mux 5))
    (if fds (caddr fds) '())
  )
))

(define tire-mux-hash-cookie! (with-docstring
  "associate the cookie <cookie> with the TIRE server on <channel>."
  (lambda (mux cookie channel)
    (hashtable-set! (vector-ref mux 4) cookie channel)
  )
))

(define tire-mux-unhash-cookie! (with-docstring
  "disassociate the cookie <cookie> with any channel."
  (lambda (mux cookie)
    (hashtable-delete (vector-ref mux 4) cookie)
  )
))

(define tire-mux-lookup-cookie (with-docstring
  "find the channel associated with cookie. return false if not found"
  (lambda (mux cookie)
  (or
    (hashtable-ref (tire-mux-get-cookie-hash mux) cookie #f)
    (tire-mux-get-default-channel mux)
  )
  )
))

(define tire-mux-select-pre (with-docstring
  "the run-select-loop select-pre function for TIRE muxes"
  (lambda (mux rfds wfds xfds)
  (let helper ((cursor (tire-mux-get-channels mux)) (r (append (tire-mux-get-listeners mux) rfds)) (w wfds) (x xfds))
    (if (null? cursor)
      (list r w x)
      (let ((sub (channel-select-pre (car cursor) mux)))
        (if sub
          (apply helper (cons (cdr cursor) (map append sub (list r w x))))
          (begin
            (tire-mux-delete-channel! mux (car cursor))
            (apply helper (cons (cdr cursor) (list r w x)))
          )
        )
      )
    )
  )
  )
))

(define tire-mux-enqueue-sync (with-wiki-docstring
  "(tire-mux-enqueue-sync MUX FROM FORM) -- Synchronous form of [[Scheme:tire-mux-enqueue*|tire-mux-enqueue*]]"
  (lambda (mux from form)
    (tire-mux-enqueue* mux from form #t)
  )
))

(define tire-mux-enqueue (with-wiki-docstring
  "(tire-mux-enqueue MUX FROM FORM) -- Asynchronous form of [[Scheme:tire-mux-enqueue*|tire-mux-enqueue*]]"
  (lambda (mux from form)
    (tire-mux-enqueue* mux from form #f)
  )
))

(define tire-mux-enqueue*
  (lambda (mux from form sync?)
    (cond
      ((and (list? form) (> (length form) 3) (eq? (car form) 't))
        ;(critmsg "TIRE-MUX-ENQUEUE: T! ~a\n" form)
        (let ()
          (define chan (tire-mux-lookup-cookie mux (list-ref form 2)))
          (if chan
            (begin
              (tire-mux-add-request! mux (tire-make-request (list-ref form 1) from chan form #f))
              (channel-enqueue chan mux form)
            )
            (begin
              (critmsg "tire-mux-enqueue* cannot enqueue form ~a to non-existant cookie ~a\n" form (list-ref form 2))
              (channel-enqueue from mux `(e ,(list-ref form 1) no-such-cookie ,(list-ref form 2)))
            )
          )
        )
      )
      ((and (list? form) (> (length form) 2) (eq? (car form) 'r))
        ;(critmsg "TIRE-MUX-ENQUEUE: R!\n")
        (let ((p (tire-mux-get-oldest-request-for-tag mux (cadr form))))
          ; (tag . ( channel . continuation ) )
          (if p
            (let ((channel (tire-request-get-src p)) (continuation (tire-request-get-continuation p)))
              ; remove the pending request
              (tire-mux-delete-request! mux p)
              (let ()
                ; write the result
                (define newform (and channel (if (> (length form) 2)
                  form
                  `(e ! tire-error ,form)
                )))
                (debugmsg "MUX-ENQUEUE: CHANNEL ENQUEUEING AN R MESSAGE: sync=~a continuation=~a\n" sync? continuation)
                (if sync?
                    (tire-channel-enqueue-sync channel mux newform)
                    (channel-enqueue channel mux newform)
                )
                (if continuation (continuation))
              )
            )
            ; else
            (begin 
              (debugmsg "tire r action for tag ~a -- no such outstanding tag\n" (list-ref form 1))
              (channel-enqueue from mux `(e ,(list-ref form 1) no-such-outstanding-tag))
            )
          )
        )
      )
      ((and (list? form) (> (length form) 1) (eq? (car form) 'i))
        (let ((p (tire-mux-get-oldest-request-for-tag mux (cadr form))))
          (if p
            (let ((channel (tire-request-get-src p)))
              ; write the result
              (and channel (if (> (length form) 2)
                (begin
                  ; snooping forms
                  (if (> (length form) 3)
                    (case (list-ref form 2)
                      ((route) (let ((cookies (list-tail form 3)))
                        (debugmsg "encountered a routing update: cookies ~a should be forwarded to channel ~a\n" cookies from)
                        (for-each (lambda (v)
                          (if (not (tire-mux-lookup-cookie mux v))
                            (tire-mux-hash-cookie! mux v from)
                          )
                        ) cookies)
                      ))
                      ((unroute) (let ((cookies (list-tail form 3)))
                        (debugmsg "encountered an unrouting update: cookies ~a should be deleted\n" cookies from)
                        (for-each (lambda (v)
                          (and (tire-mux-lookup-cookie mux v)
                               (tire-mux-unhash-cookie! mux v)
                          )
                        ) cookies)
                      ))
                    )
                    ; else this is not an I message that we snoop
                  )
                  (debugmsg "MUX-ENQUEUE: CHANNEL ENQUEUEING AN I MESSAGE: form=~a\n" form)
                  (if sync?
                    (tire-channel-enqueue-sync channel mux form)
                    (channel-enqueue channel mux form)
                  )
                )
                (critmsg "MUX-ENQUEUE: PASSED A TOO-SHORT I FORM!\n")
              ))
            )
            ; else
            (begin
              (channel-enqueue from mux `(e ! no-such-outstanding-tag ,(list-ref form 1)))
              (critmsg "received i message with tag ~a. No such outstanding request in mux (message=~a)\n" (cadr form) form)
            )
          )
        )
      )
      ((and (list? form) (> (length form) 1) (eq? (car form) 's)) (let ()
        (define tag (cadr form))
        (let stop-helper ((p (tire-mux-get-oldest-request-for-tag mux tag)))
          (if p
            (let ((d (tire-request-get-dest p))
                  (c (tire-request-get-continuation p)))
              (if d 
                (begin
                  (debugmsg "tire cancellation: got an s message that matches an outstanding tag with a destination")
                  (channel-enqueue d mux form)
                )
                (critmsg "tire cancellation: got an s message that matches an outstanding tag but has no destination")
              )
              (if c 
                (debugmsg "YOWZA! STOP MESSAGE FOR EXISTING TAG WITH AN EXISTING CONT!\n")
                (critmsg "YOWZA! STOP MESSAGE FOR EXISTING TAG WITHOUT AN EXISTING CONT!\n")
              )
              ;(critmsg "THIS COOKIE EXISTS!\n")
              ;(tire-mux-delete-request! mux p)
              ;(stop-helper (tire-mux-get-oldest-request-for-tag mux tag))
            )
            ; else take no action. we have no such tag in flight.
            ; we could decide that every cancel generates an 'e' -- but I'm not
            ; sure
            *unspecified*
          )
        )
      ))
      (else (critmsg "tire-mux-enqueue*: misunderstood tire form: ~a" form))
    )
  )
)

(define tire-mux-setup-buffering! (with-docstring
  "(tire-mux-setup-buffering! mux) -- a one-time operation on a mux to enable buffered queueing. This is necessary when forms are being enqueued from non-mux threads."
  (lambda (mux)
    (define t (make-tire-thread-channel))
    (vector-set! mux 6 (make-mutex))
    (tire-mux-add-channel! mux (cdr t))
    (tire-mux-set-buffer-enqueuer! mux (car t))
  )
))

;
; This call won't change the target of a current select call in the mux
;
(define tire-mux-buffer-enqueue (with-docstring
  "enqueue a TIRE form <form> in the TIRE mux <mux> with a return path of the channel <from>. This form may return before <form> is transmitted."
  (lambda (mux from form)
    (if (tire-mux-get-buffer-enqueuer mux)
      ((tire-mux-get-buffer-enqueuer mux) from form)
      ; enable buffering and try again.
      (begin
        (tire-mux-setup-buffering! mux)
        (tire-mux-buffer-enqueue mux from form)
      )
    )
  )
))

(define tire-mux-buffer-enqueue-sync (with-docstring
  "enqueue a TIRE form <form> in the TIRE mux <mux> with a return path of the channel <from>. This form will block until <form> is transmitted."
  (lambda (mux from form)
  (if (tire-mux-get-buffer-enqueuer mux)
    ((tire-mux-get-buffer-enqueuer mux) from form #t)
    ; else enable buffering and try again
    (begin
      (tire-mux-setup-buffering! mux)
      (tire-mux-buffer-enqueue-sync mux from form)
    )
  )
  )
))

(define tire-mux-service-channels (with-docstring
  "run the run-select-loop select-post routine for every channel in the TIRE mux <mux> that participates in select multiplexing"
  (once (lambda (mux starting-done)
    (define retval (tire-mux-channels-recurse (lambda (val done next)
      ; just pull this bus over if we return multiple times
      (define trigger-pulled #f)
      (if (channel-does-select? val)
        (let ()
          (define sub 
            (channel-select-post val mux
                        (tire-mux-get-read-fds mux)
                        (tire-mux-get-write-fds mux)
                        (tire-mux-get-exception-fds mux)))
          (if trigger-pulled
            (begin
              (critmsg "ERROR: MULTIPLE RETURN FROM CHANNEL SELECT POST:\n")
              mux
            )
            ; else
            (begin
              (set! trigger-pulled #t)

              (if sub
                (next #f)
                (begin
                  (tire-mux-delete-channel! mux val)
                  (next done)
                )
              )
            )
          )
        )
        ; else
        (next done)
      )
    ) starting-done mux))
    (if retval
      (begin
        (debugmsg "THIS TIRE MUX WILL SELECT POST NO MORE!\n")
        #f
      )
      mux
    )
  ))
))

(define tire-mux-select-post (with-docstring
  "the run-select-loop select-post function for TIRE muxes"
  (lambda (mux rfds wfds xfds)
  ; order of battle
  ;   first accept new connections
  ;   then service the channels
  ; the hope is that this will favor accepting
  ; a connection that will contain some type of
  ; cancel message.
  (define listeners (tire-mux-get-listeners mux))
  (tire-mux-set-select-fds! mux (list rfds wfds xfds))

  (if (null? (remove #f (map (lambda (v) (member v rfds)) listeners)))
    ; there are no new connections to accept
    (tire-mux-service-channels mux (null? listeners))
    ; else there are connections to accept
    (let helper ((cursor (tire-mux-get-read-fds mux)))
      (if (null? cursor)
        mux
        (if (member (car cursor) listeners)
          (let ()
            (define new (accept-io-socket (car cursor)))
            (tire-mux-add-channel! mux (make-tire-buffered-fport-channel (car new) (cdr new)))
            mux 
          )
          (helper (cdr cursor))
        )
      )
    )
  )
  )
))

(define tire-mux-flush-until (with-docstring
  "run the TIRE mux <mux> until the thunk <pred> returns non-false. Although this form can be used to flush the mux, it runs the entire mux and will allow new forms to be enqueued. returns first non-false value of <pred>."
  (lambda (mux pred)
    (let helper ()
      (or (pred)
          (begin
            (set! sleazy-flushing-thread (current-thread))
            (let ()
              (define pre (tire-mux-select-pre mux '() '() '()))
              (define beenhere 1)
                (if (not (eq? beenhere 1))
                  (critmsg "BOGUS CONTINUATION FOO!\n")
                )
                (let ()
                  (define sval (catch 'system-error (lambda ()
                    (apply studio-select pre)
                  ) (lambda args (debugmsg "interrupted select") (list '() '() '()))))
                  (set! beenhere (+ beenhere 1))
                  (apply tire-mux-select-post mux sval)
                  (or (let ((v (pred))) (if v (set! sleazy-flushing-thread #f)) v)  (helper))
                )
              ) 
            )
      )
    )
  )
))

(define tire-mux-add-tcp-listener! (with-docstring
  "add a new TCP listener to the TIRE mux <mux> on TCP port <portno>. Returns the port number actually bound"
  (lambda (mux portno)
  (let ()
    (cond-expand
      (guile
        (let ()
          (define tire-socket (socket AF_INET SOCK_STREAM 0))
          (bind tire-socket AF_INET INADDR_ANY (or portno 0))
          (listen tire-socket 5)
          (tire-mux-add-listener! mux tire-socket)
          (vector-ref (getsockname tire-socket) 2)
        )
      )
      (chicken
        (let ()
          (define tire-listener (tcp-listen (or portno 0) 5))
          (define tire-socket (tcp-listener-fileno tire-listener))
          (setprop! tire-socket 'tcp-listener tire-listener)
          (tire-mux-add-listener! mux tire-socket)
          (tcp-listener-port (getprop tire-socket 'tcp-listener))
          (generate-local-cookie)
          (tcp-listener-port (getprop tire-socket 'tcp-listener))
        )

      )
      (else
        (display "TIRE-MUX-ADD-TCP-LISTENER! UNAVAILABLE!\n")
      )
    )
  )
  )
))

;; there is a tire help article

)
