;
; Support routines for URIs URLs and URNs
;

(library (proverb uri)
  (export char->url-hex url-encode-char url-encode-string url-decode-string make-data-url)
  (import
    (rnrs)
    (rnrs arithmetic bitwise)
    (proverb doc)
    (proverb mime)
    (proverb dialect)
    (proverb types)
  )

(define char->url-hex (with-docstring
  "convert character char into the equivalent url-encoded hex string"
  (lambda-w-types ((char char?))
    (define w (char->integer char))
    (string-append 
      "%"
      (case (quotient w 16)
        ((#x0) "0") ((#x1) "1") ((#x2) "2") ((#x3) "3")
        ((#x4) "4") ((#x5) "5") ((#x6) "6") ((#x7) "7")
        ((#x8) "8") ((#x9) "9") ((#xa) "a") ((#xb) "b")
        ((#xc) "c") ((#xd) "d") ((#xe) "e") ((#xf) "f")
      )
      (case (bitwise-and w #xf)
        ((#x0) "0") ((#x1) "1") ((#x2) "2") ((#x3) "3")
        ((#x4) "4") ((#x5) "5") ((#x6) "6") ((#x7) "7")
        ((#x8) "8") ((#x9) "9") ((#xa) "a") ((#xb) "b")
        ((#xc) "c") ((#xd) "d") ((#xe) "e") ((#xf) "f")
      )
    )
  )
))

(define url-encode-char (with-docstring
  "accept a Scheme character <char> and return a string with the URL-encoded value of <char>."
  (lambda-w-types ((char char?))
    (case char
      ((#\$ #\& #\+ #\, #\/ #\: #\; #\= #\? #\@) (char->url-hex char))
      ((#\" #\< #\> #\# #\% #\{ #\} #\| #\\ #\^ #\~ #\[ #\] #\` #\space)
        (char->url-hex char)
      )
      ((#\( #\)) (char->url-hex char)) ; these irritate CSS
      (else
        (cond
          ((>= (char->integer char) 127)
            (char->url-hex char)
          )
          ((< (char->integer char) 32)
            (char->url-hex char)
          )
          (else (string char))
        )
      )
    )
  )
))

(define url-encode-string (with-docstring
  "accept a Scheme string <str> and return a transformed copy of <s> with every character requiring URL encoding replaced with its URL encoded equivalent."
  (lambda-w-types ((str string?))
    (apply string-append (map url-encode-char (string->list str)))
  )
))

(define url-decode-string (with-docstring
  "accept a Scheme string <str> and return a transformed copy of <str> with every URL-encoded substring replaced by its native equivalent."
  (lambda-w-types ((str string?))
    (let helper ((cursor (string->list str)) (racc '()))
      (if (null? cursor)
        (apply string (reverse racc))
        (cond
          ((eq? (car cursor) #\%)
            (helper (cdddr cursor) (cons (integer->char (string->number (string-append "#x" (apply string (list-head (cdr cursor) 2))))) racc))
          )
          ((eq? (car cursor) #\+)
            (helper (cdr cursor) (cons #\space racc))
          )
          (else
            (helper (cdr cursor) (cons (car cursor) racc))
          )   
        )
      )
    )
  )
))

(define make-data-url (with-docstring
  "accept a data resource as a string and return valid data-url. Scheme will attempt to guess a mime type if mime-type is false. rest is a list of additional data URL attributes, like charset"
  (lambda (data-string mime-type . rest)
    (if mime-type
      (string-append
        "data:"
        mime-type
        (string-merge (let helper ((cursor rest) (racc '()))
          (if (null? cursor)
            (if (null? racc)
              racc
              (cons "" (reverse racc)) ; to get us a leading comma
            )
            (if (null? (cdr cursor))
              (reverse racc)
              (helper (cddr cursor) (cons (string-append (symbol->string (car cursor)) "=" (cadr cursor)) racc))
            )
          )
        ) #\;)
        ","
        (url-encode-string data-string)
      )
      (make-data-url data-string (data-guess-mime-type data-string))
    )
  )
))

)
