(library (proverb wiki)
  (export
    make-wiki
    get-raw-scheme-article
  )
  (import
    (rnrs)
    (proverb doc)
    (proverb xml)
    (proverb help)
    (proverb cond-expand)
    (proverb dialect)
    (proverb debug)
    (ice-9 documentation)
    (ice-9 pretty-print)
    (ice-9 format)
    (proverb guilefoo)
    (only (guile)
      defined? procedure-source closure?
      local-eval with-input-from-string
      call-with-output-string
      current-module
      procedure-environment
      environment-module
      module-name
      resolve-interface
      resolve-module
      sort
      variable-ref
      macro?
      primitive-macro?
    )
  )

(define make-wiki (with-docstring
  "accept a Scheme string <string> in Proverb's modified MediaWiki syntax and pass it through to the Proverb client-side MediaWiki parser. <string> is HTML escaped."
  (lambda (string)
    (with-xml-tag "<wiki>" (make-raw-text (xml-body-escape-string string)))
  )
))

    (define get-guile-article (with-docstring
      "get the text of a Guile article."
      (lambda (name)
        (define t (search-documentation-files name))
        (if t
          (apply string-append (map (lambda (v)
            (string-append "&&" v "\n")
          ) (studio-string-split t #\newline)))
          #f
        )
      )
    ))

(define get-raw-scheme-article (let ()
  (define wikify-value (lambda (v)
    (call-with-output-string (lambda (p) (pretty-print v p)))
  ))
  (lambda (name)
    (cond
      ((and (symbol? name) (defined? name)) (let ()
        (define val (eval name (current-module)))
        (if (closure? val)
          (wikify-value (procedure-source val))
          #f
        )
      ))
      (else #f)
    )
  )
))

(define get-scheme-module-article (lambda (m)
  (let ((l (module-get-symbol-list m)))
    (apply string-append
      (format #f "== ~a ==\n" (module-name m))
      (map (lambda (v)
       (define val (variable-ref (cdr v)))
       (cond
         ((procedure? val) (format #f "* [[Scheme:~a|~a]] (procedure)\n" (car v) (car v)))
         ((macro? val) (format #f "* [[Scheme:~a|~a]] (macro)\n" (car v) (car v)))
         (else (format #f "* [[Scheme:~a|~a]]\n" (car v) (car v)))
       )
      ) (sort (map (lambda (v) (cons (symbol->string (car v)) (cdr v))) l) (lambda (a b) (string>? (car b) (car a)))))
    )
  )
))

(define get-scheme-article (let ()
  (with-docstring
    "return the body of an article from the 'Scheme' wiki namespace. This is a synthetic namespace backed by the top-level evaluation environment."
    (lambda (name)
      (let ()
        (define abject-failure (cons #f #f))
        (cond
          ((symbol? name) (let ()
            (define v (get-raw-scheme-article name))
            (if v
              (apply string-append (map (lambda (v)
                (string-append "%%" v "\n")
              ) (studio-string-split v #\newline)))
              (let ((value (catch #t (lambda () (eval name (current-module))) (lambda args abject-failure))))
                (define strname (symbol->string name))
                (cond
                  ((eq? value abject-failure)
                    (string-append "%%" strname " does not exist.")
                  )
                  ((and (procedure? value) (not (closure? value)))
                    (string-append "''" strname "'' is a primitive procedure. Try [[Guile:" strname "]].")
                  )
                  ((list? value)                  
                    (string-append "%%" strname " is a list.")
                  )
                  ((vector? value)                  
                    (string-append "%%" strname " is a vector.")
                  )
                  ((pair? value)                  
                    (string-append "%%" strname " is a pair.")
                  )
                  ((macro? value)
                    (cond
                      ((primitive-macro? value) (string-append "%%" strname " is part of the syntax of the language"))
                      (else (string-append "%%" strname " is a macro"))
                    )
                  )
                  (else (string-append "%%" strname " exists, but I don't know how to show it to you as a wiki article ."))
                )
              )
            )
          ))
          ((list? name) (let ()
            (if (null? name)
              (let ()
                (define m (current-module))
                (if m
                  (get-scheme-module-article m)
                  #f
                )
              )
              (let ()
                (define m (resolve-interface name))
                (if m
                  (get-scheme-module-article m)
                  #f
                )
              )
            )
          ))
          (else #f)
        )
      )
    )
  )
))

(define put-scheme-article (with-docstring
  "put the article article-name back in the Scheme help article namespace. The Scheme help article namespace does this by evaluating body as Scheme and binding the result to the label article-name."
  (lambda (article-name body)
    (define r (with-input-from-string body read))
    (if r
      (if (defined? article-name)
        (let ()
          (define v (eval article-name (current-module)))
          (if (procedure? v)
            (let ()
              (define e (procedure-environment v))
              (if e
                (let ()
                  (define m (environment-module e))
                  (if m
                    (eval `(set! ,article-name ,(local-eval r e)) m)
                    #f
                  )
                )
                (begin (critmsg "scheme article procedure has no environment") #f)
              )
            )
            (begin (critmsg "scheme article is not a procedure") #f)
          )
        )
        (begin (critmsg "scheme article is not bound") #f)
      )
      (begin (critmsg "scheme article text could not be read") #f)
    )
  )
))

    (let ()
      (define guiledoc-wiki-ns `(
        (get . ,get-guile-article)
        (search . ,(lambda (query-string)
          (define s (string->symbol query-string))
          (let ()
            (define r (search-documentation-files s))
            (if r
              (list (string->symbol query-string))
              '()
            )
          )
        ))
      ))
      (register-wiki-namespace "Guile" guiledoc-wiki-ns)
    )
    (let ()
      (define scheme-wiki-ns `(
        (get . ,get-scheme-article)
        (put . ,put-scheme-article)
      ))
      (register-wiki-namespace "Scheme" scheme-wiki-ns)
    )


)
