
; routines for XML<->scheme interactions
;
(library (proverb xml)
  (export
    make-raw
    xmlize xmlize-value
    with-xml-tag make-xml-tag
    make-raw-text make-text make-xhtml make-shortcut-anchor
    xml-escape-string xml-quote-escape-string xml-body-escape-string
    ; our passvals
    xml-html-fragment xml-xhtml xml-raw xml-text
  )
  (import
    (rnrs)
    (proverb doc) (proverb dialect)
    (proverb passval)
    (proverb uri)
    (proverb types)
    (guile errors)
    (only (guile)
      hash-fold
      object-property
      object->string
      hash-table?
      record?
      unspecified?
      struct?
      record-type?
    )
  )

(define xml-escape-char (with-docstring
  "accept a character and return the string containing an XML-safe form"
  (lambda-w-types ((char char?))
    (case char
      ((#\<) "&lt;")
      ((#\>) "&gt;")
      ((#\&) "&amp;")
      ((#\") "&quot;")
      ((#\') "&apos;")
      ; this may not be quite the right thing
      ((#\nul) "")
      (else (string char))
    )
  )
))

(define xml-escape-string (with-docstring
  "accept a string and return the string containing an HTML-safe form"
  (lambda-w-types ((str string?))
    ; chicken may prefer string-translate* here -- apply may fail with
    ; > 120 parameters. See chicken README.
    (apply string-append (map xml-escape-char (string->list str)))
  )
))

(define xml-body-escape-string (with-docstring
  "accept a string and return the string containing a form safe for the body of an XML tag. differs from xml-escape-string by inserting <null/> in place of the null character."
  (lambda-w-types ((str string?))
    (define n (integer->char 0))
    ; chicken may prefer string-translate* here -- apply may fail with
    ; > 120 parameters. See chicken README.
    (apply string-append (map (lambda (v)
      (define i (char->integer v))  
      (case i
        ((#x00 #x01 #x02 #x03 #x04 #x05 #x06 #x07
          #x08 #x09      #x0b #x0c      #x0e #x0f
          #x10 #x11 #x12 #x13 #x14 #x15 #x16 #x17
          #x18 #x19 #x1a #x1b #x1c #x1d #x1e #x1f
                                             #x7f
          #x80 #x81 #x82 #x83 #x84 #x85 #x86 #x87
          #x88 #x89 #x8a #x8b #x8c #x8d #x8e #x8f
          #x90 #x91 #x92 #x93 #x94 #x95 #x96 #x97
          #x98 #x99 #x9a #x9b #x9c #x9d #x9e #x9f
          #xa0 #xa1 #xa2 #xa3 #xa4 #xa5 #xa6 #xa7
          #xa8 #xa9 #xaa #xab #xac #xad #xae #xaf
          #xb0 #xb1 #xb2 #xb3 #xb4 #xb5 #xb6 #xb7
          #xb8 #xb9 #xba #xbb #xbc #xbd #xbe #xbf
          #xc0 #xc1 #xc2 #xc3 #xc4 #xc5 #xc6 #xc7
          #xc8 #xc9 #xca #xcb #xcc #xcd #xce #xcf
          #xd0 #xd1 #xd2 #xd3 #xd4 #xd5 #xd6 #xd7
          #xd8 #xd9 #xda #xdb #xdc #xdd #xde #xdf
          #xe0 #xe1 #xe2 #xe3 #xe4 #xe5 #xe6 #xe7
          #xe8 #xe9 #xea #xeb #xec #xed #xee #xef
          #xf0 #xf1 #xf2 #xf3 #xf4 #xf5 #xf6 #xf7
          #xf8 #xf9 #xfa #xfb #xfc #xfd #xfe #xff) (string-append "<char code='" (number->string i) "' />"))
        (else (xml-escape-char v))
      )
    ) (string->list str)))
  )
))

(define xml-quote-escape-string (with-docstring
  "accept a Scheme string <str> and escape any double quotes to the HTML quot entity"
  (lambda-w-types ((str string?))
    (let helper ((cursor (string->list str)) (racc '()))
      (if (null? cursor)
        (apply string (reverse racc))
        (cond
          ((eq? (car cursor) #\")
            (helper (cdr cursor) (append '(#\; #\t #\o #\u #\q #\&) racc))
          )
          ((eq? (car cursor) #\')
            (helper (cdr cursor) (append '(#\; #\9 #\3 #\# #\&) racc))
          )
          ((eq? (car cursor) #\&)
            (helper (cdr cursor) (append '(#\; #\p #\m #\a #\&) racc))
          )
          ((eq? (car cursor) #\<)
            (helper (cdr cursor) (append '(#\; #\t #\l #\&) racc))
          )
          (else
            (helper (cdr cursor) (cons (car cursor) racc))
          )
        )
      )
    )
  )
))


(define make-anchor (with-docstring
  "create an HTML-style anchor. When clicked, the target of the anchor, <link>, is opened in a new browser window."
  (lambda (link inner)
    (with-xml-tag `(sanchor href ,link) inner)
  )
))

(define make-eval-anchor (with-docstring
  "create a eval anchor with the action <form> and the inner XML text <inner>. Clicking an eval anchor causes <inner> to be evaluated without echoing <form> to the shell."
  (lambda (form inner)
    (with-xml-tag `(eval-anchor form ,(object->string form)) inner)
  )
))

(define make-shortcut-string-anchor (with-docstring
  "create a shortcut anchor with the evaluation text <form-string> and the inner XML text <inner>. Clicking a shortcut anchor causes <form-string> to be echoed to the terminal and <inner> to be evaluated. Shortcut anchors are designed to teach users about interesting forms that they might want to themselves type later."
  (lambda (form-string inner)
    (with-xml-tag `(shortcut-anchor form ,(xml-escape-string form-string))
      inner
    )
  )
))

(define make-shortcut-anchor (with-docstring
  "create a shortcut anchor with the action <form> and the inner XML text <inner>. Clicking a shortcut anchor has the same effect as typing its <form> at the interactive prompt. <form> is echoed to the shell and <inner> is evaluated."
  (lambda (form inner)
    (make-shortcut-string-anchor (object->string form) inner)
  )
))

(define-global xml-html-fragment (cons 'special 'xml-html-fragment))
(define make-html-fragment (with-docstring
  "accept <fragment-string> and pass in through to the browser as HTML. Proverb ordinarily requires markup to be in strict XHTML. This lets you pass along some HTML you found lying on the street."
  (lambda (fragment-string)
    (vector xml-html-fragment fragment-string)
  )
))

(define-global xml-xhtml (cons 'special 'xml-xhtml))
(define make-xhtml (with-docstring
  "accept arbitrarily many expressions and populate an extended XHTML tree"
  (lambda expressions
    (vector xml-xhtml expressions)
  )
))

(define-global xml-raw (cons 'special 'xml-raw))
(define make-raw (with-docstring
  "accept an arbitrary number of strings and use their concatenation as raw XML"
  (lambda raw-strings
    (vector xml-raw (apply string-append raw-strings))
  )
))

(define make-raw-text (with-docstring
  "accept an arbitrary number of strings and use their concatenation as raw XML text"
  (lambda raw-strings
    (apply make-raw raw-strings)
  )
))

(define-global xml-text (cons 'special 'xml-text))
(define make-text (with-docstring
  "accept a string of text and protect it from XML parsing. Text in <text-string> will still by styled by the surrounding XML context."
  (lambda (text-string)
    (vector xml-text text-string)
  )
))

(define make-xml-tag (with-docstring
  "construct a string representing an XML open tag for <tagname>. <rest> is a list of matched attribute names as symbols and Scheme objects as attribute values. Ex: (make-xml-tag 'img 'src \"foo.jpg\" 'width 64) yields <img src='foo.jpg' width='64'>. This form is can be passed directly to with-xml-tag. attribute values will be escaped as necessary for encapsulation."
  (lambda (tagname . rest)
  (string-append (apply string-append "<" (symbol->string tagname) " " (let helper ((cursor rest) (racc '()))
    (if (null? cursor)
      (reverse racc)
      (helper (cddr cursor) (cons (string-append
                                    (object->string (car cursor))
                                    "='"
                                    (xml-quote-escape-string (cond
                                      ((string? (cadr cursor)) (cadr cursor))
                                      (else (object->string (cadr cursor)))))
                                    "' ") racc))
    )
  )) ">")
  )
))

(define with-xml-tag (with-docstring
 "accept an opening tag expression (like \"<a href='foo'>\") and a list of scheme expressions and produce the xml object with body properly embedded in tag-expression. You can also now pass a list as the tag expression to have make-xml-tag applied."
  (lambda (tag-expression . body)
    (cond
      ((string? tag-expression)
        (if (and
              (> (string-length tag-expression) 2)
              (equal? (string-ref tag-expression 0) #\<)
              (equal? (string-ref tag-expression (- (string-length tag-expression) 1)) #\>)
            )
          (let ((t (studio-string-split (substring tag-expression 1 (- (string-length tag-expression) 1)) #\space)))
            (make-raw (apply string-append (cons tag-expression (cons "" (append (map xmlize-value body) `("" "</" ,(car t) ">"))))))
          )
         (error "bogus tag")
        )
      )
      ((symbol? tag-expression)
        (apply with-xml-tag (make-xml-tag tag-expression) body)
      )
      ((list? tag-expression)
        (apply with-xml-tag (apply make-xml-tag tag-expression) body)
      )
      (else (error "bogus tag"))      
    )
  )
))

(define xmlize (with-docstring
  "accept a Scheme value <value> and return an XML document as a string that presents <value>. The XML namespace is a non-standard one used by Proverb for S-expressions. The Proverb distribution includes an XSL file for translating this back into S-expressions."
  (lambda (value)
    (string-append
      "<?xml version=\"1.0\"?>\n"
      "<?xml-stylesheet type=\"text/xsl\" href=\"/sexp.xsl\"?>\n"
      "<sexp xmlns=\"http://www.w3.org/2000/sexp\">\n"

      (xmlize-value value)

      "</sexp>\n"
    )
  )
))

(define number->xml-string (with-docstring "convert a Scheme number <num> to a string for XML serialization."
  (lambda (num)
    (number->string num 10)
  )
))

(define xml-sentinels (list
  xml-html-fragment
  xml-xhtml
  xml-raw
  xml-text
))

(define xmlize-value (with-docstring
  "make an XML fragment from the Scheme value <value>. Uses the same namespace as xmlize. Scheme values created with make-raw, make-text, etc use additional non-sexp tags"
  (lambda (value)
  (cond
    ((and (vector? value) (> (vector-length value) 1)) (cond
      ((eq? (vector-ref value 0) xml-xhtml)
        (string-append
          "<xhtml>"
          (apply string-append (map xmlize-value (vector-ref value 1)))
          "</xhtml>"
        )
      )
      ((eq? (vector-ref value 0) xml-raw)
        ; just blast the raw fragment!
        (vector-ref value 1)
      )
      ((eq? (vector-ref value 0) xml-text) (let ()
        (define p (object-property value 'charset))
        (if p 
          (string-append "<text charset='" (xml-escape-string p) "'>" (xml-escape-string (vector-ref value 1)) "</text>")
          (string-append "<text>" (xml-body-escape-string (vector-ref value 1)) "</text>")
        )
      ))
      (else (apply string-append "<vector>" (append
        (map xmlize-value (vector->list value))
        '("</vector>")
      )))
    ))
    ((vector? value)
      (apply string-append "<vector>" (append
        (map xmlize-value (vector->list value))
        '("</vector>")
      ))
    )
    ((null? value)
      "<list/>"
    )
    ((list? value)
      (cond
        ((and (eq? (car value) 'quote) (eq? (length value) 2))
          (string-append "<quote>" (xmlize-value (cadr value)) "</quote>")
        )
        ((and (eq? (car value) 'quasiquote) (eq? (length value) 2))
          (string-append "<quasiquote>" (xmlize-value (cadr value)) "</quasiquote>")
        )
        ;; this is really a pair masquerading as a list
        ;((and (pair? (cdr value)) (member (cadr value) xml-sentinels))
        ;  (string-append "<pair>" (xmlize-value (car value)) (xmlize-value (cdr value)) "</pair>")
        ;)
        (else
          (apply string-append "<list>" (append
            (map xmlize-value value)
            '("</list>")
          ))
        )
      )
    )
    ((pair? value)
      (string-append "<pair>" (xmlize-value (car value)) (xmlize-value (cdr value)) "</pair>")
    )
    ((hash-table? value)
      (apply string-append "<hash>" (hash-fold (lambda (k v acc)
        (cons (xmlize-value (vector k v)) acc)
      ) '("</hash>") value))
    )
    ((string? value) (let ()
      (define p (object-property value 'charset))
      (if p 
        (string-append "<string charset='" (xml-escape-string p) "'>" (xml-escape-string value) "</string>")
        (string-append "<string>" (xml-body-escape-string value) "</string>")
      )
    ))
    ((symbol? value)
      (string-append "<symbol name=\"" (xml-quote-escape-string (symbol->string value)) "\"/>")
    )
    ((boolean? value)
      (if value "<boolean value=\"true\"/>" "<boolean value=\"false\"/>")
    )
    ((number? value)
      (string-append "<number value=\"" (number->xml-string value) "\"/>")        
    )
    ((record? value)
      (string-append "<text>" (xml-body-escape-string (object->string value)) "</text>")
    )
    ((unspecified? value)
      "<unspecified/>"
    )
    ; replace with a real tag
    ((char? value)
      (string-append "<char code='" (number->string (char->integer value)) "'>" (xml-body-escape-string (object->string value)) "</char>")
    )
    (else
      (string-append "<text>" (xml-body-escape-string (object->string value)) "</text>")
    )
  ) 
  )
))

)
