/*
 * cnl-namespace
 *
 * utility for creating a per-job namespace for CNL jobs
 *
 * Written 2008 by Daniel Ridge
 *
 * Institute for Defense Analyses
 * Center for Computing Sciences
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>

#include <signal.h>

#ifdef HAVE_UNSHARE
#include <sched.h>
#endif

#include <sys/types.h>
#include <sys/mount.h>
#include <sys/wait.h>

#include <linux/fs.h>

#ifndef MNT_DETACH
#define MNT_DETACH 0x02
#endif

/* cnl-namespace chroot-dir bind-from bind-to bind-from bind-to -- prog */

static char *chroot_dir;
int bindc; 
char **bindv;
DIR **binddirv;

int
binder(char *prefix,char *from,char *to) {
int retval;
char *target;

  target=(char *)malloc(strlen(prefix)+strlen(to)+2);
  if(target == NULL) {
    fprintf(stderr,"alloc failed!\n");
    exit(1);
  }
  target[0]=0;
  strcat(target,prefix);
  strcat(target,"/");
  strcat(target,to);

  retval=mount(from,target,"none",MS_BIND,NULL);

  if(retval != 0) {
    fprintf(stderr,"MOUNT FAILED OF %s to %s!\n",from,target);
  }

  free(target);

  return retval; 
}

int
unbinder(char *prefix,char *to) {
int retval;
char *target;

  target=(char *)malloc(strlen(prefix)+strlen(to)+2);
  if(target == NULL) {
    fprintf(stderr,"alloc failed!\n");
    exit(1);
  }
  target[0]=0;
  strcat(target,prefix);
  strcat(target,"/");
  strcat(target,to);

  retval=umount2(target,MNT_FORCE);

  if(retval != 0) {
    /* detach umounting makes cwd a bit screwy in the child. Don't do needlessly. */
    retval=umount2(target,MNT_DETACH);
    if(retval != 0) {
      fprintf(stderr,"serious umount failure of %s\n",target);
    } else {
    }
  }

  free(target);

  return retval;
}

void *sighandler(int signum) {
  int x;

  printf("SIGINT %d\n",signum);

  /* do the unbinds */
  for(x=bindc;x>0;x-=2) {
    unbinder(chroot_dir,bindv[x-1]);
  }

  exit(1);
  return NULL;
}

int
doit(char *chroot_dir,char **sub_argv) {
  int child_pid;
  int retval;
  char *retbuf;
  char current_dir[4096];

  child_pid=fork();
  if(child_pid == 0) {

    /* we are the child */

    /* do the chroot */
    if(!chroot_dir) {
      fprintf(stderr,"no chroot dir!\n");
      exit(1);
    }
    retval=chroot(chroot_dir);
    if(retval != 0) {
      fprintf(stderr,"chroot failed!\n");
      exit(1);
    }

#ifdef HAVE_UNSHARE
   retval=unshare(CLONE_FS|CLONE_NEWNS);
   if(retval != 0) {
     fprintf(stderr,"unshare failed!\n");
     exit(1);
   }
#endif

    /* drop our privileges */
    if(getgid() != getegid()) {
      setegid(getgid());
    }

    if(getuid() != geteuid()) {
      seteuid(getuid());
    }

    retbuf=getcwd(current_dir,4095);
    if(retbuf == NULL) {
      retbuf="/";
    } else {
      retbuf;
    }

    /* change dir */
    retval=chdir(retbuf);
    if(retval != 0) {
      retval=chdir("/");
      if(retval != 0) {
        fprintf(stderr,"chdir failed!\n");
        exit(1);
      }
    }

    execvp(sub_argv[0],sub_argv);

    exit(1);

  } else if(child_pid > 0) {
    /* we are the parent */
    return child_pid;
  } else {
    /* there was an error */
    return -1;
  }

}

int
main(int argc,char **argv) {
  int x;
  char **sub_argv;
  int sub_start;
  int retval;
  int child;

  if(argc<3 || !argv || !argv[1]) {
    fprintf(stderr,"invalid invocation!\n");
    exit(1);
  }

  chroot_dir=argv[1];

  sub_start=0;
  bindc=0;
  bindv=argv+2;
  for(x=2;x<argc-1;x++) {
    if(!strcmp(argv[x],"--")) {
      sub_start=x+1;
      break;
    } else {
      if((x+1 == argc-1) || !strcmp(argv[x+1],"--")) {
        fprintf(stderr,"IMBALANCED BIND ARGUMENTS!\n");
        exit(1); 
      }
      x++;
      bindc+=2;
    }
  }


  if(sub_start == 0) {
    fprintf(stderr,"invalid invocation");
    exit(1);
  }
  if(!argv[sub_start]) {
    fprintf(stderr,"invalid invocation");
    exit(1);
  }

  sub_argv=(char **)malloc(sizeof(char *)*(argc+2));
  if(!sub_argv) {
    fprintf(stderr,"alloc fail!\n");
    exit(1);
  }

  for(x=sub_start;x<argc;x++) {
    sub_argv[x-sub_start]=argv[x]; 
  }
  sub_argv[x-sub_start]=NULL;

  {
    int sigaction_retval;
    struct sigaction default_sigaction;
    default_sigaction.sa_handler=&sighandler;
    default_sigaction.sa_flags=0;
    //default_sigaction.sa_mask=0;
    
    sigaction_retval=sigaction(SIGINT,&default_sigaction,NULL);
    if(sigaction_retval != 0) {
      fprintf(stderr,"sigaction failure!\n"); 
      exit(1);
    }
  }

  /* do the binds */
  //binddirv=(DIR *)malloc(sizeof(DIR *)*(bindc+1));
  for(x=0;x<bindc;x+=2) {
    binder(chroot_dir,bindv[x],bindv[x+1]);
    //binddirv[x+1]=opendir(bindv[x+1]);
  }

  child=doit(chroot_dir,sub_argv);

  if(child > 0) {
    retval=waitpid(child,NULL,0);
    if(retval == -1) {
      fprintf(stderr,"WAITPID ERROR!\n");
    }
  }

  /* do the unbinds */
  for(x=bindc;x>0;x-=2) {
    unbinder(chroot_dir,bindv[x-1]);
  }

  exit(1);
}
