
/*
 *  An experimental radio station streaming widget that uses browser-based
 *  geolocation to map FM frequencies to streaming servers.
 *
 *  Example invocations:
 *    radio_widget(shell,"WAMU");
 *    radio_widget(shell,88.5);
 *
 */

function radio_widget(shell) {
  var hint;
  
  if(window.localStorage) {
    var i=localStorage.getItem('jsshell_radio_widget_hint');
    if(i) {
      hint=i;
    } else {
      hint=88.5;
    }
  }
  
  var stations=new Array;

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */
/*  Vincenty Inverse Solution of Geodesics on the Ellipsoid (c) Chris Veness 2002-2009            */
/*  this inclusion is LGPL and was obtained from:                                                 */
/*    http://www.movable-type.co.uk/scripts/latlong-vincenty.html                                 */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

toRad = function(n) {  // convert degrees to radians
  return n * Math.PI / 180;
}

/*
 * Calculate geodesic distance (in m) between two points specified by latitude/longitude 
 * (in numeric degrees) using Vincenty inverse formula for ellipsoids
 */
function distVincenty(lat1, lon1, lat2, lon2) {
  var a = 6378137, b = 6356752.3142,  f = 1/298.257223563;  // WGS-84 ellipsiod
  var L = toRad(lon2-lon1);
  var U1 = Math.atan((1-f) * Math.tan(toRad(lat1)));
  var U2 = Math.atan((1-f) * Math.tan(toRad(lat2)));
  var sinU1 = Math.sin(U1), cosU1 = Math.cos(U1);
  var sinU2 = Math.sin(U2), cosU2 = Math.cos(U2);
  
  var lambda = L, lambdaP, iterLimit = 100;
  do {
    var sinLambda = Math.sin(lambda), cosLambda = Math.cos(lambda);
    var sinSigma = Math.sqrt((cosU2*sinLambda) * (cosU2*sinLambda) + 
      (cosU1*sinU2-sinU1*cosU2*cosLambda) * (cosU1*sinU2-sinU1*cosU2*cosLambda));
    if (sinSigma==0) return 0;  // co-incident points
    var cosSigma = sinU1*sinU2 + cosU1*cosU2*cosLambda;
    var sigma = Math.atan2(sinSigma, cosSigma);
    var sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma;
    var cosSqAlpha = 1 - sinAlpha*sinAlpha;
    var cos2SigmaM = cosSigma - 2*sinU1*sinU2/cosSqAlpha;
    if (isNaN(cos2SigmaM)) cos2SigmaM = 0;  // equatorial line: cosSqAlpha=0 (§6)
    var C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha));
    lambdaP = lambda;
    lambda = L + (1-C) * f * sinAlpha *
      (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)));
  } while (Math.abs(lambda-lambdaP) > 1e-12 && --iterLimit>0);

  if (iterLimit==0) return NaN  // formula failed to converge

  var uSq = cosSqAlpha * (a*a - b*b) / (b*b);
  var A = 1 + uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)));
  var B = uSq/1024 * (256+uSq*(-128+uSq*(74-47*uSq)));
  var deltaSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)-
    B/6*cos2SigmaM*(-3+4*sinSigma*sinSigma)*(-3+4*cos2SigmaM*cos2SigmaM)));
  var s = b*A*(sigma-deltaSigma);
  
  s = s.toFixed(3); // round to 1mm precision
  return s;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

/*
 * method for parsing degrees or lat/long values to numeric degrees
 *
 * this is very flexible on formats, allowing signed decimal degrees, or deg-min-sec suffixed by 
 * compass direction (NSEW). A variety of separators are accepted (eg 3º 37' 09"W) or fixed-width 
 * format without separators (eg 0033709W). Seconds and minutes may be omitted. (Minimal validation 
 * is done).
 */
var parseDeg = function(str) {
  if (!isNaN(str)) return Number(str);                 // signed decimal degrees without NSEW

  var degLL = str.replace(/^-/,'').replace(/[NSEW]/i,'');  // strip off any sign or compass dir'n
  var dms = degLL.split(/[^0-9.]+/);                     // split out separate d/m/s
  for (var i in dms) if (dms[i]=='') dms.splice(i,1);    // remove empty elements (see note below)
  switch (dms.length) {                                  // convert to decimal degrees...
    case 3:                                              // interpret 3-part result as d/m/s
      var deg = dms[0]/1 + dms[1]/60 + dms[2]/3600; break;
    case 2:                                              // interpret 2-part result as d/m
      var deg = dms[0]/1 + dms[1]/60; break;
    case 1:                                              // decimal or non-separated dddmmss
      if (/[NS]/i.test(str)) degLL = '0' + degLL;       // - normalise N/S to 3-digit degrees
      var deg = dms[0].slice(0,3)/1 + dms[0].slice(3,5)/60 + dms[0].slice(5)/3600; break;
    default: return NaN;
  }
  if (/^-/.test(str) || /[WS]/i.test(str)) deg = -deg; // take '-', west and south as -ve
  return deg;
}
// note: whitespace at start/end will split() into empty elements (except in IE)

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */


  var d=document.createElement('div');

  var make_station = function(radio_url,radio_art_url) {
    return function() {
      var art=document.createElement('img');
      art.setAttribute('src',radio_art_url);

      var a=document.createElement('a');
      a.setAttribute('href',radio_url);
      a.appendChild(document.createTextNode('click to play'));
      d.appendChild(a);

      d.appendChild(art);
      d.appendChild(document.createElement('br'));

      var audio=document.createElement('audio');
      audio.setAttribute('src',radio_url);
      audio.setAttribute('controls',true);
      audio.setAttribute('autoplay',true);

      d.appendChild(audio);
    };
  };

  var do_wamu = make_station("http://wamu-1.streamguys.com:80/listen.pls","http://wamu.org/g/hdr/day/wamu885-back.gif");
  var do_kqed = make_station("http://www.kqed.org/radio/listen/kqedradio.pls","http://www.kqed.org/assets/img/nav/masthead-radio-445x51.gif");
  var do_weta = make_station("http://stream.weta.org:8000/listen.pls","http://www.weta.org/sites/all/themes/weta/images/logo.png");

  var mkstation = function(lat,lon,radius,freq,callsign,handler) {
    var station=new Object;
    station.lat=lat;
    station.lon=lon;
    station.radius=radius;
    station.freq=freq;
    station.callsign=callsign;
    station.handler=handler; 
    return station;
  };

  // there is probably a way to use http://www.radio-locator.com for this stuff
  stations.push(mkstation(37.771258,-122.44812,150000,88.5,"KQED",do_kqed));
  stations.push(mkstation(38.936379,-77.089176,150000,88.5,"WAMU",do_wamu));
  stations.push(mkstation(38.89167, -77.13194,150000,90.9,"WETA",do_weta));

  var handle_hint = function(hint) {
    if(typeof(hint) == "number" ) {
      if(navigator.geolocation) {
        var autotune=document.createTextNode('autotuning...');
        d.appendChild(autotune);
        navigator.geolocation.getCurrentPosition(function(position) {
          var i;
          for(i in stations) { 
            var station=stations[i];
            if(station.freq == hint) {
              var distance = distVincenty(position.coords.latitude,position.coords.longitude,station.lat,station.lon);
              if(distance < station.radius) {
                autotune.parentNode.removeChild(autotune);
                station.handler();
              }
            }
          }
        },function() { 
          alert('no station');
          //autotune.parentNode.removeChild(autotune);
        });
     
      } else {
        do_wamu();
      }
    } else if(typeof(hint) == "string") {
      var i,n;
      n=parseFloat(hint);
      if(n) {
        return handle_hint(n);
      } else {
        for(i in stations) { 
          var station=stations[i];
          if(station.callsign == hint) {
            station.handler();
          }
        }
      } 
    }
  }

  handle_hint(hint);

  var i=document.createElement('input');
  i.type='text';
  i.value=hint;
  i.addEventListener('change',function(e) {
    handle_hint(i.value);
  },false);

  d.appendChild(i);

  return libreplete.paint(d);
}
