#!/bin/bash
exec `which guile` -e scheduler-main -s "$0" "$@"
!#

;
; A little scheduler program intended for batch
; launching programs like Proverb
;


(define max-outstanding 128)
(define max-per-node 4)

(define xmake-hash-table (lambda (v) 
  (if (string=? "6" (minor-version))
    (make-hash-table v)
    (make-hash-table)
  )
))

(define make-index-list (lambda (s)
  (let helper ((c s) (acc '()))
    (if (zero? c)
      acc
      (let ((n (- c 1)))
        (helper n (cons n acc))
      )
    )
  )
))

(define do-experiment (lambda (workload data max-outstanding nodes)
  (define t (xmake-hash-table 32000))
  (define n (xmake-hash-table nodes))
  (define find-a-free-node (lambda ()
    (let helper ((c 0))
      (if (eq? c nodes)
        #f
        ; else
        (let ()
          (define l (hashq-ref n c))
          (if (not l)
            (begin
              (hashq-set! n c 0)
              c
            )
            (if (< l max-per-node)
              c
              (helper (+ c 1))
            )
          )
        )
      )  
    )
  ))
  (let helper ((cursor data) (outstanding 0))
    (define waiter (lambda (must-reap)
      "must reap is the number that must be reaped to return. we call the helper ourselves."
      (if (eq? must-reap 0)
        ; strange case
        0
        ; else
        (let wait-helper ((reaped 0) (may-reap outstanding))
          (define reap-helper (lambda (r blocking?)
            (define miss (hashq-ref t (car r)))
            (hashq-set! n (cdr miss) (- (hashq-ref n (cdr miss)) 1))
            (if (not (eq? (cdr r) 0))
              (let ()
                (hashq-remove! t (car r))
                (helper (append cursor (list (car miss))) (- outstanding (+ reaped 1)))
              )
              (if blocking?
                (if (>= (+ reaped 1) must-reap)
                  (helper cursor (- outstanding (+ reaped 1)))
                  (wait-helper (+ reaped 1) (- may-reap 1))
                )
                (wait-helper (+ reaped 1) (- may-reap 1))
              )
            )
          ))
          (if (> may-reap 0)
            (let ()
              (define result (waitpid -1 WNOHANG))
              (if (not (eq? (car result) 0))
                ; we got a real child
                (reap-helper result #f)
                ; we didn't
                (if (eq? reaped 0)
                  ; wait until any child dies
                  (let ()
                    (define blocking-result (waitpid -1 0))
                    (if (not (eq? (car blocking-result) 0))
                      ; we got a real child
                      (reap-helper blocking-result #t)
                      ; we didn't
                      (wait-helper reaped may-reap)
                    )
                  )
                  ; otherwise launch jobs
                  (if (>= reaped must-reap)
                    (helper cursor (- outstanding reaped))
                    (wait-helper reaped may-reap)
                  )
                )
              )
            )
            ; else
            (helper cursor (- outstanding reaped))
          )
        )
      )
    ))
    (if (null? cursor)
      (if (eq? outstanding 0)
        'done
        (waiter outstanding)
      )
      (if (> outstanding max-outstanding)
        ; then launch no more until we wait
        (let ()
          (waiter 1)
        )
        ; else
        (let ()
          (define node (find-a-free-node))
          (if node
            (begin
              (let ()
                (define p (primitive-fork))
                (case p
                  ((0)
                    (workload (car cursor) node)
                  )
                  ((-1)
                    (error "fork bombed")
                    (exit 1)
                  )
                  ; else we're the parent
                  (else
                    (hashq-set! t p (cons (car cursor) node))
                    (hashq-set! n node (+ (hashq-ref n node) 1))
                    (helper (cdr cursor) (+ outstanding 1))
                  )
                )
              )
            )
            ; else no node
            (let ()
              (waiter 1)
            )
          )
        )
      )
    )
  )
))

(define scheduler-main (lambda (args)
  (if (eq? (length args) 1)
    (let ()
      (format #t "Usage: ~a SIZE NODES WORKLOAD\n" (basename (car args)))
      (exit 1)
    )
    (let ()
      (define size (string->number (cadr args)))
      (define nodes (string->number (caddr args)))
      (if (not size)
        (begin
          (format #t "Size ~a wasn't acceptable\n" (cadr args))
          (exit 1)
        )
      )
      (do-experiment (lambda (rank node)
        (define retval (system (format #f "~a ~a ~a ~a" (cadddr args) rank size node)))
        (if (> retval 255)
          (begin
            (primitive-exit (quotient retval 256))
          )
          (primitive-exit retval)
        )
      ) (make-index-list size) max-outstanding nodes)
      (exit 0)
    )
  )
))
